/* 
 * Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; as a special exception the author gives
 * unlimited permission to copy and/or distribute it, with or without 
 * modifications, as long as this notice is preserved.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Michael Schulz (see also AUTHORS and THANKS for more)
 * Birth: 12 Juli 1998
 *
 * $Id: hook.h,v 1.25 2004/02/25 14:28:05 micha Exp $
 */

/* -------------------------------------------------------------------------
 * @addtogroup Hook Hook - A callback system
 * @{
 * @author Michael Schulz
 *
 * @} */
 
/* -------------------------------------------------------------------------
 * @addtogroup HookCore
 * @ingroup Hook
 * @{ */

#ifndef _HOOK_H
#define _HOOK_H

/* -- INCLUDES ---------------------------------------------------------- */
#include "cdg.h"
#include "stdarg.h"

/* -- MACROS ------------------------------------------------------------ */

#ifndef SWIG
/* ------------------------------------------------------------------------ 
 * callback after building constraintnodes.
 */
#define HOOK_CNBUILDNODES	0

/* ------------------------------------------------------------------------ 
 * callback after evaluation of constraints.
 */
#define HOOK_EVAL		1

/* ------------------------------------------------------------------------ 
 * callback in netsearching.
 */
#define HOOK_NSSEARCH		2

/* ------------------------------------------------------------------------ 
 * callback for printfing.
 */
#define HOOK_PRINTF		3

/* ------------------------------------------------------------------------ 
 * callback for flushing the output channels.
 */
#define HOOK_FLUSH		4

/* ------------------------------------------------------------------------ 
   notification that parses, constraint nets etc. have become invalid
   because of grammar changes.
 */
#define HOOK_RESET              5

/* ------------------------------------------------------------------------ 
 * callback for user interaction in the gls module. 
 */
#define HOOK_GLSINTERACTION	6

/* ------------------------------------------------------------------------ 
 * callback in to get a string from the user.
 */
#define HOOK_GETS		7

/* ------------------------------------------------------------------------ 
 * callback to indicate progress in a heavy computation.
 */
#define HOOK_PROGRESS		8

/* ------------------------------------------------------------------------ 
 * callback to hand over partial results.
 */
#define HOOK_PARTIALRESULT      9                                          

/* ------------------------------------------------------------------------ 
 * callback for user interaction in the incrementalcompletion module.
 */
#define HOOK_ICINTERACTION     10


#endif

/* ------------------------------------------------------------------------ 
 * identification of the \c HINT channel.
 * \todo The \c HINT channel is bogus and never used.
 */
#define CDG_HINT (1L << 0)

/* ------------------------------------------------------------------------ 
 * identification of the \c INFO channel.
 * Most of the output of the \c CDG library is tagged as \c INFO. A message
 * to the \c INFO channel is \e informal. 
 */
#define CDG_INFO (1L << 1)

/* ------------------------------------------------------------------------ 
 * identification of the \c WARNING channel.
 * A message to this channel informs about an unexpected situation. Operations
 * continue. If a normal continuation is not possible the message should
 * be emitted on the \c ERROR channel.
 */
#define CDG_WARNING (1L << 2)

/* ------------------------------------------------------------------------ 
 * identification of the \c PROLOG channel.
 * \todo The \c PROLOG channel is bogus and never used.
 */
#define CDG_PROLOG (1L << 3)

/* ------------------------------------------------------------------------ 
 * identification of the \c EVAL channel.
 * \todo The \c EVAL channel is bogus and never used.
 */
#define CDG_EVAL (1L << 4)

/* ------------------------------------------------------------------------ 
 * identification of the \c SEARCHRESULT channel.
 * Results in a parser are emited on this channel. All parsing falvours
 * should obey to this channel semantics. Most of them do.
 */
#define CDG_SEARCHRESULT (1L << 5)

/* ------------------------------------------------------------------------ 
 * identification of the \c PROFILE channel.
 * This channel is used to emit profiling information, that is time statistics
 * on computations.
 */
#define CDG_PROFILE (1L << 6)

/* ------------------------------------------------------------------------ 
 * global hook flag.
 * This flag switches on/off the complete hook system.
 * \todo By no means this is an output channel. This define should be 
 * removed and replaced with a proper variable that is altered by the 
 * hook command.
 */
#define CDG_HOOK (1L << 7)

/* ------------------------------------------------------------------------ 
 * identification of the \c ERROR channel.
 * This channel transmits messages about unexpected events that lead to
 * an abortion of further computation. Compare to \c WARNING.
 */
#define CDG_ERROR (1L << 8)

/* ------------------------------------------------------------------------ 
 * identification of the \c DEBUG channel.
 * This message increases the normal output verbosity by offering debug
 * information.
 */
#define CDG_DEBUG (1L << 9)

/* ------------------------------------------------------------------------ 
 * identification of the \c DEFAULT channel.
 * Unclassified messages of messages that should only be supressed in very
 * rare cases are emitet on the \c DEFAULT channel.
 */
#define CDG_DEFAULT (1L << 10)

/* ------------------------------------------------------------------------ 
 * identification of the \c PROGRESS channel.
 * Messages about progress are emited on this channel and might be 
 * redirected over the \c HOOK_PROGRESS callback.
 */
#define CDG_PROGRESS (1L << 11)

/* ------------------------------------------------------------------------ 
 * identification of the \c PROGRESS channel.
 * All xml output goes in this channel. See the write module for more 
 * infofmation on xml output.
 */
#define CDG_XML (1L << 12)


/* -- TYPE DEFINITIONS -------------------------------------------------- */
#ifndef SWIG

/* ------------------------------------------------------------------------ 
 * type of a callback function.
 * This function is installed into a callback hook and executed whenever
 * this callback is triggered. This is only a indicator definition without
 * any grants on the arguments. But be aware of the arguments that are 
 * actually needed by every callback.
 */
typedef void HookFunction();

/* ------------------------------------------------------------------------ 
 * callback information.
 * This structure bundles the state of a callback. 
 */
typedef struct {
   int no;			       /**< index in vector hkHooks */
   unsigned long count;		       /**< number of past invocations */
   String name;			       /**< identifier required to be unique among all hooks*/
   String cmd;			       /**< command to be executed in tcl */
   Boolean active;		       /**< flag indicating the status of the hook */
   HookFunction *function;	       /**< function which is called with a va_list
				            argument */
} HookStruct;
typedef HookStruct *Hook; /**< type of a Hook */
#endif

/* -- VARIABLES --------------------------------------------------------- */

#ifdef SWIG
#ifdef OLD_SWIG
%readwrite
#else
%mutable;
#endif
#endif
extern unsigned long int hkVerbosity;
#ifdef SWIG
#ifdef OLD_SWIG
%readonly;
#else
%immutable;
#endif
#endif
extern Vector hkHooks;

/* -- FUNCTIONS --------------------------------------------------------- */

#ifndef SWIG
extern void hkInitialize(void);
extern void hkFinalize(void);
extern int hkFindNoOfHook(String name);
extern void cdgExecHook(int hookNo, ...);
extern void cdgFlush(void);
extern void cdgPrintf(int mode, String format, ...);
extern void cdgGetString(String buffer, int size); 
extern void hkCallback(String name, Boolean *var);
extern Boolean hkValidate(String name, String value, Boolean *var);
#endif

/* ---------------------------------------------------------------------- */
/** @} */
#endif /* don't insert anything after this #endif */

