/* 
 * File: parse.h
 * Author: Jens Grabarske
 * Purpose: data structures for the solution trees
 * Birth: 9/8/99
 *
 * $Id: parse.h,v 1.58 2004/02/19 15:27:32 engelman Exp $
 */

#ifndef PARSE_H
#define PARSE_H

/* -- INCLUDES ---------------------------------------------------------- */
#include <configure.h>
#include <my_time.h>

#include <blah.h>
#include "constraintnet.h"
#include "cdg.h"
#include "eval.h"
#include "levelvalue.h"

/* -- MACROS ------------------------------------------------------------ */
#ifndef SWIG

#define PC_SILENT     0     /* do not produce any messages      */
#define PC_SUMMARY    1     /* produce only summary report      */
#define PC_MEASURES   2     /* produce only measure information */
#define PC_ALL        3     /* produce detailed information     */

#define PC_STRICT     0     /* words not found at end of tested word list are errors */
#define PC_NONSTRICT  1     /* words not found at end of tested word list are not errors */

#endif
 
/* -- TYPE DEFINITIONS -------------------------------------------------- */
typedef struct {
  String  id;                /* id of the parse in the list of global parses */

  String  latticeId;         /* Name of the word graph      */
  String  userName;          /* Who did the run?            */
  time_t  dateOfCreation;    /* When?                       */
  List    grammarFiles;      /* With which grammars?        */
  String  searchStrategy;    /* With which search strategy? */
  String  comment;           /* comment about the solution  */

  int     solutionNo;        /* This is solution number X   */
  int     noOfSolutions;     /* from Y                      */
  
  Number  score;             /* What score did it reach?    */
  Boolean optimal;           /* Is it optimal?              */
  List    violatedConstraints; /* List of violated constraints (Strings) */

  Vector  words;             /* Vector of all word nodes        */
  List    levels;            /* List of the used levels (Strings) */
  List    labels;            /* List of the used labels (Strings) */

  Vector verticesStructure;  /* holds the elder node for each node i in level j */
  Vector    verticesLabels;  /* holds the label for every vertex (i, j) */

  int           evalUnary;       /* Number of unary evaluations         */
  int           evalBinary;      /* Number of binary evaluations        */
  unsigned long numberOfHits;     /* Score cache statistic               */
  unsigned long evaluationTime;   /* How long did the evaluation take?   */
  String        statisticComment; /* Comment about the statistic data    */

  /* The following parts are not required, but useful for many operations.
     They cannot be assumed to exist, and they are never saved or restored.
  */
  
  Vector LVs;                 /* holds the constituting LVs */
  LexemGraph lg;              /* lexeme graph used by the LVs */
  List violations;            /* List of full ConstraintViolation structures */
  Badness badness;            /* generalized score */
  Vector map;                 /* maps vertex no -> LV index */
  Vector rmap;                /* maps LV index -> vertex no */

  
} ParseStruct;
typedef ParseStruct *Parse;

typedef struct
{
  int     from;           /* Interval in the                       */
  int     to;             /* lattice graph                         */

  String  word;           /* Which word does this node represent?  */
  String  description;    /* == description of the currently corresponding LexiconItem */
} WordStruct;
typedef WordStruct *Word;

typedef struct {
  int errorextrawordref;  /* nr of errors: extra word in reference */
  int errorextrawordtest; /* nr of errors: extra word in test */
  int errorwordlex;       /* nr of errors: words have different lexical readings */
  int errorgovword;       /* nr of errors: governors are different words */
  int errorgovlex;        /* nr of errors: governors have different lexical readings */
  int erroredgelabel;     /* nr of errors: edges have different label */
  int matchedgebase;      /* nr of matches: edges are identical (not counting gov's) */
  int matchedgegov;       /* nr of matches: edges are identical (same gov's words) */
  int matchedgetot;       /* nr of matches: edges are identical (same gov's read.) */
  int totalextrawordref;  /* nr of checks: extra word in reference */
  int totalextrawordtest; /* nr of checks: extra word in test */
  int totalwordlex;       /* nr of checks: words have different lexical readings */ 
  int totalgovword;       /* nr of checks: governors are different words */
  int totalgovlex;        /* nr of checks: governors have different lexical readings */
  int totaledgelabel;     /* nr of checks: edges have different label */
  int totaledgebase;      /* nr of checks: edges are identical (not counting gov's) */
  int totaledgegov;       /* nr of checks: edges are identical (same gov's words) */
  int totaledgetot;       /* nr of checks: edges are identical (same gov's read.) */
  int totaledgetest;      /* total number of edges in test      */ 
  int totaledgeref;       /* total number of edges in reference */
  float accuracy;
  float recall;
  float fmeasure;
} ParseCmpStatStruct;
typedef ParseCmpStatStruct *ParseCmpStat;

typedef struct {
  Parse parse;		     /* the parse */
  AnnoEntry annoEntry;	     /* the annotation */
  Hashtable tries;	     /* maps levels to tries */
  Hashtable structHits;	     /* maps levels to strucural hits */
  Hashtable labelledHits;    /* maps levels to labelled hits */
  Hashtable lexicalHits;     /* maps levels to lexical hits */
  Hashtable strictHits;	     /* maps levels to strict hits */

  Vector lexErrors;       /* mark wrong lexemes */
  Vector labelErrors;     /* mark wrong labels */
  Vector structErrors;    /* mark wrong strucure */
  
} ParseVerificationStruct;
typedef ParseVerificationStruct *ParseVerification;
  
/* -- FUNCTIONS --------------------------------------------------------- */

/* NULL if no parse has been created, or a pointer to the last one */
extern Parse parseMostRecentlyCreatedParse;

extern void parseSwapLabel(Parse parse, int where, String label);
extern void parseShiftEdge(Parse parse, int where, int to);
extern void parseSwapWord(Parse parse, int where, String id);
extern Boolean parseIsAbstract(Parse parse);
extern Parse parseFind(String id);
extern int parseHeight(Parse parse, String levelname);
extern int parseWordHeight(Parse p, String levelname, int wordNo);
extern int parseWordDepth(Parse p, Level level, int wordNo);
extern String parseDate(Parse parse);
extern LevelValue parseGetLV(Parse parse, int where);
extern List parseGetLexemNodes(Parse parse);
extern List parseGetModifiers(Parse parse, Level level, int wordIndex);
extern Boolean parseDecorate(Parse p, LexemGraph lg, AnnoEntry ae);
extern void parseEval(Parse parse);
extern AnnoEntry parse2annotation(Parse parse);
extern Parse parseFromAnno(AnnoEntry ae);
extern void parseRegister(Parse p);
extern void parseMirror(Parse p);
extern int parseMap(Parse p, int lvindex);
extern int parseRmap(Parse p, int lvindex);
extern ParseVerification parseVerify(Parse p, AnnoEntry ae);
extern void parseVerificationDelete(ParseVerification pv);
extern int parseIndex(Parse p, String levelname, int wordno);
extern LexiconItem anno2le(Annotation an);
extern String parseOptimizeWord(Parse p, int wordno);
extern String parseOptimizeLabel(Parse p, String levelname, int wordno);

#ifndef SWIG
extern Parse parseClone(Parse parse);
extern void parseReEval(Parse parse, List changes);
extern int parseCompare(Parse ref, Parse test, ParseCmpStat stat, int verbosity, int strict);
       /* for verbosity and strict arguments see macros above */
extern Boolean parseEquals(Parse a, Parse b);
extern Parse parseNew(String id);
extern void parseDelete(Parse p);
extern Word newWord();
extern void deleteWord(Word w);
extern LexemNode parseFindLn(Parse p, Word word, AnnoEntry ae);
extern void parseComputeStructure(Parse p, Vector LVs, ByteVector isDeleted);
extern void parseStrip(Parse p);
extern void parsePrint(Parse p);
extern void parseFinalize();
extern Boolean parseIsUnderspecified(Parse p);
extern Parse parseFromLVs(ConstraintNet net, Vector lvs, String strategy);
extern int parseMinFrom(Parse p);
extern int parseMaxTo(Parse p);
extern void parseInitialize(void);
extern void parseFinalize(void);
extern ParseVerification parseVerificationNew(Parse p);
#endif


/* -- VARIABLES --------------------------------------------------------- */

/* ---------------------------------------------------------------------- */
#endif
