/*
 * File: predicates.c
 * Author: Ingo Schroeder
 * Purpose: CDG predicates
 * Birth: 17/3/97
 *
 * $Id: predicates.c,v 1.56 2004/02/19 15:27:33 engelman Exp $
 */

/* -- INCLUDES ---------------------------------------------------------- */
#include <configure.h>

#include <stdio.h>
#include <string.h>
#include "cdg.h"
#include "eval.h"
#include "predicates.h"
#include "increment.h"
#include "hook.h"
#include "chunker.h"

/* -- MACROS ------------------------------------------------------------ */
/* -- PROTOTYPES -------------------------------------------------------- */
/*
Boolean predNthCousin(List args, LexemGraph lg);
Boolean predAncestor(List args, LexemGraph lg);
*/
/* -- VARIABLES --------------------------------------------------------- */
CDGPredicateStruct eval_predicates[] =
  {
    {"start", predStart, "predStart", FALSE, "1",
     "returns TRUE if lattice starts with word"},
    {"stop", predStop, "predStop", FALSE, "1",
     "returns TRUE if lattice ends with word"},
    {"root", predRoot, "predRoot", FALSE, "1",
     "returns TRUE if argument is the root"},
    {"spec", predSpec, "predSpec", FALSE, "1",
     "returns TRUE if argument is specific"},
    {"nonspec", predNonSpec, "predNonSpec", FALSE, "1",
     "returns TRUE if argument is nonspecific"},
    {"print", predPrint, "predPrint", FALSE, ">0",
     "prints string or number argument and returns TRUE"},
    {"subsumes", predSubsumes, "predSubsumes", FALSE, "3",
     "returns TRUE if 2nd arguments subsumes 3rd"},
    {"exists", predExists, "predExists", FALSE, "1",
     "returns TRUE if feature exists in lexical entry"},
    {"compatible", predCompatible, "predCompatible", FALSE, "3",
     "equivalent to `( subsumes(a, b, c) | subsumes(a, c, b) )'"},
    {"chunk_head", predChunkHead, "predChunkHead", FALSE, "1",
     "returns TRUE if lattice is a chunk head"},
    {"occur", predOccur, "predOccur", FALSE, "1",
     "returns TRUE if the specified word form occurs in the sentence somewhere"},
    {"has", predHas, "predHas", TRUE, ">1",
     "returns TRUE if the LV has a child edge with the specified label"},
    {"is", predIs, "predIs", TRUE, ">1",
     "returns TRUE if the word is subordinated somewhere with the specified label"},
    {"cyclic", predCycle, "predCycle", TRUE, "1",
     "returns TRUE if the named level is cyclical"},
    {"between", predBetween, "predBetween", FALSE, "3",
     "test if one of the specified punctuation marks separates the two words"},
    {"under", predUnder, "predUnder", TRUE, "2",
     "test if word A is indirectly a modifier of word B"},
    {"connected", predConnected, "predConnected", TRUE, "2",
     "test if A and B have the same root"},
    /*
      {"nth_cousin", predNthCousin, "are there common ancestors not more then n levels back?"},
      {"ancestor_at_level", predAncestor, "are there common ancestors at level x or later?"}, 
    */   
    {(String) NULL, (BooleanFunction *) NULL, (String)NULL, FALSE, (String) NULL}
  };

/* -- FUNCTIONS --------------------------------------------------------- */

/* ----------------------------------------------------------------------
 * return a pointer to the function implementing a functor
 */
CDGPredicate predGetFunction(String functor)
{
  int i;

  for (i = 0; eval_predicates[i].functor != NULL; i++) {
    if (strcmp(functor, eval_predicates[i].functor) == 0) {
      return &eval_predicates[i];
      break;
    }
  }

  return (CDGPredicate)NULL;
}

/* ----------------------------------------------------------------------
 * return the name of the function implementing a functor
 */
String predGetFunctionName(String functor)
{
  int i;

  for (i = 0; eval_predicates[i].functor != NULL; i++) {
    if (strcmp(functor, eval_predicates[i].functor) == 0) {
      return eval_predicates[i].functionName;
      break;
    }
  }

  return NULL;
}

/* ----------------------------------------------------------------------
 * is #1 a start node?
 */
Boolean predStart(List a, LexemGraph lg, Vector context)
{
   ValueStruct vas;
   Value val;


   val = evalTerm((Term) listElement(a), &vas, lg, context);
   if (val->type != VTLexemNode) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: argument of predicate `start' not an id\n");
      return (FALSE);
   }

   if(val->data.lexemnode == NULL || val->data.lexemnode == NONSPEC) {
     return FALSE;
   }

   /* detect word that is truly the start */
   if(((LexemNode) val->data.lexemnode)->arc->from == lg->min) {
     return TRUE;
   }

   /* detect word that is the start of an embedded clause */
   if(inputSentenceBoundary(
         inputLatticePrevArc( lg->lattice,
                              ((LexemNode)val->data.lexemnode)->arc))) {
     return TRUE;
   }
   
   return FALSE;
}

/* ----------------------------------------------------------------------
 * is #1 an end node? 
 */
Boolean predStop(List a, LexemGraph lg, Vector context)
{
   ValueStruct vas;
   Value val;


   val = evalTerm((Term) listElement(a), &vas, lg, context);
   if (val->type != VTLexemNode) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: argument of predicate `stop' not an id\n");
      return (FALSE);
   }

   return (val->data.lexemnode != NULL && 
           val->data.lexemnode != NONSPEC &&
	   ((LexemNode)val->data.lexemnode)->arc->to == lg->max);
}

/* ----------------------------------------------------------------------
 * is #1 a specified node? 
 */
Boolean predSpec(List a, LexemGraph lg, Vector context)
{
   Value val;
   ValueStruct vas;


   val = evalTerm((Term) listElement(a), &vas, lg, context);

   if (val->type != VTLexemNode) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: argument of predicate `spec' not an id\n");
      return (FALSE);
   }

   return (val->data.lexemnode != NONSPEC);
}

/* ----------------------------------------------------------------------
 * is #1 an underspecified node? 
 */
Boolean predNonSpec(List a, LexemGraph lg, Vector context)
{
  Value val;
  ValueStruct vas;


   val = evalTerm((Term) listElement(a), &vas, lg, context);
   if (val->type != VTLexemNode) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: argument of predicate `nonspec' not an id\n");
      return (FALSE);
   }

   return (val->data.lexemnode == NONSPEC);
}

/* ----------------------------------------------------------------------
 * is #1 the root node? 
 */
Boolean predRoot(List a, LexemGraph lg, Vector context)
{
  Value val;
  ValueStruct vas;


   val = evalTerm((Term) listElement(a), &vas, lg, context);
   if (val->type != VTLexemNode) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: argument of predicate `root' not an id\n");
      return (FALSE);
   }

   /* tv->lexemnode is NULL if and only if X^id points to the root */
   return (val->data.lexemnode == NULL);
}

/* ----------------------------------------------------------------------
 * is #1 a chunk head?
 */
Boolean predChunkHead(List a, LexemGraph lg, Vector context)
{
  ValueStruct vas;
  Value val;
  Chunk chunk;
  LexemNode ln;


  val = evalTerm((Term) listElement(a), &vas, lg, context);
  if (val->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: argument of predicate `chunkhead' not an id\n");
    return FALSE;
  }

  if (val->data.lexemnode == NULL || val->data.lexemnode == NONSPEC) {
    return FALSE;
  }

  ln = (LexemNode) val->data.lexemnode;
  chunk = ln->grapheme->chunk;
  if (!chunk) {
    return FALSE;
  } else {
    return (chunk->head == ((LexemNode) val->data.lexemnode)->grapheme);
  }
}

/* ----------------------------------------------------------------------
 * print #1 
 */
Boolean predPrint(List a, LexemGraph lg, Vector context)
{
  Value val;
  ValueStruct vas;
  LexemNode ln;
  
   while (a != NULL && listElement(a) != NULL) {
     val = evalTerm((Term) listElement(a), &vas, lg, context);
      switch(val->type) {
	 case VTString:
	    cdgPrintf(CDG_DEFAULT, "%s", val->data.string);
	    break;
	 case VTNumber:
	    cdgPrintf(CDG_DEFAULT, "%f", val->data.number);
	    break;
	 case VTLexemNode:
	    ln = (LexemNode)val->data.lexemnode;
            if(!ln) {
              cdgPrintf(CDG_DEFAULT, "NIL");
            } else if (ln == NONSPEC) {
		  cdgPrintf(CDG_DEFAULT, "NONSPEC");
            } else {
              cdgPrintf(CDG_DEFAULT, "%s@%p",
                        ((LexemNode)val->data.lexemnode)->arc->word,
                        val->data.lexemnode);
	    }
	    break;
	 default:
	    cdgPrintf(CDG_WARNING, 
	       "WARNING: type mismatch for predicate `print'\n");
	    return (FALSE);
      }
      a = listNext(a);
   }

   return (TRUE);
}

/* ----------------------------------------------------------------------
 * Does #2 subsume #3 in hierarchy #1?
 */
Boolean predSubsumes(List l, LexemGraph lg, Vector context)
{
   ValueStruct vsa, vsb, vsh;
   Value va, vb, vh;

   Hierarchy h;
   int a, b;
   String constraint =
     evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*";


   vh = evalTerm((Term) listNthElement(l, 1), &vsh, lg, context);
   if (vh->type != VTString) {
      cdgPrintf(CDG_WARNING,
                "WARNING: invalid 1st arg in `");
      printFormula(CDG_WARNING, evalCurrentFormula, FTConstant, 0);
      cdgPrintf(CDG_WARNING, "' in constraint `%s'\n",
                constraint);
      return FALSE;
   }
   va = evalTerm((Term) listNthElement(l, 2), &vsa, lg, context);
   if (va->type != VTString) {
      cdgPrintf(CDG_WARNING,
                "WARNING: invalid 2nd arg in `");
      printFormula(CDG_WARNING, evalCurrentFormula, FTConstant, 0);
      cdgPrintf(CDG_WARNING, "' in constraint `%s'\n",
                constraint);
      return FALSE;
   }
   vb = evalTerm((Term) listNthElement(l, 3), &vsb, lg, context);
   if (vb->type != VTString) {
     cdgPrintf(CDG_WARNING,
                "WARNING: invalid 3rd arg in `");
      printFormula(CDG_WARNING, evalCurrentFormula, FTConstant, 0);
      cdgPrintf(CDG_WARNING, " 'in constraint `%s'\n",
                constraint);
      return FALSE;
   }

   /* get hierarchy */
   if(NULL == (h = findHierarchy(vh->data.string))) {
     return FALSE;
   }

   /* find both sorts */
   if ( 0 > ( a = (int)hashGet( h->types, va->data.string ) - 1 ) ) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: type `%s' not found in hierarchy `%s' in `%s'\n",
	 va->data.string, vh->data.string,
         constraint);

      return (FALSE);
   }
   if ( 0 > ( b = (int)hashGet( h->types, vb->data.string ) - 1 ) ) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: type `%s' not found in hierarchy `%s' in `%s'\n",
	 vb->data.string, vh->data.string,
         constraint);

      return (FALSE);
   }

   /*
   if ( a == b || (int)arrayElement( h->distance, a, b ) > 0 ) {
     cdgPrintf(CDG_WARNING,
	       "WARNING: predicate `subsumes' true for %s and %s\n",
	       tva->string, tvb->string);
   }
   */

   /* return canned answer */
   return ( (int)arrayElement( h->distance, a, b ) >= 0 ); 
}

/* ----------------------------------------------------------------------
 * Does #2 subsume #3, or vice versa, in #1?
 */
Boolean predCompatible(List l, LexemGraph lg, Vector context)
{
   ValueStruct vhs, vas, vbs;
   Value vh, va, vb;
   Hierarchy h;
   int a, b;
   String constraint =
     evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*";


   vh = evalTerm((Term) listNthElement(l, 1), &vhs, lg, context);
   if (vh->type != VTString) {
     cdgPrintf(CDG_WARNING,
                "WARNING: invalid 1st arg in `");
      printFormula(CDG_WARNING, evalCurrentFormula, FTConstant, 0);
      cdgPrintf(CDG_WARNING, "' in constraint `%s'\n",
                constraint);
      return FALSE;
   }
   va = evalTerm((Term) listNthElement(l, 2), &vas, lg, context);
   if (va->type != VTString) {
     cdgPrintf(CDG_WARNING,
                "WARNING: invalid 2nd arg in `");
      printFormula(CDG_WARNING, evalCurrentFormula, FTConstant, 0);
      cdgPrintf(CDG_WARNING, "' in constraint `%s'\n",
                constraint);
      return FALSE;
   }
   vb = evalTerm((Term) listNthElement(l, 3), &vbs, lg, context);
   if (vb->type != VTString) {
     cdgPrintf(CDG_WARNING,
                "WARNING: invalid 3rd arg in `");
      printFormula(CDG_WARNING, evalCurrentFormula, FTConstant, 0);
      cdgPrintf(CDG_WARNING, "' in constraint `%s'\n",
                constraint);
      return FALSE;
   }

   h = findHierarchy(vh->data.string);
   if(!h) {
     cdgPrintf(CDG_WARNING,
               "WARNING: hierarchy `%s' not found\n", vh->data.string);
     return (FALSE);
   }

   /* find both sorts */
   if ( 0 > ( a = (int)hashGet( h->types, va->data.string ) - 1 ) ) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: type `%s' not found in hierarchy `%s' in `%s'\n",
                va->data.string, vh->data.string,
    	 evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return (FALSE);
   }
   if ( 0 > ( b = (int)hashGet( h->types, vb->data.string ) - 1) ) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: type `%s' not found in hierarchy `%s' in `%s'\n",
                vb->data.string, vh->data.string,
	 evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return (FALSE);
   }

   return ( (int)arrayElement( h->distance, a, b ) >= 0 ||
	    (int)arrayElement( h->distance, b, a ) >= 0  );
}

/* ----------------------------------------------------------------------
 * Is #1 defined? 
 */
Boolean predExists(List l, LexemGraph lg, Vector context)
{
   Term t;
   Value val;
   ValueStruct vas;

   
   t = (Term) listElement(l);
   if (t->type != TTTopPeek && t->type != TTBottomPeek) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: predicate `exists' needs a lexical access\n");
      return (FALSE);
   }

   val = evalTerm(t, &vas, lg, context);

   return (val->type != VTError);
}

/* ----------------------------------------------------------------------
   Is the edge LV a direct or indirect descendant of LN?

   Used by predHas() in recursive mode.
*/
Boolean descendant(LevelValue lv, LexemNode ln,
                   Vector context, int counter,
                   Hierarchy h, String s) {

  Boolean result = FALSE;
  int nrlevels = inputCurrentGrammar->noOfLevels;
  LevelValue parent;
  /* break out of loops */
  if(counter > lv->modifier->lexemgraph->max) {
    return FALSE;
  }

  /* answer FALSE for NIL LVs */
  if(!spec(lv->modifiee)) {
    return FALSE;
  }

  /* reject wrong labels */
  if(counter && s && !inputSubsumes(h, s, lv->label)) {
    return FALSE;
  }

  /* detect trivial success */
  if(lv->modifiee->arc->from == ln->arc->from) {
    return TRUE;
  }

  /* maybe call self */
  parent =
    vectorElement(context,
                    nrlevels * lv->modifiee->arc->from + lv->level->no);
  if(parent) {
    result = descendant(parent, ln, context, counter+1, h, s);
  }
  
  return result;
  
}

/* ----------------------------------------------------------------------
   Does the LV #1 have a daughter edge with label #2?

   `has' checks whether one of the edges below the specified lexeme node
   has particular properties.

   If there are two arguments, the first is the lexeme node and the second
   is a label that must be present.

   If the second argument is the name of a unary constraint, its
   formula will be used for matching instead of the edges' labels.
   
   If there is a third argument, the third is a hierarchy which should be
   used for label matching. Thus, a constraint can say `has(X@id,
   OBJA_OBJC, Labels)', and (assuming a fitting `Labels' hierarchy) either
   OBJA or OBJC will satisfy the predicate.

   If there is a fourth argument, the search is conducted recursively. The
   value of the argument is a node in the previously specified hierarchy. 
   This node subsumes all the labels which are *allowed* on the path
   through the tree. So to express that anything which has a relative
   pronoun in its scope is a relative clause, you can say

     X@cat = vfin & has(X@id, find_prel, Labels, AUX)
     ->
     X.label = REL
 
   This constraint forces a verb to be a relative clause if one word in its
   VP has a relative pronoun as a subordinate, but not if a verb in a
   subordinated clause has a relative pronoun, because then at least one
   label other than `AUX' will intervene.
     
   If there is a fifth argument, it must be the string `left' or `right'.
   It the restricts the search to the portions to the left or to the right
   of the anchor word in the sentence.

*/
Boolean predHas(List l, LexemGraph lg, Vector context) {
  LevelValue lv;
  String label = NULL;
  String hname = NULL;
  String labels = NULL;
  Hierarchy h = NULL;
  int nrlevels = inputCurrentGrammar->noOfLevels;
  int i;
  LexemNode ln;
  ValueStruct vas;
  Value val;
  int cache_index = -1;
  Level level = inputGetMainlevel(inputCurrentGrammar);
  int mainno = level ? level->no : 0;
  List args;
  Constraint c = NULL;
  Boolean recursive = FALSE;
  Boolean result = FALSE;
  int ordering = 0;
  
  /* if there is no context at all, we can only assume that
     the edge might be found if we had it. */
  if(!context) {
    return TRUE;
  }

  args = l;

  val = evalTerm((Term) listElement(args), &vas, lg, context);
  if (val->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: argument of predicate `has' not an id in `%s'\n",
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return FALSE;
  }
  ln = (LexemNode) val->data.lexemnode;
  if (!spec(ln)) {
    return FALSE;
  }

  /* read cache */
  if(lock_counter) {
    int result;
    cache_index = evalCurrentFormula->index * lock_width + ln->arc->from;
    result = has_cache[cache_index];
#ifdef DEBUGHAS
    cdgPrintf(CDG_DEBUG, "has_cache[%d] == %d\n", cache_index, result);
#endif
    if(2 == result) {
      return TRUE;
    }
    if(1 == result) {
      return FALSE;
    }
  }
  
  /* read second parameter */
  args = listNext(args);
  val = evalTerm((Term) listElement(args), &vas, lg, context);
  if(val->type == VTString) {
    label = val->data.string;
  }
  /* complain */
  else {
    cdgPrintf(CDG_WARNING,
              "WARNING: type mismatch (2rd arg) for `has' in `%s'\n",
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTString);
    return FALSE;
  }

#ifdef DEBUGHAS
  cdgPrintf(CDG_DEBUG, "DEBUG: has(`%s', `%s'", ln->lexem->word, label);
#endif
  
  /* if the string is the name of a unary constraint,
     use this constraint instead of label match  */
  c = hashGet(inputCurrentGrammar->constraints, label);
  if(c && 1 != listSize(c->vars)) {
    c = NULL;
  }
  
  /* read third parameter and find Hierarchy to use */
  if(NULL != (args = listNext(args))) {

    val = evalTerm((Term) listElement(args), &vas, lg, context);
    if(val->type == VTString) {
      hname = val->data.string;
    } else {
    cdgPrintf(CDG_WARNING,
              "WARNING: type mismatch (3rd arg) for `has' in `%s'\n",
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTString);
      return FALSE;
    }
    if(!(h = findHierarchy(hname))) {
      cdgPrintf(CDG_WARNING,
                "WARNING: hierarchy `%s' not found.\n", hname);
      return FALSE;
    }
#ifdef DEBUGHAS
  cdgPrintf(CDG_DEBUG, ", `%s'", hname);
#endif
  }
  

  /* read fourth parameter */
  if(args && (args = listNext(args))) {
    recursive = TRUE;
    val = evalTerm((Term) listElement(args), &vas, lg, context);
    if(val->type == VTString) {
      labels = val->data.string;
    } else {
      cdgPrintf(CDG_WARNING,
                "WARNING: type mismatch (4th arg) for `has' in `%s'\n",
                evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      printTypeError(CDG_WARNING, val, VTString);
      return FALSE;
    }

#ifdef DEBUGHAS
  cdgPrintf(CDG_DEBUG, ", `%s'", labels);
#endif

  }

  /* read fifth parameter */
  if(args && (args = listNext(args))) {
    val = evalTerm((Term) listElement(args), &vas, lg, context);
    if(val->type == VTString &&
       !strcmp("left", val->data.string)) {
      ordering = -1;
    } else if(val->type == VTString &&
              !strcmp("right", val->data.string)) {
      ordering = 1;
    } else {
      cdgPrintf(CDG_WARNING,
                "WARNING: 5th arg for `has' must be `left' or `right' in `%s'\n",
                evalCurrentConstraint ?
                evalCurrentConstraint->id : "*UNKNOWN*");
      return FALSE;
    }
    
#ifdef DEBUGHAS
    cdgPrintf(CDG_DEBUG, ", `%s'", val->data.string);
#endif
  }
 
  
#ifdef DEBUGHAS
  cdgPrintf(CDG_DEBUG, ")? ");
#endif

  /* look for a suitable child edge */
  result = FALSE;
  for(i = 0; i < lg->max; i++) {
    int index = nrlevels * i + mainno;
    if(index >= vectorSize(context)) {
      break;
    }
    lv = vectorElement(context, index);

#ifdef DEBUGHAS
    if(lv) cdgPrintf(CDG_DEBUG, "@%d ", lvIndex(lv));
#endif

    if(!lv || !spec(lv->modifiee)) continue;
    if(1 == ordering && lv->modifier->arc->from <= ln->arc->from)
      continue;
    if(-1 == ordering && lv->modifier->arc->from >= ln->arc->from)
      continue;
    if((!recursive && lv->modifiee->arc->from == ln->arc->from) ||
       (recursive && descendant(lv, ln, context, 0, h, labels))) {
      
      if(c) {

        /* call constraint */
        Constraint thisConstraint = evalCurrentConstraint;
        if(evalUnaryConstraint(c, NULL, context, lv)) {
          evalCurrentConstraint = thisConstraint;
          result = TRUE;
          break;
        }
        evalCurrentConstraint = thisConstraint;
          
      } else {
      
        /* compare label */
        if((!h && lv->label == label)
           ||
           (h && inputSubsumes(h, label, lv->label))) {
          result = TRUE;
          break;
        }
      }
    }
  }

#ifdef DEBUGHAS
  cdgPrintf(CDG_DEBUG, " ...%s.\n", result ? "yes" : "no");
#endif

  /* write cache */
  if(lock_counter) {

#ifdef DEBUGHAS
    cdgPrintf(CDG_DEBUG, "has_cache[%d] = %d\n",
              cache_index, 1 + (result == TRUE));
#endif

    has_cache[cache_index] = 1 + (result == TRUE);
  }

  return result;
  
}


/* ----------------------------------------------------------------------
   Is the word #1 subordinated with label #2?

   `is' checks whether the specified word is directly subordinated with the
   specified label.

   The first parameter is a lexeme node in the graph, and
   the second parameter is the label that it should carry.

   If there is a third parameter, it is the name of the level on which the
   subordination is checked. If there is no third parameter, the level that
   is first mentioned in the enclosing constraint is used.

*/
Boolean predIs(List l, LexemGraph lg, Vector context) {

  LevelValue lv;
  String label = NULL;
  Constraint c = evalCurrentConstraint;
  int nrlevels = inputCurrentGrammar->noOfLevels;
  int i;
  int levelno = -1;
  LexemNode ln;
  ValueStruct vas;
  Value val;
  
  /* if there is no context at all, we can only assume that
     the edge might be found if we had it. */
  if(!context) {
    return TRUE;
  }


  val = evalTerm((Term) listElement(l), &vas, lg, context);
  if (val->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: argument of predicate `is' not an id\n");
    return FALSE;
  }
  ln = (LexemNode) val->data.lexemnode;
  if (!ln || ln == NONSPEC) {
    return FALSE;
  }
  
  /* look for a particular label */
  l = listNext(l);
  val = evalTerm((Term) listElement(l), &vas, lg, context);
  if(val->type == VTString) {
    label = val->data.string;
  }
  /* complain */
  else {
    cdgPrintf(CDG_WARNING,
              "WARNING: type mismatch (2rd arg) for `is' in `%s'\n",
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTString);
    return FALSE;
  }
    
  /* get optional third parameter */
  l = listNext(l);
  if(l) {
    val = evalTerm((Term) listElement(l), &vas, lg, context);
    if(val->type == VTString) {
      Level level = findLevel(val->data.string);
      if(!level) {
        cdgPrintf(CDG_WARNING,
                  "WARNING: no level `%s' defined in `%s'\n",
                  val->data.string,
                  evalCurrentConstraint ?
                  evalCurrentConstraint->id :
                  "*UNKNOWN*");
        return FALSE;
      }
      levelno = level->no;
    }
    /* complain */
    else {
      cdgPrintf(CDG_WARNING,
                "WARNING: type mismatch (3rd arg) for `is' in `%s'\n",
                evalCurrentConstraint ?
                evalCurrentConstraint->id :
                "*UNKNOWN*");
      printTypeError(CDG_WARNING, val, VTString);
      return FALSE;
    }

  } else {
    VarInfo vi = listElement(c->vars);
    levelno = vi->level->no;
  }
  
  /* check corresponding LV for the correct label */
  i = ln->arc->from;

  
  lv = vectorElement( context, nrlevels * i + levelno);
  if(lv && lv->label == label) {
    return TRUE;
  }
  return FALSE;
  
}

/* ----------------------------------------------------------------------
 *  Does word #1 occur somewhere in the sentence?
 */
Boolean predOccur(List l, LexemGraph lg, Vector context) {
  ValueStruct vas;
  Value val;
  

  val = evalTerm((Term) listElement(l), &vas, lg, context);
  if (val->type != VTString) {
    cdgPrintf(CDG_WARNING,
              "WARNING: type mismatch (1st arg) for `occur' in `%s'\n",
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTString);
    return FALSE;
  }

  return lgContains(lg, val->data.string);
}

/* ----------------------------------------------------------------------
   Does a punctuation mark contained in S occur between two words?

   The first and second parameters are lexeme nodes, and the third parameter
   is a string. TRUE is returned if one of the words between the two lexeme
   nodes is one character long and consists of one of the characters in the
   string.

   The two lexeme nodes may be given in either order indifferently.

   Either of the lexeme nodes may be replaced by a Number, which is then
   used as an index into the lexeme graph instead.
 */
Boolean predBetween(List l, LexemGraph lg, Vector context) {
  ValueStruct vas;
  Value val;
  int a,b;
  LexemNode ln;
  String s;
  
  val = evalTerm((Term) listElement(l), &vas, lg, context);
  if (val->type == VTLexemNode) {
    ln = (LexemNode) val->data.lexemnode;
    if (!ln || ln == NONSPEC) {
      return FALSE;
    }
    a = ln->arc->from;
  } else if (val->type == VTNumber) {
    a = val->data.number;
    if(a < lg->min) {
      a = lg->min;
    }
    if(a > lg->max) {
      a = lg->max;
    }
  } else {
    cdgPrintf(CDG_WARNING,
              "WARNING: type mismatch (1st arg) for `between' in `%s'\n",
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTLexemNode);
    return FALSE;
  }
  
  l = listNext(l);
  val = evalTerm((Term) listElement(l), &vas, lg, context);
  if (val->type == VTLexemNode) {
    ln = (LexemNode) val->data.lexemnode;
    if (!ln || ln == NONSPEC) {
      return FALSE;
    }
    b = ln->arc->from;
  } else if (val->type == VTNumber) {
    b = val->data.number;
    if(b < lg->min) {
      b = lg->min;
    }
    if(b > lg->max) {
      b = lg->max;
    }
  } else {
    cdgPrintf(CDG_WARNING,
              "WARNING: type mismatch (2nd arg) for `between' in `%s'\n",
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTLexemNode);
    return FALSE;
  }

  if(a>b) {
    int c = a;
    a = b, b = c;
  }

  l = listNext(l);
  val = evalTerm((Term) listElement(l), &vas, lg, context);
  if (val->type != VTString) {
    cdgPrintf(CDG_WARNING,
              "WARNING: type mismatch (3rd arg) for `between' in `%s'\n",
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTString);
    return FALSE;
  }
  s = val->data.string;

  
#ifdef DEBUGBETWEEN
  cdgPrintf(CDG_DEBUG, "between(%d, %d, %s)?", a,b,s);
#endif
  for(l = lg->lattice->arcs; l != NULL; l = listNext(l)) {
    Arc arc = listElement(l);
#ifdef DEBUGBETWEEN
  cdgPrintf(CDG_DEBUG, " %s(%d)", arc->word, arc->from);
#endif
    if(arc->from <= a || arc->from > b) continue;
    if(strlen(arc->word) != 1 ||
       !strspn(arc->word, s)) continue;
#ifdef DEBUGBETWEEN
  cdgPrintf(CDG_DEBUG, " yes.\n");
#endif
    return TRUE;
  }
#ifdef DEBUGBETWEEN
  cdgPrintf(CDG_DEBUG, " no.\n");
#endif
  return FALSE;
}

/* ----------------------------------------------------------------------
 *  Is A indirectly a modifier of B?
 */
Boolean predUnder(List l, LexemGraph lg, Vector context) {
  ValueStruct vas;
  Value val;
  int a,b;
  LexemNode ln;
  LevelValue lv;
  VarInfo vi = listElement(evalCurrentConstraint->vars);
  int nrlevels = inputCurrentGrammar->noOfLevels;
  int j = 0;
  
  val = evalTerm((Term) listElement(l), &vas, lg, context);
  if (val->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
              "WARNING: type mismatch (1st arg) for `under' in `%s'\n",
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTLexemNode);
    return FALSE;
  }
  ln = (LexemNode) val->data.lexemnode;
  if (!ln || ln == NONSPEC) {
    return FALSE;
  }
  a = ln->arc->from;
  val = evalTerm((Term) listElement(listNext(l)), &vas, lg, context);
  if (val->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
              "WARNING: type mismatch (2nd arg) for `under' in `%s'\n",
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTLexemNode);
    return FALSE;
  }
  ln = (LexemNode) val->data.lexemnode;
  if (!ln || ln == NONSPEC) {
    return FALSE;
  }
  b = ln->arc->from;
#ifdef DEBUGUNDER
  cdgPrintf(CDG_DEBUG, "under(%d, %d)?", a,b);
#endif

  do {
    lv = vectorElement(context, nrlevels * a + vi->level->no);
    if(!lv || !spec(lv->modifiee)) {
#ifdef DEBUGUNDER
      cdgPrintf(CDG_DEBUG, " no.\n");
#endif
      /* failure */
      return FALSE;
    }
    if(lv->modifiee->arc->from == b) {
#ifdef DEBUGUNDER
      cdgPrintf(CDG_DEBUG, " yes.\n");
#endif
      /* success */
      return TRUE;
    }
    
    if(j > lg->max) {
#ifdef DEBUGUNDER
      cdgPrintf(CDG_DEBUG, " no.\n");
#endif
      /* cycle */
      return FALSE;
    }
    a = lv->modifiee->arc->from;
    j++;
#ifdef DEBUGUNDER
      cdgPrintf(CDG_DEBUG, " %d", a);
#endif
  } while(lv);

  return FALSE;
}

/* ----------------------------------------------------------------------
   Are A and B part of the same tree?
 */
Boolean predConnected(List l, LexemGraph lg, Vector context) {
  ValueStruct vas;
  Value val;
  int a, b, roota = -1, rootb = -1;
  LexemNode lna, lnb;
  LevelValue lv;
  VarInfo vi = listElement(evalCurrentConstraint->vars);
  int nrlevels = inputCurrentGrammar->noOfLevels;
  int max = lg->max - lg->min;
  int count;
  
  val = evalTerm((Term) listElement(l), &vas, lg, context);
  if (val->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
              "WARNING: type mismatch (1st arg) for `connected' in `%s'\n",
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTLexemNode);
    return FALSE;
  }
  lna = (LexemNode) val->data.lexemnode;
  if (!lna || lna == NONSPEC) {
    return FALSE;
  }
  a = lna->arc->from;
  l = listNext(l);

  val = evalTerm((Term) listElement(l), &vas, lg, context);
  if (val->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
              "WARNING: type mismatch (1st arg) for `connected' in `%s'\n",
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTLexemNode);
    return FALSE;
  }
  lnb = (LexemNode) val->data.lexemnode;
  if (!lnb || lnb == NONSPEC) {
    return FALSE;
  }
  b = lnb->arc->from;

  count = 0;
  lv = vectorElement(context, nrlevels * a + vi->level->no);
  while(lv) {
    if(!spec(lv->modifiee)) {
      roota = lv->modifier->arc->from;
      break;
    }
    else if(NONSPEC == lv->modifiee) {
      return FALSE;
    }
    else if(count++ > max) {
      roota = a;
      break;
    }
    else {
      lv = vectorElement(context, nrlevels * lv->modifiee->arc->from + vi->level->no);
    }
  }

  count = 0;
  lv = vectorElement(context, nrlevels * b + vi->level->no);
  while(lv) {
    if(!spec(lv->modifiee)) {
      rootb = lv->modifier->arc->from;
      break;
    }
    else if(NONSPEC == lv->modifiee) {
      return FALSE;
    }
    else if(count++ > max) {
      rootb = b;
      break;
    }
    else {
      lv = vectorElement(context, nrlevels * lv->modifiee->arc->from + vi->level->no);
    }
  }

  return roota == rootb;
}

/* ----------------------------------------------------------------------
 * Does the level contain a cycle?
 */
Boolean predCycle(List l, LexemGraph lg, Vector context) {
  ByteVector cleared;
  ByteVector seen;
  Level level;
  LevelValue lv;
  Value val;
  ValueStruct vas;
  int i;
    
  /* if there is no context at all, we cannot check this condition,
     so we must assume that it is satisfied. */
  if(!context) {
    return FALSE;
  }


  val = evalTerm((Term) listElement(l), &vas, lg, context);
  if (val->type != VTString) {
        cdgPrintf(CDG_WARNING,
                  "WARNING: type mismatch (1st arg) for `cycle' in `%s'\n",
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTString);

    return FALSE;
  }
  lv = NULL;

  /* find level */
  level = findLevel(val->data.string);
  if(!level) {
    cdgPrintf(CDG_ERROR, "ERROR: no level `%s'\n", val->data.string);
    return TRUE;
  }
  
  /* check all edges */
  seen = bvNew(lg->max);
  bvSetElements(seen, FALSE, 0, lg->max);
  cleared = bvNew(lg->max);
  bvSetElements(cleared, FALSE, 0, lg->max);

  for(i = 0; i < vectorSize(context); i++) {
    LevelValue lv = vectorElement(context, i);
    if(!lv) {
      continue;
    }

    /* skip LVs of other levels */
    if(i % inputCurrentGrammar->noOfLevels != level->no) {
      continue;
    }
    
#ifdef DEBUGPREDCYCLE
    cdgPrintf(CDG_ERROR, "DEBUG: checking time point %d\n", lv->modifier->arc->from);
#endif
    /* search connection to NIL */
    while(lv) {
      int index;
#ifdef DEBUGPREDCYCLE
      cdgPrintf(CDG_ERROR, "DEBUG: loop reaches time point %d.\n", lv->modifier->arc->from);
#endif
      /* detect loop */
      if(bvElement(seen, lv->modifier->arc->from)) {
        bvDelete(seen);
        bvDelete(cleared);
#ifdef DEBUGPREDCYCLE
        cdgPrintf(CDG_ERROR, "DEBUG: loop detected\n", lv->modifier->arc->from);
#endif
        return TRUE;
      }
      bvSetElement(seen, TRUE, lv->modifier->arc->from);


      /* find already cleared LV */
      if(!spec(lv->modifiee) || bvElement(cleared, lv->modifier->arc->from)) {

        /* set seen LVs to cleared also */
        int j;
        for(j = 0; j < bvSize(seen); j++) {
          if(bvElement(seen, j)) {
            bvSetElement(cleared, TRUE, j);
          }
        }
#ifdef DEBUGPREDCYCLE
        cdgPrintf(CDG_ERROR, "DEBUG: time point %d is ok.\n", lv->modifier->arc->from);
#endif
        bvSetElements(seen, FALSE, 0, lg->max);
        break;
      }
      
      /* step upward */
      index =
        inputCurrentGrammar->noOfLevels * lv->modifiee->arc->from +
        lv->level->no;
      lv = vectorElement(context, index);
    }
  }

  bvDelete(seen);
  bvDelete(cleared);
  return FALSE;
}

/* ----------------------------------------------------------------------
 * In hierarchy #1, is there a common ancestor of sorts #2 and #3 
 * that is not more than #4 levels away?
 */
/*
Boolean predNthCousin(List l, LexemGraph lg, Vector context) {
   TermValueUnion tvuh, tvua, tvub, tvun;
   TermValue tvh = &tvuh;
   TermValue tva = &tvua;
   TermValue tvb = &tvub;
   TermValue tvn = &tvun;
   EvalTermType tth,tta,ttb,ttn;
   Hierarchy h;
   Sort s;
   int i, a, b, n;
   int levela=0, levelb=0, leveli;

   if (listSize(l) != 4) {
     cdgPrintf(CDG_WARNING, 
	       "WARNING: predicate `nth_cousin' needs four arguments\n");
     return (FALSE);
   }
   tth = evalTerm((Term) listNthElement(l, 1), tvh, lg, Vector context);
   if (tth != ETTString) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: type mismatch (1st arg) for `nth_cousin' in `%s'\n",
	 evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return (FALSE);
   }
   tta = evalTerm((Term) listNthElement(l, 2), tva, lg, Vector context);
   if (tta != ETTString) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: type mismatch (2nd arg) for `nth_cousin' in `%s'\n",
	 evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return (FALSE);
   }
   ttb = evalTerm((Term) listNthElement(l, 3), tvb, lg, Vector context);
   if (ttb != ETTString) {
      if (!evalSloppySubsumesWarnings)
	 cdgPrintf(CDG_WARNING,
	    "WARNING: type mismatch (3rd arg) for `nth_cousin' in `%s'\n",
	    evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return (FALSE);
   }
   ttn = evalTerm((Term) listNthElement(l, 4), tvn, lg, Vector context);
   if (ttn != ETTNumber) {
      if (!evalSloppySubsumesWarnings)
	 cdgPrintf(CDG_WARNING,
	    "WARNING: type mismatch (4rd arg) for `nth_cousin' in `%s'\n",
	    evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return (FALSE);
   }
   n = tvun.number;

   for (l = inputCurrentGrammar->hierarchies; l != NULL; l = listNext(l))
      if (strcmp(tvh->string, ((Hierarchy) listElement(l))->id) == 0)
	 break;
   if (l == NULL) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: hierarchy `%s' not found\n", tvh->string);
      return (FALSE);
   }
   h = (Hierarchy) listElement(l);

   a = b = -1;
   for (i = 0; i < vectorSize(h->sorts); i++) {
      s = (Sort) vectorElement(h->sorts, i);
      if (strcmp(s->id, tva->string) == 0) {
	a = i;
	levela = s->depth;
      }
      if (strcmp(s->id, tvb->string) == 0) {
	 b = i;
	 levelb = s->depth;
      }
   }
   if (a < 0) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: sort `%s' not found in hierarchy `%s'\n",
	 tva->string, tvh->string);
      return (FALSE);
   }
   if (b < 0) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: sort `%s' not found in hierarchy `%s'\n",
	 tvb->string, tvh->string);
      return (FALSE);
   }

   if(abs(levela-levelb) > n) 
   return(FALSE);
 
 
   for (i = 0; i < vectorSize(h->sorts); i++) {
     s = (Sort) vectorElement(h->sorts, i);
     leveli = s->depth;

     if ( levela-leveli <= n &&
	  levelb-leveli <= n &&
	 (Boolean) arrayElement(h->subsumption, i, a) &&
	 (Boolean) arrayElement(h->subsumption, i, b) 
	  ) 
       return (TRUE);
     
   }
   return(FALSE);
}
*/
/* ----------------------------------------------------------------------
 * In hierarchy #1, is there a common ancestor of sorts #2 and #3 
 * at level #4 or lower?
 */
/*
Boolean predAncestor(List l, LexemGraph lg, Vector context) {
   TermValueUnion tvuh, tvua, tvub, tvun;
   TermValue tvh = &tvuh;
   TermValue tva = &tvua;
   TermValue tvb = &tvub;
   TermValue tvn = &tvun;
   EvalTermType tth,tta,ttb,ttn;
   Hierarchy h;
   Sort s;
   int i, a, b, n;
   int levela=0, levelb=0, leveli;

   if (listSize(l) != 4) {
     cdgPrintf(CDG_WARNING, 
       "WARNING: predicate `ancestor_at_level' needs four arguments\n");
     return (FALSE);
   }
   tth = evalTerm((Term) listNthElement(l, 1), tvh, lg, Vector context);
   if (tth != ETTString) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: type mismatch (1st arg) for `ancestor_at_level' in `%s'\n",
	 evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return (FALSE);
   }
   tta = evalTerm((Term) listNthElement(l, 2), tva, lg, Vector context);
   if (tta != ETTString) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: type mismatch (2nd arg) for `ancestor_at_level' in `%s'\n",
	 evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return (FALSE);
   }
   ttb = evalTerm((Term) listNthElement(l, 3), tvb, lg, Vector context);
   if (ttb != ETTString) {
      if (!evalSloppySubsumesWarnings)
	 cdgPrintf(CDG_WARNING,
	    "WARNING: type mismatch (3rd arg) for `ancestor_at_level' in `%s'\n",
	    evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return (FALSE);
   }
   ttn = evalTerm((Term) listNthElement(l, 4), tvn, lg, Vector context);
   if (ttn != ETTNumber) {
      if (!evalSloppySubsumesWarnings)
	 cdgPrintf(CDG_WARNING,
	    "WARNING: type mismatch (4rd arg) for `ancestor_at_level' in `%s'\n",
	    evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return (FALSE);
   }
   n = tvun.number;

   for (l = inputCurrentGrammar->hierarchies; l != NULL; l = listNext(l))
      if (strcmp(tvh->string, ((Hierarchy) listElement(l))->id) == 0)
	 break;
   if (l == NULL) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: hierarchy `%s' not found\n", tvh->string);
      return (FALSE);
   }
   h = (Hierarchy) listElement(l);

   a = b = -1;
   for (i = 0; i < vectorSize(h->sorts); i++) {
      s = (Sort) vectorElement(h->sorts, i);
      if (strcmp(s->id, tva->string) == 0) {
	a = i;
	levela = s->depth;
      }
      if (strcmp(s->id, tvb->string) == 0) {
	 b = i;
	 levelb = s->depth;
      }
   }
   if (a < 0) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: sort `%s' not found in hierarchy `%s'\n",
	 tva->string, tvh->string);
      return (FALSE);
   }
   if (b < 0) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: sort `%s' not found in hierarchy `%s'\n",
	 tvb->string, tvh->string);
      return (FALSE);
   }


 if(n > levela) 
   return(FALSE);
 
 if(n > levelb) 
   return(FALSE);
 
   for (i = 0; i < vectorSize(h->sorts); i++) {
     s = (Sort) vectorElement(h->sorts, i);
     leveli = s->depth;

     if ( leveli >= n &&
	 (Boolean) arrayElement(h->subsumption, i, a) &&
	 (Boolean) arrayElement(h->subsumption, i, b)) 
       return (TRUE);
     
   }
   return(FALSE);
}
*/

/* ---------------------------------------------------------------------- */
/* -- ENDOFFILE --------------------------------------------------------- */
