/*
 * File: variables.h
 * Author: Horia F. Pop
 * Purpose: header file for variables encapsulation
 * Birth: February 9, 2000
 * 
 * $Id: variables.h,v 1.15 2003/06/11 16:06:09 micha Exp $
 */

#ifndef VARIABLES_H
#define VARIABLES_H

/* -- INCLUDES ---------------------------------------------------------- */
#include <blah.h>

/* -- MACROS ------------------------------------------------------------ */
#define VAR_NONE    	0
#define VAR_BOOL    	1
#define VAR_INT     	2
#define VAR_FLOAT   	3
#define VAR_STRING  	4
#define VAR_UNSLONG	5
#define VAR_BIT  	6
#define VAR_ENUM  	7
#define VAR_USER    	8

/* -- TYPE DEFINITIONS -------------------------------------------------- */

struct VariableStruct;
typedef struct VariableStruct VariableStruct, *Variable;

/*
 * Opaque type definition for iterator object used by
 * - varIteratorNew
 * - varIteratorNextName
 * - varIteratorFree
 */
struct VarIteratorStruct;
typedef struct VarIteratorStruct VarIteratorStruct, *VarIterator;

/* -- VARIABLES --------------------------------------------------------- */
/* -- FUNCTIONS --------------------------------------------------------- */

/* ----------------------------------------------------------------------
 * Makes the parameter a new valid iterator object for iterating
 * through all registered variables. The general use is as follows:
 *
 * String s;
 * VarIterator vi;
 *
 * varIteratorNew(&vi);
 * while (s=varIteratorNextName(&vi)) { do something with s }
 * varIteratorFree(&vi);
 */
extern void varIteratorNew(VarIterator *);

/* ----------------------------------------------------------------------
 * Returns the name of the next variable. Cf. above.
 */
extern String varIteratorNextName(VarIterator *);

/* ----------------------------------------------------------------------
 * Finally deletes an VarIterator object. Cf. above.
 */
extern void varIteratorFree(VarIterator *);

/* ----------------------------------------------------------------------
 * Returns the variable name
 */
extern String varName(Variable v);

/* ----------------------------------------------------------------------
 * Returns the variable type 
 */
extern int varType(Variable v);

/* ----------------------------------------------------------------------
 * Returns the variable pointer 
 */
extern Pointer varPtr(Variable v);

/* ----------------------------------------------------------------------
 * Returns the variable validation function 
 */
extern BooleanFunction * varValidate(Variable v);

/* ----------------------------------------------------------------------
 * Returns the variable callback function 
 */
extern VoidFunction * varCallback(Variable v);

/* ----------------------------------------------------------------------
 * Returns the variable userset function 
 */
extern VoidFunction * varUserSet(Variable v);


/* ----------------------------------------------------------------------
 * Allocate a Variable and sets its fields
 */
extern Variable varAllocate(String name, int type, Pointer ptr, 
			    BooleanFunction *funcValidate,
			    VoidFunction *funcCallback,
			    VoidFunction *funcUserSet);

/* ----------------------------------------------------------------------
 * Registers a variable. The arguments are the name, type, address, and 
 * address of validation function and callback function
 * o the validation function is used to confirm that a value is valid
 * o the callback function is called after the value has been set
 */
extern Boolean varRegister(String name, int type, Pointer ptr, 
			   BooleanFunction *funcValidate,
			   VoidFunction *funcCallback,
			   VoidFunction *funcUserSet);

/* ----------------------------------------------------------------------
 * The second argument must be of the type associated with the given name
 * Returns TRUE if the named variable does exist and if the value is valid.
 * It is varSet's responsibility to set the value; the validating function
 * only validates it.
 */
extern Boolean varSet(String name, ...);

/* ----------------------------------------------------------------------
 * The second argument must be an address to the type associated with the
 * given name. Returns TRUE if the named variable does exist. On the exit 
 * of this function, the second argument will hold the value of that 
 * variable.
 */
extern Boolean varGet(String name, Pointer ptr);

/* ----------------------------------------------------------------------
 * Returns a list of Strings holding the names of all variables.
 */
extern List varReturnListOfNames();

/* ----------------------------------------------------------------------
 * Returns the type of the given variable, or VAR_NONE otherwise.
 */
extern int varReturnType(String name);

extern void varFinalize();

/* ---------------------------------------------------------------------- */
#endif /* don't insert anything after this #endif */
