/*

    The BLAH library, a container library
    Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: blah@nats.informatik.uni-hamburg.de

    $Id: ringbuffer.c,v 1.4 2004/03/23 14:21:57 micha Exp $

*/

/* ------------------------------------------------------------------- */
/* ----------------------------------------------------------------------
 * @defgroup Ringbuffer Ringbuffers
 * Implementation of ring buffers.
 *
 * It is similar to the vector that stores a series of objects,though the head
 * and the tail are attached inorder to enhance cyclic operations.
 * @{
 */

#include <stdio.h>
#include "blah.h"

/* ----------------------------------------------------------------------
 * internal representation of the ring buffer.
 */
struct RingBufferStruct {
  int first;		     /**< index of first item */
  int last;		     /**< index of the item after the last item */
  int count;		     /**< number of entries */
  int capacity;		     /**< capacity of table */
  Pointer *entries;	     /**< table of entries */
};

/* ----------------------------------------------------------------------
 * creates a new ringbuffer.
 *
 * @param capacity specifies the capacity with which the new ringbuffer has to be created.
 * @returns a new ringBuffer with the size \a capacity.
 */
RingBuffer rbNew(int capacity)
{
  RingBuffer rb;

  if(capacity <= 0) {
    capacity = 1;
  }

  rb = (RingBuffer)memMalloc(sizeof(RingBufferStruct));
  rb->first = 0;
  rb->last = 0;
  rb->count = 0;
  rb->capacity = capacity;
  rb->entries = (Pointer *)memMalloc(sizeof(Pointer)*capacity);

  return (rb);
}

/* ----------------------------------------------------------------------
 * deletes a ringbuffer.
 *
 * @param rb the ringbuffer that is to be deleted.
 */
void rbDelete(RingBuffer rb)
{
  memFree(rb->entries);
  memFree(rb);
}

/* ----------------------------------------------------------------------
 * checks if the given ringbuffer is empty.
 *
 * @param rb ringbuffer that is to be checked.
 * @returns \a TRUE if the buffer is empty and \a FALSE otherwise.
 */
Boolean rbIsEmpty(RingBuffer rb)
{
  return (rb->count == 0);
}

/* ----------------------------------------------------------------------
 * checks if a given ring buffer is full.
 *
 * @param rb  ringbuffer that is to be checked
 * @returns \a TRUE if buffer is full and \a FALSE otherwise.
 */
Boolean rbIsFull(RingBuffer rb)
{
  return (rb->count == rb->capacity);
}

/* ----------------------------------------------------------------------
 * adds a new element to the top of a ring buffer.
 *
 * @param rb ringbuffer to which addition has to be done.
 * @param element element that has to be added to the ring buffer.
 * @returns the number of elements in the buffer
 */
int rbAddTopElement(RingBuffer rb, Pointer element)
{
  if (rbIsFull(rb)) {
    /* a buffer has limited capacity, so no resizing here */
    fprintf(stderr, "WARNING: rbAddTopElement: buffer overflow\n");
    abort();
  }

  rb->entries[rb->last] = element;
  rb->last++;
  if(rb->last == rb->capacity)
    rb->last = 0;

  return (++rb->count);
}

/* ----------------------------------------------------------------------
 * adds a new element to the bottom of a ring buffer.
 *
 * @param rb  ringbuffer to which addition has to be done.
 * @param element element that has to be added to the ring buffer.
 * @returns the number of elements in the buffer.
 */
int rbAddBottomElement(RingBuffer rb, Pointer element)
{
  if (rbIsFull(rb)) {
    /* a buffer has limited capacity, so no resizing here */
    fprintf(stderr, "WARNING: rbAddBottomElement: buffer overflow\n");
    abort();
  }

  rb->first--;
  if (rb->first < 0)
    rb->first = rb->capacity-1;

  rb->entries[rb->first] = element;

  return (++rb->count);
}

/* ----------------------------------------------------------------------
 * returns the last inserted element without removing it.
 *
 * @param rb ringbuffer from which top element is returned.
 * @returns the last inserted element from the ring buffer rb.
 */
Pointer rbTopPeek(RingBuffer rb)
{
  if (rbIsEmpty(rb)) {
    fprintf(stderr, "WARNING: rbTopPeek: buffer empty\n");
    abort();
  } else {
    if (rb->last)
      return (rb->entries[rb->last-1]);
    else
      return (rb->entries[rb->capacity-1]);
  }
}

/* ----------------------------------------------------------------------
 * returns the first inserted element without removing it.
 *
 * @param rb ringbuffer from which the bottom element is returned.
 * @returns the first inserted element from the ringbuffer rb.
 */
Pointer rbBottomPeek(RingBuffer rb)
{
  if (rbIsEmpty(rb)) {
    fprintf(stderr, "WARNING: rbBottomPeek: buffer empty\n");
    abort();
  } else {
    return (rb->entries[rb->first]);
  }
}

/* ----------------------------------------------------------------------
 * finds the current size of the buffer.
 *
 * @param rb  ringbuffer whose size is returned.
 * @returns the current number of elements in the buffer.
 */
int rbSize(RingBuffer rb)
{
  return (rb->count);
}

/*----------------------------------------------------------------------
 * finds the maximum capacity of the specified buffer.
 *
 * @param rb ringbuffer whose maximum capacity is returned.
 * @returns the maximum number of elements the buffer rb can hold.
 */
int rbCapacity(RingBuffer rb)
{
  return (rb->capacity);
}

/* ----------------------------------------------------------------------
 * removes the element from the top of the buffer.
 *
 * @param rb  ringbuffer from which the removal is done.
 * @returns the last inserted element in the buffer rb.
 */
Pointer rbRemoveTopElement(RingBuffer rb)
{
  if (rbIsEmpty(rb)) {
    fprintf(stderr, "WARNING: rbRemoveTopElement: buffer empty\n");
    abort();
  } else {
    if(rb->last)
      rb->last--;
    else
      rb->last = rb->capacity-1;

    rb->count--;
    return (rb->entries[rb->last]);
  }
}

/* ----------------------------------------------------------------------
 * removes the element from the bottom of the buffer.
 *
 * @param rb ringbuffer from which the removal is done.
 * @returns the first inserted element in the buffer rb.
 */
Pointer rbRemoveBottomElement(RingBuffer rb)
{
  if (rbIsEmpty(rb)) {
    fprintf(stderr, "WARNING: rbRemoveBottomElement: buffer empty\n");
    abort();
  } else {
    Pointer element = rb->entries[rb->first];
    if(++rb->first == rb->capacity)
      rb->first = 0;
    rb->count--;
    return element;
  }
}

/* ----------------------------------------------------------------------
 * checks if the ringbuffer contains a specified element.
 *
 * @param rb ringbuffer  in which the search has to be made.
 * @param element the element to be seached in the ringbuffer rb.
 * @returns \a TRUE if the buffer contains element, \a FALSE otherwise.
 */
Boolean rbContains(RingBuffer rb, Pointer element)
{
  int i;

  i = rb->first; 
  while(i != rb->last) {
    if (rb->entries[i] == element)
      return TRUE;
    if(++i == rb->capacity) 
      i = 0;
  }

  return FALSE;
}

/* ----------------------------------------------------------------------
 * makes a copy of the container from \a src to \a dst.
 *
 * @param src this refers to the source ring buffer that has to be copied to the destination buffer.
 * @param dst this refers to the destination buffer into which the copying has been done.
 * @returns the dst ringbuffer after copying.
 */
RingBuffer rbCopy(RingBuffer dst, RingBuffer src)
{
   if (dst->capacity < src->capacity) {
      dst->entries = (Pointer *)
	 memRealloc(dst->entries, sizeof(Pointer) * src->capacity);
   }

   memcpy(dst->entries, src->entries, sizeof(Pointer) * src->capacity);

   dst->count = src->count;
   dst->capacity = src->capacity;
   dst->first = src->first;
   dst->last = src->last;

   return dst;
}

/* ----------------------------------------------------------------------
 * makes a new copy of the ring buffer.
 *
 * @param src source of the ringbuffer to be cloned.
 * @returns a new copy of a ring buffer after cloning.
 */
RingBuffer rbClone(RingBuffer src)
{
  RingBuffer dst;

  dst = rbNew(src->capacity);
  return (rbCopy(dst, src));
}

/* ----------------------------------------------------------------------
 * iterate over all items in the ring buffer and apply a function
 *
 * @param rb this is the ring buffer in which the iteration is done.
 * @param f this is the function that is used for the iteration in the ring buffer rb.
 * @param data determines the data in the ringbuffer.
 */
void rbForEachWithData(RingBuffer rb, VoidFunction *f, Pointer data)
{
  int i;

  i = rb->first;
  while(i != rb->last) {
    f(rb->entries[i], data);
    if(++i == rb->capacity)
      i = 0;
  }
}

/* ----------------------------------------------------------------------
 * empties the buffer.
 *
 * @param rb refers to the buffer that is to be emptied.
 */
void rbClear(RingBuffer rb)
{
  rb->first = rb->last = rb->count = 0;
}

/* ---------------------------------------------------------------------- */
/* -- ENDOFFILE --------------------------------------------------------- */
/** @} */


