#!/usr/bin/env perl
# /data/foth/distort-grammar.new/distort-grammar.pl

use warnings;


# TODO: insert baseform declarations where they miss!

my $grammarfile;

my ($latcount, $annocount, $wfcount);

my %ids;          # map "der" => ("der_nom_sg_masc", ...)
my %baseforms;    # map "hielten" => "halten"
my %paradigms;    # map "halten" => ("halte_sg_1", "hltst_sg_2", ...)

my %lattices;
my %annotations;
my %lat2anno;
my $mainlevel;

my $nrtokens;

# get args
use Getopt::Std;
getopts('d:v:');
my $distance = defined ($opt_d) ? $opt_d : 0;
my $verbose = defined ($opt_v) ? $opt_v : 0;

use strict;

die "usage: $0 [-v NUM] -d NUM <grammar file>\n" if $#ARGV;

$grammarfile = shift;

if($distance == 0) {
  system "cat $grammarfile";
  exit;
}

print "/*\n";

open IN, $grammarfile or die "Can't open $grammarfile: $!";
while (<IN>) {

  # find main level
  if (/mainlevel\s*=\s*yes/) {
    /^\s*(\S+)/;
    $mainlevel = $1;
  }

  # parse lexical entry
  elsif (/(\S+)\s*:=\s*(\S+)\s*:/) {
    my $id = $1;
    my $form = $2;
    push @{$ids{$form}}, $id;

    my $line = $_;
    while ($line !~ /;/) {
      if ($line =~ /baseform:\s*([^,\]]+)/) {
        my $baseform = $1;
        print "$id --> $baseform\n" if $verbose > 1;
        $baseforms{$form} = $baseform;
        if (!exists($paradigms{$baseform})) {
          $wfcount++;
          print "Found word field `$baseform'\n" if $verbose;
        }
        push @{$paradigms{$baseform}}, $form;
        last;
      }
      $line = <IN>;
    }
  }
  # parse lattice
  elsif (/^\s*([\"\'])(.*)\1\s*:\s*([\'\"])(.*)\3\s*:(.*)$/) {
    my ($latticeid, $annoid, $rest) = ($2, $4, $5);
    my @lattice;
    my $line;
    # in-line sentence
    if ($rest =~ /;/) {
      $rest =~ s/,\s*/ /g;
      $rest =~ s/^\s+//g;
      $rest =~ s/;//g;
      my $i;
      foreach (split(/\s+/, $rest)) {
        my %arc;
        $arc{from} = $i++;
        $arc{to} = $i;
        $arc{word} = $_;
        push @lattice, { %arc };
      }
    }
    # enumerated sentence
    else {
      do {
        $line = <IN>;
        if ($line =~ /^,?\s*(\d+)\s+(\d+)\s+(.*)\s*$/) {
          my %arc;
          $arc{from} = $1;
          $arc{to} = $2;
          $arc{word} = $3;
          push @lattice, { %arc };
        }
      } while $line !~ /;/;
    }
    $lattices{$latticeid} = [ @lattice ];
    $lat2anno{$latticeid} = $annoid;
    $latcount++;
    print "Found lattice `$latticeid'.\n" if $verbose;

  }
  # parse annotation
  elsif (/^\s*([\'\"])(.*)\1\s*<->/) {
    my $annoid = $2;
    my $line;
    my @annotation;
    # insert dummy entry so we can count from 1 like cdgp does
    push @annotation, 0;
    my %annoentry;
    do {
      $line = <IN>;
      if ($line =~ /^\s*(\d+)\s+(\d+)\s+(\S+)/) {
        $annoentry{from} = $1;
        $annoentry{to} = $2;
        $annoentry{word} = $3;
      } elsif ($line =~ m|id/'(.*)'|) {
        $annoentry{id} = $1;
      } elsif ($line =~ /^\s*$mainlevel\s*->\s*'(.*)'\s*->\s*(\d+)/) {
        $annoentry{label} = $1;
        $annoentry{parent} = $2;
        push @annotation, { %annoentry };
        $nrtokens++;
      }
    } while $line !~ /;/;

    $annotations{$annoid} = [ @annotation ];
    $annocount++;

    print "Found annotation `$annoid'.\n" if $verbose;

  }

}

close IN;

print "Found main level $mainlevel.\n" if(defined $mainlevel);
print "Found $wfcount word fields.\n";
print "Found $latcount lattices.\n";
print "Found $annocount annotations.\n";

foreach (keys %lattices) {
  if (!defined $annotations{$lat2anno{$_}}) {
    print "WARNING: Annotation `$lat2anno{$_}' for lattice `$_' not found!\n";
    delete $lattices{$_};
  }
}

# distort
my $target = $nrtokens / $distance;
my $work = 0;
my $count;
while ($work < $target ) {
  my $x = rand();
  if($x < 0.5) {
    $work += exchange_word(one_of(keys %lattices));
  } elsif ($x < 0.8) {
    $work += delete_word(one_of(keys %lattices));
  } else {
    $work += swap_children(one_of(keys %lattices));
  }
 last if $count++ > 10*$target;
}

print "WARNING: only " if $work < $target;
print "$work distortions done\n";
print "*/\n";


# dump
foreach my $id (keys %lattices) {
  print_lattice($id);
  print_anno($lat2anno{$id});

  check_sentence($id);
}


sub rebuild_lattice {
  my $id = shift;
  my $anno = $annotations{$id};
  my @l;
  foreach my $entry (@$anno) {
    next if $entry == 0;
    my %arc;
    $arc{from} = $$entry{from};
    $arc{to} = $$entry{to};
    $arc{word} = $$entry{word};
    push @l, { %arc };
  }
  $lattices{$id} = [ @l ];

}


# swap two subtrees in the analysis
# returns number of swap operations performed
sub swap_children {
  my $name = shift;
  my $pos;
  my $a = $annotations{$lat2anno{$name}};
  my $nrwords = $#$a;

  print "Exchanging constituents in $name\n" if $verbose;

  # extract tree structure from anno
  my @children;
  my @parents;
  my $count;
  foreach my $entry (@$a) {
    # skip dummy
    next unless $count;
    my $down = $count;
    my $up = $$entry{parent};
    $parents[$down] = $up;
    push @{$children[$up]}, $down;
    print "$down modifies $up\n" if $verbose > 2;
  } continue {
    $count++;
  }

  # find parent (root is disallowed)
  if (defined $_[0]) {
    $pos = shift;
  } else {
    $pos = one_of(0..$nrwords);
  }
  if ($pos > $nrwords) {
    die "Anno for `$name' has no word $pos.\n";
  }

  # step along until a suitable node is found
  my ($i, $found);
  while ($i++ <= $nrwords && !$found) {
    $found = $#{$children[$pos]} > 0;
    last if $found;
    $pos = (($pos + 1) % ($nrwords + 1));
  }

  print_anno($lat2anno{$name}) if $verbose > 2;

  if ($#{$children[$pos]} < 1) {
    print "WARNING: can't exchange constituents in $name.\n";
    return 0;
  }

  # find children
  my @candidates = @{$children[$pos]};
  my $scylla = one_of(@candidates);

  my $charybdis = $scylla;
  while ($scylla == $charybdis) {
    $charybdis = one_of(@candidates);
  }

  # guarantee order
  if ($scylla>$charybdis) {
    ($scylla, $charybdis) = ($charybdis, $scylla);
  }

  print "parent: $pos\ncandidates: @candidates\nexchanging: $scylla $charybdis\n"
    if $verbose > 2;

  # find spheres of interest
  my @desca = subtree($scylla, \@children);
  my @descb = subtree($charybdis, \@children);
  my $amin = $desca[0];
  my $amax = $desca[-1];
  my $awidth = $amax - $amin + 1;
  my $bmin = $descb[0];
  my $bmax = $descb[-1];
  my $bwidth = $bmax - $bmin + 1;

  print "descendants of $scylla:  @desca\n" if $verbose > 2;
  print "descendants of $charybdis:  @descb\n" if $verbose > 2;

  # complain about non-projective subtrees
  if($#desca != $amax - $amin) {
    print "WARNING: nonprojective structures in $name under node $scylla!\n";
    return 0;
  }
  if($#descb != $bmax - $bmin) {
    print "WARNING: nonprojective structures in $name under node $charybdis!\n";
    return 0;
  }

  print "amin = $amin  amax = $amax awidth = $awidth\n" if $verbose > 2;
  print "bmin = $bmin  bmax = $bmax bwidth = $bwidth\n" if $verbose > 2;

  # build map of new node numbers:
  # - nodes before amin remain
  # - increment nodes amin through amax by bmax-amax
  # - increment nodes amax+1 through bmin-1 by bwidth-awidth
  # - increment nodes bmin through bmax by amin-bmin
  my @map;
  foreach my $node (0..$nrwords) {
    if($node == 0) {
      $map[0] = $node
    } elsif($node < $amin) {
      $map[$node] = $node;
    } elsif ($node <= $amax) {
      $map[$node + $bmax - $amax] = $node;
    } elsif ($node < $bmin) {
      $map[$node + $bwidth - $awidth] = $node;
    } elsif ($node <= $bmax) {
      $map[$node + $amin - $bmin] = $node;
    } else {
      $map[$node] = $node;
    }
  }
  print "Reordering map: @map\n" if $verbose > 2;
  my @pam;
  for($i = 0; $i <= $#map; $i++) {
    $pam[$map[$i]] = $i;
  }
  print "Inverse reordering map: @pam\n" if $verbose > 2;

  # reorder entries of the annotation,
  # apply map for governor shift,
  # and reassign trivial timepoints
  my @b;
  for($i = 0; $i <= $nrwords; $i++) {
    my $entry = $$a[$map[$i]];
    # skip dummy
    if($i) {
      $$entry{parent} = $pam[$$entry{parent}];
      $$entry{from} = $i-1;
      $$entry{to} = $i;
    }
    $b[$i] = $entry;
  }
  $annotations{$lat2anno{$name}} = [ @b ];

  rebuild_lattice($name);
  return 1;
}


sub delete_word {
  my $name = shift;
  my $pos;
  my $a = $annotations{$lat2anno{$name}};
  my $nrwords = $#$a;

  # don't delete last word
  return 0 unless $#$a > 1;
  print "Deleting word from $name\n" if $verbose;

  # delete which word?
  if (defined $_[0]) {
    $pos = shift;
  } else {
    $pos = one_of(1..$nrwords);
  }

  if ($pos > $nrwords) {
    print "WARNING: Anno for `$name' has no word $pos.\n";
    return 0;
  }

  # find doomed word
  my $entry = $$a[$pos];
  my $parent = $$entry{parent};
  my $from =  $$entry{from};
  my $to =  $$entry{to};
  $parent-- if $parent >= $pos;

  if($verbose > 1) {
    print "  deleting `$a->[$pos]->{word}'\n";
  }

  # delete from annotation
  @$a = @$a[0..$pos-1, $pos+1..$#$a];

  # adjust others
  my $i = 0;
  foreach $entry(@$a) {
    next if $entry == 0;
    $$entry{from} = $i;
    $i = $$entry{to};
    if ($$entry{parent} == $pos) {
      $$entry{parent} = $parent;
    } elsif ($$entry{parent} > $pos) {
      $$entry{parent}--;
    }
  }
  rebuild_lattice($name);

  return 1;

}


sub exchange_word {
  my $name = shift;
  my $pos;
  my $a = $annotations{$lat2anno{$name}};
  my $nrwords = $#$a;

  print "Exchanging word in $name\n" if $verbose;

  # exchange which word?
  if (defined $_[0]) {
    $pos = shift;
  } else {
    $pos = one_of(1..$nrwords);
  }

  if ($pos > $nrwords) {
    print "Anno for `$name' has no word $pos.\n";
    return 0;
  }

  # step along until a changeable form is found
  my ($oldid, $oldform);
  my ($i, $found, $baseform);
  while ($i++ <= $nrwords && !$found) {
    $oldid = $$a[$pos]{id};
    $oldform = $$a[$pos]{word};
    $baseform = $baseforms{$oldform};
    $found = defined($baseform);
    last if $found;
    $pos++;
    if($pos > $nrwords) {
      $pos = 1;
    }
  }

  if (!$found) {
    print "WARNING: can't inflect any word in $name.\n";
    return 0;
  } else {
    print "SUCCESS: can inflect a word in $name.\n";
  }

  # find new form
  my @p = @{$paradigms{$baseform}};
  my $newform = one_of(@p);
  my $newid = one_of(@{$ids{$newform}});

  # distort annotation
  $$a[$pos]{word} = $newform;
  $$a[$pos]{id} = $newid;

  print "$name: $oldform --> $newform\n" if $verbose > 1;

  rebuild_lattice($name);

  return 1;

}


sub print_lattice {
  my $name = shift;
  my $first;
  my $l = $lattices{$name};
  printf "\"%s\" : \"%s\" :", $name, $lat2anno{$name};
  foreach my $a (@$l) {
    print $first++ ? ",\n" : "\n";
    printf "  %2d %2d %s", $$a{from}, $$a{to}, $$a{word};
  }
  print ";\n";
}

sub print_words {
  my $l = shift;
  foreach my $a (@$l) {
    print " $$a{word}";
  }
  print "\n";
}


sub print_anno {
  my $name = shift;
  my $a = $annotations{$name};
  my $first;
  print "'$name' <->";
  foreach my $e (@$a) {
    # skip dummy
    next if $e == 0;
    print $first++ ? ",\n" : "\n";
    print "  /* node $first */\n";
    printf "  %d %d %s\n", $$e{from}, $$e{to}, $$e{word};
#    printf "    id/'%s'\n", $$e{id};
    printf "    %s -> '%s' -> %d", $mainlevel, $$e{label}, $$e{parent};
  }
  print ";\n\n";
}

sub print_arc {
  my $a = shift;
  printf "  %2d %2d %s\n", $$a{from}, $$a{to}, $$a{word};
}

sub subtree {
  my $node = shift;
  my $children = shift;
  my @result = ($node);

  my @tmp = @result;
  my $progress = 1;
  while ($progress) {
     foreach my $parent (@result) {
       foreach my $child (@{${$children}[$parent]}) {
       add_to_list(\@tmp, $child);
     }
     $progress = (length(@tmp) > length(@result));
     @result = @tmp;
   }
  }
  return sort {$a<=>$b} @result;
}

sub contains {
  my $list = shift;
  my $elem = shift;

  foreach (@$list) {
    return 1 if $_ == $elem;
  }
  return 0;
}

sub add_to_list {
  my $list = shift;
  my $elem = shift;
  if (!contains($list, $elem)) {
    push @$list, $elem;
  }
}

sub check_sentence {
  my $name = shift;
  my $l = $lattices{$name};
  my $a = $annotations{$lat2anno{$name}};
  my ($amin, $amax) = my ($lmin, $lmax) = (999,0);

  foreach my $a (@$l) {
    $lmin = min($lmin, $$a{from});
    $lmax = max($lmax, $$a{to});
  }

  foreach my $e (@$a) {
    next if $e == 0;
    $amin = min($amin, $$e{from});
    $amax = max($amax, $$e{to});
  }
  die unless $lmin == $amin;
  die unless $lmax == $amax;

}

sub min() { $_[0] < $_[1] ? $_[0] : $_[1]; }
sub max() { $_[0] > $_[1] ? $_[0] : $_[1]; }

sub one_of { $_[int(rand($#_+1))] }
