#!/usr/bin/env perl
# cdg/grammar/heise/eval-tagger.pl

# Measure correctness of a tagger on an annotated corpus.
use warnings;
use strict;

use File::Spec;

our @corpus;
our %stat;

my $tntdir = $ENV{TNTDIR} || '/data/linux/opt/tnt';

# structure of defined taggers;
# you only need to give a command for invocation really.
our %taggers =
  (
   TreeTagger =>
   {
    invocation =>
    '/data/linux/opt/TreeTagger/bin/tree-tagger' .
    ' -quiet -token -threshold 0.1 -prob . ' .
    '/data/linux/opt/TreeTagger/lib/papa-german.par'

   },

   tnt =>
   {
    invocation =>
    "$tntdir/tnt -v0 -b$tntdir/models/papa.lex -z200 $tntdir/models/negra.tnt -",
   },

   papa =>
   {
    invocation => "deutsch-tagger.pl"
   }
  );

die "Usage: $0 <tagger name> <corpus>\n" unless $#ARGV == 1;

our $tagger = shift;
die "Defined taggers: " . (join " ", keys %taggers) . "\n"
  unless exists $taggers{$tagger};

our %similar;
for(qw (
      FM:XY NN:XY ADJA:NN ADJA:NN ADJD:ADV ADJD:PTKVZ ADJD:VVPP
      ADV:PTKVZ CARD:NN FM:NE FM:NN NE:ADJA NE:FM NE:NN NE:XY ADJA:FM
      ADJA:NN ADJA:XY NN:CARD NN:FM NN:NE NN:PIS PIAT:PIDAT PIDAT:PIAT
      PRELS:PWS PWS:PRELS VMFIN:VVFIN VVPP:ADJD
       )) {
  $similar{$_} = 1;
}



sub interpret {
  my ($tuple, $line, $cat) = @_;
  while ($line =~ s/(\S+)\s+(\S+)\s+//) {
    my($c,$s) = ($1,$2);
    $tuple->{$c} = $s;
    if(!exists $tuple->{max} ||
       $$tuple{$$tuple{max}} < $s ||
       $$tuple{$$tuple{max}} == $s && $c eq $cat) {
      $tuple->{max} = $c
    }
  }
}

# remove quotes from a string
sub parse_string {
  my $s = shift;
  if ($s =~ /^([\'\"])(.*)\1$/) {
    $s = $2;
    $s =~ s|\\||g;
  }
  return $s;
}


# get annotated categories
print "Creating stream of words...\n";

our $corpus = shift;
if (-d $corpus) {
  open (ANNOS, "cd $corpus;ls|grep cda|xargs cat|")
    or die "Can't cat annotation files.\n";
  my $last = 'cat';
  my %pair;
  while (<ANNOS>) {
    chomp;
    if (/^ *(\d+) +\d+ +(.+)$/) {
      die "WARNING: Expected category, got `$_'\n"
        if $last eq 'word';
      my($from,$word) = ($1,$2);
      $pair{word} = parse_string $word;
      $last = 'word';
      if(0 == $from) {
        push @corpus, { word => '#EOS#', cat => 'XY' };
      }
    } elsif (/^\'cat\' *\/ *\'([^ ]+)\'/) {
      die "WARNING: Expected word, got `$_'\n"
        if $last eq 'cat';
      $pair{cat} = parse_string $1;
      $last = 'cat';

      # guard against words with embedded whitespace
      $pair{word} =~ s/ /_/g;
      push @corpus, { %pair };
    }
  }
} else {

  open (ANNOS, $corpus) or die "Can't open corpus.\n";
  my $last = 'cat';
  my %pair;
  my $on = 0;
  while (<ANNOS>) {
    chomp;
    if (/^#BOS/) {
      $on = 1;
      next;
    }
    if (/^#EOS/) {
      push @corpus, { word => '#EOS#', cat => 'XY' };
      $on = 0;
      next;
    }
    if (!$on) {
      next;
    }
    if (/^([^#\t]+)\t+([^\t]+)/) {
      %pair = ( word => $1, cat => $2 );
      # guard against words with embedded whitespace
      $pair{word} =~ s/ /_/g;
      push @corpus, { %pair };
    }
  }
}

# deduce base name of the corpus
$corpus = File::Spec->rel2abs($corpus);
if($corpus =~ m|/([^/]+)/annos/?$|) {
  $corpus = $1;
} elsif($corpus =~ m|/([a-z]+)([-_][^/]+).export|) {
  $corpus = $1;
} else {
  my ($volume,$directories,$file) = File::Spec->splitpath($corpus);
  $corpus = $file;
}

# write list of words only
open(WORDS, ">words") or die "Can't write to `words'\n";
for (@corpus) {
  if($$_{word} eq '#EOS#') {
    print WORDS "\n";
  } else {
    print WORDS "$_->{word}\n";
  }
}
close(WORDS);

# call tagger
print "Calling tagger...\n";
open(TAGGER,
     "cat words |" .
     " $taggers{$tagger}{invocation} 2>/dev/null |tee ${tagger}.${corpus}.output |")
  or die "Can't call tagger\n";
my $i = 0;

while (<TAGGER>) {
  next unless s/^([^\t]+)\t+//;

  my $word = $1;

  my $pair = $corpus[$i++];

  if($$pair{word} eq '#EOS#') {
    $pair = $corpus[$i++];
  }

  # TreeTagger sometimes overrules our tokenization,
  # turning gibt+'s into gibt+'+s.
  # There's nothing we can do about this except skip the '.
  if ($word ne $$pair{word}) {
    warn "Mismatch: got `$word', expected `$$pair{word}'\n";
    warn "Have to skip `$pair->{word}'\n";
    <TAGGER>;
    next;
  }

  # actually process one line
  my %tagged = ();
  interpret(\%tagged, $_, $$pair{cat});
  $corpus[$i-1]{tagger} = { %tagged };
}
close(TAGGER);

# print transcript
my $words = 0;
my $errors = 0;
my $baderrors = 0;
my $p = 0;

my $logname = "$tagger.$corpus.log";
open(FILE, ">$logname") or die "Can't write `>$logname': $!\n";
for my $w (@corpus) {

  if($$w{word} eq '#EOS#') {
    print FILE "\n";
    next;
  }

  $words++;

  if (!exists($w->{tagger})) {
    print FILE "Tokenizer error for $w->{word}\n";
    $errors++;
  }

  else {
    printf FILE "%-24s", "$w->{cat}:$w->{tagger}{max}";
    printf FILE "%-25s", "$w->{word}";
    for (keys %{$w->{tagger}}) {
      next if /^max$/;
      if($w->{cat} eq $_) {
        print FILE "$_ [$w->{tagger}{$_}] ";
      } else {
        print FILE "$_ $w->{tagger}{$_} ";
      }
    }
    print FILE "\n";

  }

  # find errors
  if (exists $w->{tagger}) {
    if ($w->{tagger}{max} ne $w->{cat}) {
      $errors++;
      $stat{"$w->{cat}:$w->{tagger}{max}"}++;
      if(!exists($similar{"$w->{cat}:$w->{tagger}{max}"})) {
        $baderrors++;
      }
    }
    if (exists($w->{tagger}{$w->{cat}})) {
      $p += $w->{tagger}{$w->{cat}};
    }
  }
}

# print statistics about what errors happen
for (sort { $stat{$a} <=> $stat{$b} } keys %stat) {
  print FILE exists($similar{$_}) ? ' ' : '*';
  print FILE "$_: $stat{$_}\n";
}

# give recall percentage
my $correct = $words - $errors;
printf FILE "         single:      %4.3f%% ($correct/$words)\n", 100 * $correct/$words;
printf "         single:      %4.3f%% ($correct/$words)\n", 100 * $correct/$words;

# give average p of the correct tag
printf FILE "         multi:       %4.3f%% \n", 100 * $p/$words;
printf "         multi:       %4.3f%% \n", 100 * $p/$words;

$correct = $words - $baderrors;
printf FILE "nonstrict single:      %4.3f%% ($correct/$words)\n", 100 * $correct/$words;
printf "nonstrict single:      %4.3f%% ($correct/$words)\n", 100 * $correct/$words;

printf FILE "nonstrict multi:       %4.3f%% \n", 100 * $p/$words;
printf "nonstrict multi:       %4.3f%% \n", 100 * $p/$words;



close(FILE);
