#!/usr/bin/env perl
# cdg/utils/posdebug.pl

# reads a transcript of POS tagging performance and answers requests
# about occurring patterns of tags

use warnings;
use strict;

use Term::ReadLine;
our $term = new Term::ReadLine 'posdebug';
our @words;
our @stretches;
our %pos;
our %pos2char;
our %char2pos;

our $interrupted;
our $debug = 0;

sub print_pos {
  my $s = shift;
  for my $w(@{$$s{tokens}}) {
    print "$$w{word}/$$w{pos} ";
  }
  print "\n";
}

sub print_recog {
  my $s = shift;
  for my $w(@{$$s{tokens}}) {
    print "$$w{word}/$$w{recog} ";
  }
  print "\n";
}

sub translate_request {
  my $r = shift;
  $r =~ s/^ +//;
  $r =~ s/ +$//;
  my $result = '#';

  for my $atom (split(/ +/, $r)) {

    # regex quantifiers and anchors stand for themselves
    if($atom =~ /^[\+\*\?]$/) {
      $result .= "$atom";
    }

    elsif($atom eq '^') {
      $result = "^$result";
    }
    elsif($atom eq '$') {
      $result = "$result\$";
    }

    # word specifications are translated into a format that allows us
    # to use perl's regex mechanism. If the user says `der/PRELS' it
    # is translated into `der#x#' because PRELS happens to be the 24th
    # POS seen by the program.
    else {
      my $cat;
      my $word;

      if($atom =~ m|^(.+)/(.+)$|) {
        ($word,$cat) = ($1,$2);
      } else {
        ($cat,$word) = ($atom,'[^#]+');
      }

      # If the user says `de./ART', we must not actually allow
      # any last character -- # is always reserved for the program.
      $word =~ s/\./[^#]/g;

      my $class = '';
      if($cat =~ s/^!//) {
        $class = '^';
      }
      for my $p (keys %pos) {
        if($p =~ /$cat/) {
          $class .= $pos2char{$p};
        }
      }
      if(!$class) {
        print "ERROR: no POS tag matches `$cat'.\n";
        return '';
      }
      $result .= "($word#[$class]#)";
    }
  }
  return $result;
}

sub highlight {
  my $s = shift;
  my $i = 0;
  for my $w(@{$$s{tokens}}) {
    if(defined $_[0] && $_[0] == $i) {
      shift;
      print "_$$w{word}_ ";
    } else {
      print "$$w{word} ";
    }
    $i++;
  }
  print "\n";
}

# Check whether S is of the type requested.
#
# Legal requests are:
#
# pos PRELS                 ==> find everything with a PRELS in it
# pos VVINF VAINF           ==> find VVINF directly followed by VAINF
# pos ^ VVINF               ==> find initial VVINF
# pos VVINF $               ==> find final VVINF
# pos !^V + VVINF           ==> find VVINF not preceded by any verb
sub interpret {
  my $request = shift;
  $request =~ s/^ +//;
  $request =~ s/ +$//;
  my $mode;
  my $no = 0;

  if($request eq '+') {
    $debug++;
    return;
  }
  if($request eq '-') {
    $debug--;
    return;
  }

  if($request =~ s/^ *(pos|recog) *//) {
    $mode = $1;
    $request = translate_request($request);
    return unless $request;
    my $re = qr/$request/;
    $interrupted = 0;
    for my $s (@stretches) {
      my $test = $mode eq 'pos' ? $$s{posstring} : $$s{recogstring};
      if($test =~ $re) {
        $no++;
        $mode eq 'pos' ? print_pos $s : print_recog $s;
        last if $interrupted;
      }
    }
  }

  elsif($request =~ /error +([^ ]+):([^ ]+)/) {
    my ($word, $correct, $incorrect) = (qr/./,$1,$2);
    if($correct =~ /^(.+):(.+)$/) {
      ($word,$correct) = (qr/$1/,$2);
    }
    for my $s(@stretches) {
      my $i = 0;
      my @result = ();
      for my $w(@{$$s{tokens}}) {
        if($$w{pos} eq $correct &&
           $$w{recog} eq $incorrect &&
           $$w{word} =~ $word) {
          push @result, $i;
          $no++;
        }
        $i++;
      }
      if(@result) {
        highlight $s, @result;
      }
    }
  } else {
    warn "Don't understand `$request'.\n";
  }

  if($no) {
    print "$no matches\n";
  }

}


# read_events
my $logfile = shift;
open(IN, "$logfile") or die "Can't read `$logfile': $!\n";
while (<IN>) {
  if (my ($pos,$recog,$word) = /^([^ ]+):([^ ]+) +([^ ]+) +/) {
    push @words, {pos=>$pos, recog=>$recog, word=>$word};
    $pos{$pos}++;
    $pos{$recog}++;
  } else {
    push @words, { word => '#EOS#'};
  }
}
close(IN);

# cut into stretches
my @s;
for(@words) {
  if($$_{word} =~ /^([,.:;\-!?]|--+|#EOS#)$/) {
    if(@s) {
      push @stretches, {tokens=>[@s]};
      @s=();
    }
    next;
  }
  else {
    push @s, $_;
  }
}

# build pos <-> character map
my $s = 'a';
for(keys %pos) {
  $pos2char{$_} = $s;
  $char2pos{$s} = $_;
  $s++;
  if($s eq 'aa') { $s = 'A'};
  if($s eq 'AA') { $s = '0'};

}

# translate all stretches into character strings
for my $s (@stretches) {
   my $result = '#';
   for my $w(@{$$s{tokens}}) {
     $result .= $$w{word};
     $result .= '#';
     $result .= $pos2char{$$w{pos}};
     $result .= '#';
   }
   $$s{posstring} = $result;

   $result = '#';
   for my $w(@{$$s{tokens}}) {
     $result .= $$w{word};
     $result .= '#';
     $result .= $pos2char{$$w{recog}};
     $result .= '#';
   }
   $$s{recogstring} = $result;
}

$SIG{INT} = sub {
  $interrupted++;
};

# read user's command history
my $historyfile = "$ENV{HOME}/.posdebugrc";
if (-r $historyfile) {
  for (`cat $historyfile`) {
    chomp;
    $term->addhistory($_) if /[^ ]/;
  }
}


# interact
while ( defined ($_ = $term->readline("Find pattern: >"))) {
  interpret $_;
}

# write updated command history
if (open(SAVEHIST,">$historyfile")) {
  print SAVEHIST (join "\n",$term->history_list());
}
