#!/usr/bin/env perl
# cdg/grammar/heise/search-annotations.pl

# Finds particular phenomena in the annotations. Example:
#
#  Find right-branching appositions between proper nouns:
#  X/SYN & X@cat = NN & X^cat = NN & X.label = APPO
#
# Automatic transformation of analyses is also possible. Say that the
# grammar writer changed the verb frame of `sagen' from transitive to
# bitransitive. That means that all its children with label C2 need to
# be converted to C3, and the mirror edges changed accordingly. Here's
# how to do that:
#
#  X.label = C2 & X^baseform = sagen & Y:OBL2 & Y/X & Z|OBL3 & Z/X
#   -> X.label := C3 & Z := Y & Y := NIL
#
# Each changed annotation will be written back to disk immediately in
# the cwd, under the name <anno id>.cda. If this is the file you
# originally loaded the annotation from, it is overwritten.
use warnings;
use strict;

BEGIN {
 if(!defined $ENV{CDGHOME}) {
   die "Please set CDGHOME.\n";
 }
 $ENV{CDGPERL} = "$ENV{CDGHOME}/bindings";
}

use lib "$ENV{CDGPERL}";
use lib "$ENV{CDGPERL}/.libs";
use CDG;
use CDG::Utils qw(:statistics :functions);

our %options =
  (
   # display debug output?
   debug => 0,

   # when displaying dependency edges,
   # should we print the modifier at the left,
   # or the earlier word at the left?
   print_edge_direction => 'modifier',

   # How much information to show about the solution sets
   verbosity => 1
  );

our %annotations;

# get args
#
# -d: debug on
# -p: pipe commands from STDIN
# -v: set verbosity level
#
use Getopt::Std;
our ($opt_d, $opt_p, $opt_v);
getopts('dpv:');
if($opt_v) {
  $options{verbosity} = $opt_v;
}

$options{debug} = defined $opt_d;
sub debug {
  print @_ if $options{debug};
}
sub printif {
  my $level = shift;
  print @_ if $options{verbosity} >= $level;
}

# constants for edge connexions
our ($CTOP, $CBOTTOM, $COVER, $CUNDER, $CINVERSE, $CIDENT, $CNONE)
  = qw (/\      \/        /       \        ~=         ==      ||);

our $conn_re = join "|", map { quotemeta }
  ($CTOP,$CBOTTOM,$COVER,$CUNDER,$CINVERSE,$CIDENT,$CNONE);

our $dir_re = '[!|\\\/:]';
our $op_re = '=~|=|!=|!~|<|>|<=|>=|\.';
our $var_re = '[A-Z]';             # 26 variables ought to be enough for anybody
our $string_re =
  '(?:\'(?:(?:[^\\\\\']|\\\\.)*)\'|' .
  '"(?:(?:[^\\\\"]|\\\\.)*)"|' .
  '[a-zA-Z_\x80-\xff][a-zA-Z_0-9\x80-\xff]*)';
our $number_re = '[0-9]+';
our $att_re = "(?:$string_re|$number_re)(?::(?:$string_re|$number_re))*";

# lengthy queries can be interrupted and set this flag
our $interrupted = 0;

# I wanted to pull this from command.c, but SWIG wouldn't let me.
our %cdg_commands;
for (qw( activate annotation anno2parse arcconsistency chunk
         compareparses compile constraint deactivate distance edges frobbing
         genetic gls help hierarchy hook incrementalcompletion inputwordgraph
         isearch level levelsort lexicon license load ls net netdelete
         netsearch newnet nodeconsistency nonspeccompatible parsedelete
         printparse printparses pruning quit renewnet reset search section
         set showlevel status tagger testing useconstraint uselevel version
         verify weight wfst wordgraph writenet writeparses writewordgraph )) {
  $cdg_commands{$_}++;
}
our @cdg_commands = keys %cdg_commands;


# read terminal input comfortably
use Term::ReadLine;
our $term = new Term::ReadLine 'search-annotations';
$term->Attribs->{attempted_completion_function} = \&searchannos_completion;

# readline completion code stolen from soundgrab
# (http://rawrec.sourceforge.net/soundgrab/)
{
  my $list_index;
  my @name;
  my $current_line;
  my $current_start;
  my $current_end;

  sub command_generator {
    my ($text, $state) = @_;

    # If this is a new word to complete, initialize now.  This
    # includes saving the length of TEXT for efficiency, and
    # initializing the index variable to 0.
    unless ( $state ) {
      $list_index = 0;

      my $current_prefix = substr($current_line,0,$current_end);
      if ($current_prefix =~/^ *([a-z]*)$/) {
        # complete CDG shell command
        @name = @cdg_commands;
      } elsif ($current_prefix =~/^ *set +([^ ]*)$/) {
        # complete option
        @name = keys(%options);
      } elsif ($current_prefix =~ /\b[A-Z][@^] *[^ ]*$/) {
        # complete attribute
        @name = keys(%features);
      } elsif ($current_prefix =~ /[A-Z] *[@^] *([^ ]+) *($op_re|:=) *[^ ]* *$/) {
        # complete value
        @name = keys(%{$features{$1}});
      } elsif ($current_prefix =~ /\b[A-Z] *[\/\\|!:] *[^ ]* *$/) {
        # complete level
        @name = keys(%levels);
      } elsif ($current_prefix =~ /\b([A-Z]) *\. *label *:?= *[^ ]* *$/) {
        # complete label
        my $v = $1;
        # find out what level this variable belongs to
        my $level =
          $current_line =~ /\b$v *$dir_re *([^ ]+)/
            ? $1 : 'all';
        @name = keys(%{$labels{$level}})
      } else {
        # complete word form
        @name = keys(%words);
      }
    }
	
    # Return the next name which partially matches from the command list.
    while ( $list_index <= $#name ) {
      $list_index++;
      if ($name[$list_index - 1] =~ /^$text/) {
        return $name[$list_index - 1];
      }
    }
	
    # If no names matched, then return NULL.
    return undef;
  }

  # Attempt to complete the contents of TEXT.  START and END bound
  # the region of rl_line_buffer that contains the word to complete.
  # TEST is the word to complete.  We can use the entire contents of
  # rl_line_buffer in case we want to do some simple parsing.  Return
  # the array of matches, or NULL if there aren't any.
  sub searchannos_completion {
    my ($text, $line, $start, $end) = @_;
    $current_line = $line;
    $current_start = $start;
    $current_end = $end;
    my @matches = $term->completion_matches ($text, \&command_generator);
  }
}

# print one dependency edge on STDOUT
sub print_edge {
  my $e = shift;
  my $annotation = $$e{anno};

  my $down = $$e{child};
  my $up = $$e{parent};

  my $modifier = $annotation->{words}[$down];
  my $modifiee = $annotation->{words}[$up];

  if ($options{print_edge_direction} eq 'modifier' or !$up or $down < $up) {
    printf "%10s: %s--%s-->%s",
      $$e{level}, $$modifier{word}, $$e{label}, $$modifiee{word};
  } else {
    printf "%10s: %s<--%s--%s",
      $$e{level}, $$modifiee{word}, $$e{label}, $$modifier{word};
  }

}

# print entire annotation on STDOUT
sub print_anno {
  my $name = shift;
  my $annotation = $annotations{$name};

  my $first;
  print "'$name' <->";

  foreach (@{$annotation->{edges}}) {
    print $first++ ? "," : "";
    print_edge $_;
    print "\n";

  }
  print ";\n\n";
}

# return the word at the lower end of $e
sub modifier {
  my $e = shift;
  return undef unless defined $e;
  return $$e{'@'};
}

# return the word at the upper end of $e (or undef)
sub modifiee {
  my $e = shift;
  return undef unless defined $e;
  return $$e{'^'};
}

# return the edge above $e (or undef)
sub upper_edge {
  my $e = shift;
  my $level = shift;
  $level = $$e{level} unless defined $level;

  my $annotation = $$e{anno};
  foreach my $e2 (@{$annotation->{edges}}) {
    next if $$e2{level} ne $level;
    next unless $$e2{child} == $$e{parent};
    return $e2;
  }
  return undef;
}

# return all edges below $e
sub lower_edges {
  my $e = shift;
  my $level = shift;
  my @result = ();

  $level = $$e{level} unless defined $level;

  my $annotation = $$e{anno};
  foreach my $e2 (@{$annotation->{edges}}) {
    next if $$e2{level} ne $level;
    next unless $$e{child} == $$e2{parent};
    push @result, $e2;
  }
  return \@result;
}

# how long is $e?
# null edges have length 0,
# rightward subordinations have length > 0
sub edge_length {
  my $e = shift;
  if($$e{parent}) {
    return $$e{parent} - $$e{child};
  }
  return 0;
}

# does $e point in direction $dir?
sub match_dir {
  my $e = shift;
  my $dir = shift;

  return 1 if $dir eq ':';

  my $m1 = modifier($e);
  my $m2 = modifiee($e);
  my $down = $$m1{no};
  my $up = $$m2{no};
  $up = 0 unless defined $up;
  if ($dir eq '|') {
    return $up == 0;
  }
  if ($dir eq '!') {
    return $up != 0;
  }
  if ($up == 0) {
    return 0;
  }
  if ($dir eq '/') {
    return $down < $up;
  }
  return $down > $up;
}

# are $a and $b connected like $conn specifies?
sub match_conn {

  my $conn = shift;
  my $a = shift;
  my $b = shift;

  if ($conn eq $COVER) {
    return $$a{child} == $$b{parent}
  }

  if ($conn eq $CUNDER) {
    return $$b{child} == $$a{parent}
  }

  if ($conn eq $CTOP) {
    return $$b{parent} == $$a{parent}
  }

  if ($conn eq $CBOTTOM) {
    return $$b{child} == $$a{child}
  }

  if ($conn eq $CINVERSE) {
    return $$a{child} == $$b{parent} && $$a{parent} == $$b{child}
  }

  if ($conn eq $CIDENT) {
    return $$a{child} == $$b{child} && $$a{parent} == $$b{parent}
  }

  if ($conn eq $CNONE) {
    return $$a{child} != $$b{child}
      && $$a{parent} != $$b{parent}
        && $$a{child} != $$b{parent}
          && $$a{parent} != $$b{child}
        }

  return 0;
}

# implement CDG predicate `has'
sub has {
  my $edge = shift;
  my $label = shift;

  for my $e (@{$edge->{anno}->{edges}}) {
    if($e->{label} eq $label &&
       $e->{level} eq $edge->{level} &&
       $e->{parent} == $edge->{child}) {
      return 1;
    }
  }
  return 0;
}


# sort like ls -v does
sub byVName {

  # don't change the names themselves
  my ($x,$y) = ($a,$b);

  while ($x ne "" && $y ne "") {

    # skip identical alphabetic parts
    while (substr($x,0,1) eq substr($y,0,1) &&
           $x !~ /^[0-9]/) {
      substr($x,0,1) = "";
      substr($y,0,1) = "";
    }

    # compare alphabetical parts normally
    if ($x !~ /^\d/ || $y !~ /^\d/) {
      return $x cmp $y;
    }

    # compare numeric parts numerically
    $x=~s/^(\d+)//;
    my $nx = $1;
    $y=~s/^(\d+)//;
    my $ny = $1;
    my $result = $nx <=> $ny;
    next unless $result;
    return $result;
  }
}

# is $x numeric?
sub number { shift =~ /^ *-*[0-9]+(\.[0-9]*)? *$/ }

# is the equation $lhs $op $rhs true?
sub eval_op {
  my ($op,$lhs,$rhs) = @_;
  die unless defined $op;
  return 0 unless defined $lhs;
  return 0 unless defined $rhs;
  if ($op eq '=') {
    return 0 if $lhs ne $rhs;
  } elsif ($op eq '!=') {
    return 0 if $lhs eq $rhs;
  } elsif ($op eq '=~') {
    return 0 unless $lhs =~ m/$rhs/;
  } elsif ($op eq '!~') {
    return 0 unless $lhs !~ m/$rhs/;
  } elsif ($op eq '<') {
    die unless number $lhs;
    die unless number $rhs;
    return 0 unless $lhs < $rhs;
  } elsif ($op eq '>') {
    die unless number $lhs;
    die unless number $rhs;
    return 0 unless $lhs > $rhs;
  } elsif ($op eq '<=') {
    die unless number $lhs;
    die unless number $rhs;
    return 0 unless $lhs <= $rhs;
  } elsif ($op eq '>=') {
    die unless number $lhs;
    die unless number $rhs;
    return 0 unless $lhs >= $rhs;
  } elsif ($op eq '.') {
    die unless number $lhs;
    die unless number $rhs;
    return 0 unless $lhs+1 == $rhs;
  } else {
    print
      "WARNING: operator `$op' can be parsed but not interpreted. Weird.\n";
    return 0;
  }
  return 1;
}

# Look up a feature in the lexicon.
#
# This is much harder to do, and much more expensive, than simply
# looking up the feature in the annotation. Consider this scenario:
#
# der := [cat:ART, def:yes, case:n, gen:m, num:s];
# der := [cat:ART, def:yes, case:g, gen:f, num:s];
# der := [cat:ART, def:yes, case:d, gen:f, num:s];
# der := [cat:ART, def:yes, case:g, gen:bot, num:p];
# der := [cat:ART, def:yes, case:d, gen:bot, num:p];
# der := [cat:PREL, case:n, gen:m, num:s];
# der := [cat:PREL, case:g, gen:f, num:s];
# der := [cat:PREL, case:d, gen:f, num:s];
#
# Now imagine that a sentence with `der' in it is annotated. The
# annotation might say something like this:
#
#  0 1 der
# 'case' / 'n'
# 'cat' / 'ART'
#
# since that is enough to identify the word uniquely. But to
# reconstruct the fact that the feature `def' is `yes' we have to look
# into the lexicon, determine which item the annotation describes, and
# then lookup its value for `def'. That is complicated, but luckily
# the code already exists, and we only have to reuse it.
#
# We do this by recreating the AnnotationStruct that we read from disk
# and handing that to anno2word(). It would be faster to store the
# pointer to the AnnotationStruct originally read, but that would make
# the perl annotation dependent on the existence of a C structure that
# might have been unloaded in the mean time. So we play it safe and
# retain a LexicalEntry that is all our own.
sub lex_lookup {
  my $w = shift;
  my $feature = shift;

  # try very hard to get a pointer to the actual LexicalEntry
  # and cache it for future queries
  if(!exists($$w{le})) {
    my $an = newAnnotation(0,0,$$w{word});
    for my $a (keys %$w) {
      next if $a =~ /^(word|no)$/;
      my $s =
        newSpecification($STTag, $a, $$w{$a}, 0);
      inputSpecificationAdd $an, $s;
    }
    $$w{le} = anno2le $an;
  }

  # query it for this feature
  return inputFeatureOfLI($$w{le}, $feature);
}


# namespace for common variables
# of the parsing and interpreting routines
{

  my %vars;          # variables mentioned in the query and their attributes
  my %conns;         # connections between the variables
  my %features;      # unary equations about features
  my %equations;     # binary equations about features
  my %distances;     # binary equations about distance

  my @commands;                    # commands specified in the query
  # This is an array because commands have side-effects,
  # and therefore order is important.

  my @statistics;                  # statistics ordered by the query

  # holds edges fitting all unary conditions for each variable
  my %candidates;

  # access feature of a bound variable
  # e.g. eval_path($e,'^','case') ==> 'acc'
  # e.g. eval_path($e,'@','label') ==> 'SUBJ'
  sub eval_path {
    my ($edge,$end,$att) = @_;

    if($att eq 'label') {
      return $$edge{label};
    }

    my $word = $edge->{$end};
    my $result = $word->{$att};
    return $result;
  }

  # print the words in an annotation
  sub print_sentence {
    my $annotation = shift;
    print "\n";
    my @words = @{$$annotation{words}};
    # do not print word #0, which is `NIL'
    for (@words[1..$#words]) {
      print "$_->{word} ";
    }
    print "\n";
  }

  # print all solution sets on STDOUT
  sub print_sets {
    my $x = shift;
    my $writeback = 0;

    for my $sol (@$x) {
      my $annotation;
      my $first = 1;

      # print the edges that constitute the set
      for my $v (sort keys %$sol) {
        next if $v eq 'contains';
        if ($first) {
          $annotation = $sol->{$v}->{anno};
          if($options{verbosity}) {
            print "\n\n"
          } else {
            print " ";
          }
          print "$annotation->{id}:";
        }
        if ($options{verbosity}) {
          print "\n$v: ";
          print_edge $$sol{$v};
        } else {
          print "," unless $first;
          print "$$sol{$v}->{pos}";
        }
        $first = 0;
      }
      if(!$options{verbosity}) {
        print " ";
      }

      # maybe print the sentence itself
      if ($options{verbosity} > 1) {
        print_sentence $annotation;
      }

      # execute the commands
      for my $c (@commands) {
        $writeback = 1;

        # `X.label := FOO'
        if ($$c{type} eq 'label') {
          my $e = $$sol{$$c{var}};
          $$e{label} = $$c{label};
        }

        # `X.label := Y.label'
        elsif ($$c{type} eq 'labelexchange') {
          my $e = $$sol{$$c{dst}};
          my $e2 = $$sol{$$c{src}};
          $$e{label} = $$e2{label};
        }

        # `X@cat := NN'
        elsif ($$c{type} eq 'feature') {
          my $e = $$sol{$$c{var}};
          my $w = $$c{end} eq '^' ? modifiee $e : modifier $e;
          $$w{$$c{att}} = $$c{val};
        }

        # `X@case := X^case'
        elsif ($$c{type} eq 'relative_feature') {
          my $e1 = $$sol{$$c{var1}};
          my $e2 = $$sol{$$c{var2}};
          my $w1 = $$c{end1} eq '^' ? modifiee $e1 : modifier $e1;
          my $w2 = $$c{end2} eq '^' ? modifiee $e2 : modifier $e2;
          if (exists $$w2{$$c{att2}}) {
            $$w1{$$c{att1}} = $$w2{$$c{att2}};
          }
        }

        # `delete(X@case)'
        elsif ($$c{type} eq 'delete_feature') {
          my $e = $$sol{$$c{var}};
          my $w = $$c{end} eq '^' ? modifiee $e : modifier $e;
          if (exists $$w{$$c{att}}) {
            delete $$w{$$c{att}};
          }
        }

        # redirect
        elsif ($$c{type} eq 'redirect') {
          my $lhs = $$sol{$$c{lhs}};

          # `X := NIL'
          if ($$c{rhs} eq 'NIL') {
            $$lhs{parent} = 0;
            $$lhs{'^'} = $a->{words}[0];
          }

          # `X := Y@':
          # redirect X to Y's modifier
          elsif ($$c{rhs} =~ /($var_re)@/) {
            my $name = $1;
            my $rhs = $$sol{$name};
            $$lhs{parent} = $$rhs{child};
            $lhs->{'^'} = $rhs->{'@'};
          }

          # `X := Y':
          # redirect X to Y's modifiee
          else {
            my $rhs = $$sol{$$c{rhs}};
            $$lhs{parent} = $$rhs{parent};
            $lhs->{'^'} = $rhs->{'^'};
          }
        }

      }
      if ($writeback) {
        write_anno $annotation;
      }
    }
  }

  # count what the user wanted counted
  sub count_sets {
    my $x = shift;

    for my $sol (@$x) {
      for my $s (@statistics) {
        my $edge = $$sol{$$s{var}};
        if(exists($$s{end})) {
          my $word = $$edge{$$s{end}};
          my $att = $$word{$$s{att}} || '<undef>';
          my $h = $$s{counts};
          $$h{$att}++;
        } elsif ($$s{att} eq 'label') {
          my $h = $$s{counts};
          $$h{$$edge{label}}++;
        } elsif ($$s{att} eq 'length') {
          my $h = $$s{counts};
          my $value = edge_length($edge);
          $$h{"$value"}++;
        }
      }
    }
  }

  # select sets of edges from the candidates
  # that fulfill all specifications.
  sub find_sets {
    my $unbound = shift;
    my $sets = shift;
    debug "find_sets() called\n";
    debug "  # of variables left: " . (1+$#$unbound) . "\n";
    debug "  # of partial solutions: " . (1+$#$sets) . "\n";
    return $sets unless scalar @$unbound;

    my @new_sets = ();

    # take first unbound variable
    my $var = pop @$unbound;
    my $new = $$var{name};
    debug " binding var `$new'\n";

    my $i;
    # check it against all partial sets
  SET:
    for my $s (@$sets) {
      debug "  extending set: [ "; print_sets [ $s ] if $options{debug}; debug "  ]\n";
    EDGE:
      for my $e (@{$candidates{$new}}) {
        debug "  checking edge "; print_edge $e if $options{debug}; debug "...";
        last SET if $interrupted;
        my $new_s = { %$s };
        # make deep copy of the `contains' hash so it does not
        # accumulate entries across different alternatives
        $new_s->{contains} = { %{$s->{contains}} };


        # suppress edges already present
        if (exists $new_s->{contains}{$$e{no}}) {
          debug "   already present!\n";
          next EDGE;
        }

        # check connexions
        for my $v (keys %$s) {
          if (exists $conns{"$v:$new"}) {
            if (!match_conn($conns{"$v:$new"}, $$s{$v}, $e)) {
              debug "   wrong connection!\n";
              next EDGE;
            }
          }
          if (exists $conns{"$new:$v"}) {
            if (!match_conn($conns{"$new:$v"}, $e, $$s{$v})) {
              debug "   wrong connection!\n";
              next EDGE;
            }
          }
        }

        # check binary equations
        for my $old (keys %$s) {
          for my $eq ( @{$equations{"$old$new"}}) {
            my $lhs = eval_path($$s{$old}, $$eq{end1}, $$eq{att1});
            my $rhs = eval_path($e, $$eq{end2}, $$eq{att2});
            if (!eval_op($$eq{op}, $lhs, $rhs)) {
              debug "   failed equation!\n";
              next EDGE;
            }
          }
          for my $eq ( @{$equations{"$new$old"}}) {
            my $lhs = eval_path($e, $$eq{end1}, $$eq{att1});
            my $rhs = eval_path($$s{$old}, $$eq{end2}, $$eq{att2});
            if (!eval_op($$eq{op}, $lhs, $rhs)) {
              debug "   failed equation!\n";
              next EDGE;
            }
          }
        }

        # check binary distance
        for my $old (keys %$s) {
          my $e2 = $$s{$old};
          for my $eq ( @{$distances{"$old$new"}}) {
            my $lhs = $$e{$$eq{end1}}{no} - $$e2{$$eq{end2}}{no};
            my $rhs = $$eq{d};
            if (!eval_op($$eq{op}, $lhs, $rhs)) {
              debug "   failed distance criterion!\n";
              next EDGE;
            }
          }
          for my $eq ( @{$distances{"$new$old"}}) {
            my $e2 = $$s{$old};
            my $lhs = $$e2{$$eq{end2}}{no} - $$e{$$eq{end1}}{no};
            my $rhs = $$eq{d};
            if (!eval_op($$eq{op}, $lhs, $rhs)) {
              debug "   failed distance criterion!\n";
              next EDGE;
            }
          }
        }

        # add edge
        debug "  OK.\n";
        $new_s->{$var->{name}} = $e;
        $new_s->{contains}{$$e{no}} = 1;
        push @new_sets, $new_s;
      }
    }

    # call self
    debug "  passing " . (1+$#new_sets) . " partial solutions.\n";
    return find_sets($unbound,[@new_sets]);
  }

  # return variable by name if it exists
  sub query_var {
    my $name = shift;
    if (exists $vars{$name}) {
      return $vars{$name};
    }
    print "No such variable: $name\n";
    return 0;
  }

  # return variable by name, creating if necessary
  sub get_var {
    my $name = shift;
    unless (exists $vars{$name}) {
      $vars{$name} = { name => $name };
      debug "  variable `$name' declared implicitly.\n";
    }
    return $vars{$name};
  }

  # analyse transformation commands into @commands
  sub parse_commands {
    for (split /&/, shift) {

      # label change
      if (/^ *($var_re) *. *label *:= *($string_re|$number_re) *$/) {
        my($var,$lab) = ($1,parse_string $2);
        return 0 unless query_var $var;
        push @commands, { type=>'label', var=>$var, label=>$lab };
      }

      # label exchange
      elsif (/^ *($var_re) *. *label *:= *($var_re) *. *label *$/) {
        my($dst,$src) = ($1,$2);
        return 0 unless query_var $dst;
        return 0 unless query_var $src;
        push @commands, { type=>'labelexchange', dst=>$dst, src=>$src };
      }

      # feature change
      elsif (/^ *($var_re) *([@^]) *($att_re) *:= *($string_re|$number_re) *$/) {
        my($v,$end,$att,$val) = ($1,$2,$3,$4);
        return 0 unless query_var $v;
        push @commands,
          { type=>'feature', var=>$v, end=>$end, att=>$att, val=>$val };
      }

      # relative feature change
      elsif (/^ *($var_re) *([@^]) *($att_re) *:= *($var_re) *([@^]) *($att_re) *$/) {
        my($v1,$end1,$att1,$v2,$end2,$att2) = ($1,$2,$3,$4,$5,$6);
        return 0 unless query_var $v1;
        return 0 unless query_var $v2;
        push @commands,
          {
           type=>'relative_feature',
           var1=>$v1, end1=>$end1, att1=>$att1,
           var2=>$v2, end2=>$end2, att2=>$att2
          };
      }

      # feature deletion
      elsif (/^ *delete *\( *($var_re) *([@^]) *($att_re) *\) *$/) {
        my($v,$end,$att) = ($1,$2,$3);
        return 0 unless query_var $v;
        push @commands,
          { type=>'delete_feature', var=>$v, end=>$end, att=>$att };
      }

      # structure change
      elsif (/^ *($var_re) *:= *(($var_re)@?|NIL) *$/) {
        my($v1,$v2) = ($1,$2);
        return 0 unless query_var $v1;
        (my $name2 = $v2) =~ s/@//;
        return 0 unless ($name2 eq 'NIL' or query_var $name2);
        push @commands, { type=>'redirect', lhs=>$v1, rhs=>$v2 };
      } else {
        print "Invalid command: `$_'\n";
        return 0;
      }
    }
    return 1;
  }

  # analyse query into %vars, %conns, %features and %equations
  sub parse_formula {
    debug "parsing: $_[0]\n";

    for (split /&/, shift) {
      debug "parsing spec `$_'\n";
      # variable declaration
      if (/^ *($var_re) *($dir_re) *([A-Z0-9]+) *$/ &&
          exists($levels{$3})) {
        my($name,$dir,$level) = ($1,$2,$3);
        debug "  variable $name declared\n";
        $vars{$name} = { name => $name,
                         dir => $dir,
                         level => { op =>'=', rhs=>$level }};
        next;
      }

      # connexion
      elsif (m!^ *($var_re) *($conn_re) *($var_re) *$!) {
        my($v1,$conn,$v2) = ($1,$2,$3);
        get_var $v1;
        get_var $v2;
        debug "  connection $v1$conn$v2 declared\n";
        $conns{"$v1:$v2"} = $conn;
      }

      # binary label comparison
      elsif (/^ *($var_re) *\. *label *($op_re) *($var_re) *\. *label *$/) {
        my($var1,$op,$var2) = ($1,$2,$3);
        my $v1 = get_var $var1;
        my $v2 = get_var $var2;
          push @{$equations{"$var1$var2"}},
            { end1 => '@', att1 => 'label',
              end2 => '@', att2 => 'label',
              op => $op };
        debug "  $var1.label $op $var2.label declared\n";
      }

      # label
      elsif (/^ *($var_re) *\. *label *($op_re) *([^ ]+) *$/) {
        my($var,$op,$lab) = ($1,$2,parse_string($3));
        my $v = get_var $var;
        push @{$v->{label}}, { op => $op, rhs => $lab };
        debug "  $var.label $op `$lab' declared\n";
      }

      # length
      elsif (/^ *($var_re) *\. *length *($op_re) *([-0-9]+) *$/) {
        my($var,$op,$l) = ($1,$2,$3);
        my $v = get_var $var;
        push @{$v->{length}}, { op => $op, rhs => $l };
        debug "  $var.length $op `$l' declared\n";
      }

      # `has' condition
      elsif (/^ *(~?) *has *\( *($var_re) *, *([^ ]+) *\) *$/) {
        my ($neg,$var,$label) = ($1,$2,$3);
        my $v = get_var $var;
        if(!$v) {
          print " Undeclared variable: `$var'\n";
          return 0;
        }
        push @{$v->{has}}, { neg=>$neg, label=>$label };
      }

      # distance test
      elsif (/^\s*distance\s*\(\s*($var_re)\s*([@^])\s*,\s*($var_re)\s*([@^])\s*\)\s*($op_re)\s*(-?[0-9]+)\s*$/) {

        my ($var1,$end1,$var2,$end2,$op,$d) = ($1,$2,$3,$4,$5,$6);
        my $v1 = get_var $var1;
        if(!$v1) {
          print " Undeclared variable: `$var1'\n";
          return 0;
        }
        my $v2 = get_var $var2;
        if(!$v2) {
          print " Undeclared variable: `$var2'\n";
          return 0;
        }
        push @{$distances{"$var1$var2"}},
          { end1 => $end1, end2 => $end2, op => $op, d => $d };
      }



      # existence test
      elsif (/^ *(~?) *exists *\( *($var_re) *([@^]) *($att_re) *\) */) {
        my($neg,$var,$end,$att) = ($1,$2,$3,parse_string($4));
        if (!get_var $var) {
          print " Undeclared variable: `$var'\n";
          return 0;
        }
        push @{$features{$var}},
            { end => $end,
              att => $att,
              op => $neg eq '' ? '!=' : '=',
              val => '<undef>' };
        debug "  exists($var$end$att) declared\n";
      }

      # lexical lookup
      elsif (/^ *lookup *\( *($var_re) *([@^]) *($att_re) *\) *($op_re) *($string_re) *$/) {
        my($var,$end,$att,$op,$val) = ($1,$2,parse_string($3),$4,parse_string($5));
        if (!get_var $var) {
          print " Undeclared variable: `$var'\n";
          return 0;
        }
        push @{$features{$var}},
          { end => $end,
            lookup => $att,
            op => $op,
            val => $val };
        debug "  lookup($var${end}word,$att) = $val declared\n";
      }

      # lexical features
      elsif (/^ *($var_re) *([@^]) *($att_re) *($op_re) *([^ ]+) *$/) {
        my($var1,$end1,$att1,$op,$val) =
          ($1,$2,parse_string($3),$4,parse_string($5));

        # binary equation: `X@no < Y@no'
        if ($val =~ /^($var_re) *([@^]) *($att_re) *$/) {
          my ($var2,$end2,$att2) = ($1,$2,$3);
          if (!get_var $var2) {
            print " Undeclared variable: `$var2'\n";
            return 0;
          }
          push @{$equations{"$var1$var2"}},
            { end1 => $end1, att1 => $att1,
              end2 => $end2, att2 => $att2,
              op => $op };
          debug "  $var1$end1$att1 $op $var2$end2$att2 declared\n";
        }

        # ordinary feature test: `X@no < 3'
        else {
          if (!get_var $var1) {
            print " Undeclared variable: `$var1'\n";
            return 0;
          }
          push @{$features{$var1}},
            { end => $end1, att => $att1, op => $op, val => $val };
          debug "  feature $var1$end1$att1 $op $val declared\n";
        }
      } elsif (/^ *($var_re) *($dir_re) *([A-Z0-9]+) *$/ && !exists($levels{$3})) {
        printif 1, "\nHINT: The level $3 is never mentioned in your annotations,\n";
        printif 1, "HINT: so your query cannot succeed.\n";
      }

      # statistics about features
      elsif(/^ *stat *\( *($var_re) *([@^]) *($att_re) *\) *$/) {
        my($var,$end,$att) = ($1,$2,$3);
        if (!get_var $var) {
          print " Undeclared variable: `$var'\n";
          return 0;
        }
        push @statistics,
          { var => $var, end => $end, att => $att, counts => {} };
      }

      # statistics about edges
      elsif(/^ *stat *\( *($var_re) *\. *(label|length) *\)$/) {
        my ($var,$att) = ($1,$2);
        if (!get_var $var) {
          print " Undeclared variable: `$var'\n";
          return 0;
        }
        push @statistics,
          { var => $var, att => $att, counts => {} };
      }

      # catchall
      else {
        print "Invalid specification: `$_'\n";
        return 0;
      }
    }
    return 1;
  }

  # parse entire user query into scope-local variables
  sub parse_query {

    %vars = ();
    %conns = ();
    %features = ();
    %equations = ();
    %distances = ();
    @commands = ();
    @statistics = ();

    my $line = shift;
    if ($line =~ /^(.*)->(.*)$/) {
      my($f,$c) = ($1,$2);
      return undef unless parse_formula $f;
      return undef unless parse_commands $c;
    } else {
      return undef unless parse_formula $line;
    }
    return 1;
  }

  # apply the query to annotation $a
  sub process {
    my $annotation = shift;
    my $count = 0;
    %candidates = ();

    # find all possible variable bindings
  VAR:
    for my $v (values %vars) {
      my $name = $$v{name};
      debug "searching candidates for $name...\n";

    EDGE:
      for my $e (@{$annotation->{edges}}) {
        last VAR if $interrupted;
        debug "  checking edge "; print_edge $e if $options{debug}; debug "...\n";
      SPEC:
        for my $spec (keys %$v) {
          next SPEC if $spec eq 'name';

          my $val = $v->{$spec};

          # check direction
          if ($spec eq 'dir') {
            debug "    checking dir $val ...\n";
            next EDGE unless match_dir($e,$val);
          }

          # check label
          elsif ($spec eq 'label') {
            for my $l (@{$$v{label}}) {
              my($rhs,$op) = ($$l{rhs},$$l{op});
              debug "    checking label $op $rhs...\n";
              next EDGE unless eval_op($op, $e->{label}, $rhs)
            }
          }

          # check length
          elsif ($spec eq 'length') {
            for my $l (@{$$v{length}}) {
              my($rhs,$op) = ($$l{rhs},$$l{op});
              debug "    checking length $op $rhs...\n";
              next EDGE unless eval_op($op, edge_length($e), $rhs)
            }
          }

          # check `has'
          elsif ($spec eq 'has') {
            for my $h (@{$$v{has}}) {
              debug "    checking `$$h{neg}has($$h{label})...\n";
              if($$h{neg} eq '') {
                next EDGE unless has($e,$$h{label});
              } else {
                next EDGE if has($e,$$h{label});
              }
            }
          }

          # check features
          else {
            my($rhs,$op) = ($$val{rhs},$$val{op});
            debug "    checking feature $spec $op $rhs...\n";
            next EDGE unless exists $e->{$spec};
            next EDGE unless eval_op($op, $e->{$spec}, $rhs)
          }
        }

        # check binary equations that are really unary: `X@cat = X^cat'
        for my $eq ( @{$equations{"$name$name"}}) {
          debug "    checking feature $name$$eq{end1}$$eq{att1} $$eq{op} $name$$eq{end2}$$eq{att2}...\n";
          my $lhs = eval_path($e, $$eq{end1}, $$eq{att1});
          my $rhs = eval_path($e, $$eq{end2}, $$eq{att2});
          if (!eval_op($$eq{op}, $lhs, $rhs)) {
            debug "   failed equation!\n";
            next EDGE;
          }
        }

      EQ:
        for my $eq (@{$features{$name}}) {

          # check lexical lookup
          if(exists($$eq{lookup})) {
            debug "    checking lookup($$eq{end}$$eq{lookup}) = `$$eq{val}'...\n";
            next EDGE unless exists $e->{$$eq{end}};
            my $word = $e->{$$eq{end}};
            my $lhs = lex_lookup($word, $$eq{lookup});
            next EDGE unless defined $lhs;
            debug "    lookup returns `$lhs'\n";
            next EDGE unless eval_op($$eq{op}, $lhs, $$eq{val});
          } else {

            # check feature
            debug "    checking feature $$eq{end}$$eq{att} $$eq{op} `$$eq{val}'...\n";
            my $op = $$eq{op};
            next EDGE unless exists $e->{$$eq{end}};
            my $word = $e->{$$eq{end}};
            my $lhs = $word->{$$eq{att}} || '<undef>';
            debug "    value is `$lhs'\n";
            next EDGE unless eval_op ($$eq{op}, $lhs, $$eq{val});
          }
        }


        # check unary distance
        for my $d (@{$distances{"$name$name"}}) {
          my $lhs = $$e{$$d{end1}}{no} - $$e{$$d{end2}}{no};
          my $rhs = $$d{d};
          if (!eval_op($$d{op}, $lhs, $rhs)) {
            debug "   failed distance criterion!\n";
            next EDGE;
          }
        }

        # OK, this edge may fit
        debug "candidate found.\n";
        push @{$candidates{$v->{name}}}, $e;
      }

      # fail immediately due to lack of candidates
      if (! exists $candidates{$v->{name}}) {
        debug "No candidates for $name.\n";
        return 0;
      }
      debug "Found " . scalar @{$candidates{$name}} . " candidates for $name.\n";
    }

    # find actual solution sets
    my $result = find_sets([ values %vars ], [{ contains=>{}}]);

    # boast with it
    if(@statistics) {
      count_sets($result);
    } else {
      print_sets($result);
    }

    return scalar @$result;
  }

  # handle one line of user input
  sub interpret {
    my $count = 0;
    my $line = shift;

    print "Interpreting: $line" if $opt_p;

    chomp $line;

    return unless $line =~ /\S/;

    # interpret `s154'
    if (exists $annotations{$line}) {
      print_anno $line;
      return;
    }

    # interpret control commands
    if ($line =~ /^ *set +([^ ]+) +([^ ]+)/) {
      if (exists $options{$1}) {
        $options{$1} = $2;
      } else {
        print "No option `$1'.";
      }
      return;
    }

    # execute normal CDG command
    elsif($line =~ /^ *([a-z0-9-]+)\b/ &&
          exists($cdg_commands{$1})) {
      my $command = $1;
      commandEval $line;
      %annotations = import_annos() if 'load' eq $command;
    }

    # execute search command
    elsif (parse_query $line) {
      my $processed;
      for (sort byVName keys %annotations) {
        $processed++;
        $count += process $annotations{$_};
        if ($interrupted) {
          last;
        }
      }

      if(!$options{verbosity}) {
        print "\n";
      }

      printif 1, "\n\n# of matching structures: $count\n";
      if ($interrupted) {
        my $x = int($count * (scalar keys %annotations) / $processed);
        printif 1, "I probably would have found $x matches if you had let me.\n";
      }
      $interrupted = 0;

      # print result of statistics collected
      for my $s(@statistics) {
        my %h = %{$$s{counts}};
        if(exists($$s{end})) {
          printf "Values of $$s{var}$$s{end}$$s{att} (%d different):\n", scalar keys %h;
          for my $k(sort {$h{$b}<=>$h{$a}} keys %h) {
            print " $k $h{$k} ";
          }
        } elsif($$s{att} eq 'label') {
          print "Values of $$s{var}.label: ";
          for my $k(sort {$h{$b}<=>$h{$a}} keys %h) {
            print " $k $h{$k} ";
          }
        } elsif($$s{att} eq 'length') {
          print "Length of $$s{var}:\n\n";
          for my $k(sort {$a<=>$b} keys %h) {
            print " $k: $h{$k}\n";
          }
        }
        print "\n";
      }
    }
  }

}


### MAIN ###

# slurp input
commandEval "load @ARGV";

%annotations = import_annos();

# catch C-c
$SIG{INT} = sub {
  $interrupted++;
};

# loop
if ($opt_p) {
  while (<>) {
    interpret $_;
  }
} else {

  # read user's command history
  my $historyfile = "$ENV{HOME}/.cdgsearchrc";
  if (-r $historyfile) {
    for (`cat $historyfile`) {
      chomp;
      $term->addhistory($_) if /[^ ]/;
    }
  }

  # add to command history
  while ( defined ($_ = $term->readline("Find edge: >"))) {
    interpret $_;
  }

  # write updated command history
  if (open(SAVEHIST,">$historyfile")) {
    # TODO: remove duplicates, impose maximum size
    print SAVEHIST (join "\n",$term->history_list());
  }

}


