#!/usr/bin/env perl
# cdg/utils/tokenize.pl

use warnings;
use strict;

use Getopt::Std;
getopts("p:w:");
our $opt_p;
our $opt_w;
our $prefix = $opt_p ? "$opt_p-" : '';

# Some words, usually proper names, have whitespace within them, but
# should be treated as one word anyway. Usually these sequences that
# are clearly recognisable, since otherwise they would be very
# misleading even for the human reader. Therefore we allow the user to
# specify a file of such sequences, and each occurrence of them will
# be reassembled upon sight.
our @whitespace_tokens = ();
if($opt_w && -r $opt_w) {
  open(IN, "$opt_w") or die "Can't read `$opt_w': $!\n";
  while(<IN>) {
    (my $new = $_) =~ s/ /INTERNALWHITESPACE/g;
    push @whitespace_tokens, { from => $_, to => $new };
  }
}
close(IN);

our $abbrevs = join '|',
  qw (

      Abs Az BGBl Corp Dr Inc Mio Mr Mrd Nr Prof St Tel [A-Z] a bzw ca
      etc evtl excl exkl i incl inkl lt q u usf usw vgl z [123]?[0-9][a-k]?

     );

our $keywords = 'true|false|level|label';

our $counter = 0;
our $marker = '##EOS##';

our @files = ();

our $lastsentence = '';
sub write_lattice {

  my @x = @_;

  return unless @x;

  # Yet another abbreviation heuristic: Although a sentence may end
  # with an abbrevation, it is very unlikely that it is an ordinal
  # abbrevation. Therefore `3.' is split if it is the last word.
  if($x[-1] =~ /^(\d+)\.$/ && $#x > 1) {
    pop @x;
    push @x, $1;
    push @x, '.';
  }

  # detect sentences that consist only in a quotation mark; this is a
  # sure sign that the computation of quotation levels has been thrown
  # off, probably because of inconsistent quotation marks in the input.
  if(1 == @x && $x[0] =~ /^[\'\"]$/) {
    warn "Possible incorrect quotation mark after `$lastsentence'!\n";
  }

  # insert quotes only where necessary
  my @quoted;
  for (@x) {

    s/INTERNALELLIPSIS/.../g;
    s/INTERNALWHITESPACE/ /g;

    if (/[^a-zA-Z]|^$keywords$/) {

      # escape internal quotes
      s/([\'\"\\])/\\$1/g;

      # add external quotes
      $_ = "'$_'";
    }
    push @quoted, $_;
  }

  $counter++;
  print "'${prefix}s$counter' : ", join(', ', @quoted);
  print ";\n\n";
  $lastsentence = "@x";
}

# build list of files to tokenize
for my $what (@ARGV) {

  if ( -d $what ) {

    # read each specified file
    opendir(DIR, $what);
    for my $file (readdir(DIR)) {
      push @files, "$what/$file";
    }
    closedir(DIR);
  } elsif ( -f $what ) {
    push @files, $what;
  }
}

# process list
for my $file(@files) {

  my @words = ();
  open(IN, $file)
    or die "Can't read `$file': $!\n";

  # Read lines of input and apply some initial transformations. Note
  # that line breaks can intervene between any two words, so most
  # transformations must be done later. At this stage we can only
  # apply those transformations that insert space, or those that apply
  # to patterns that cannot possibly be broken across lines.
  while (<IN>) {
    chomp;
    s/^\s+//g;
    s/\s+$//g;

    # Bad Nauheim --> 'Bad Nauheim'
    for my $w (@whitespace_tokens) {
      s/$$w{from}/$$w{to}/g;
    }

    # notice word-final apostrophe, those might be genitive markers
    s/(ch|[xcysz])\'(\W|\z)/$1 APOSTROPHEORGEN $2/g;

    # notice sentence-internal ...
    s/([a-zA-Z])\.\.\.([a-zA-Z])/$1 INTERNALELLIPSIS $2/g;

    # reassemble consecutive full stops
    s/\.(\s*\.)+/ ... /g;

    # ja---aber --> ja --- aber
    s/--+/ - /g;

    # insert space between words and punctuation:
    # (Wohin steuern die USA?). --> ( Wohin steuern die USA ? ) .

    # Vorwrts! --> Vorwrts !
    s/(.*?)([.,;:!?]|\.\.+|!!+|\?\?+)(\s+|\z|[)\]"`'])/$1 $2 $3/g;

    # Isolate quotation marks from the words that they mark:
    # "put options" --> " put options "
    1 while s/(\s|^)([`'"(])(\S)/ $2 $3/g;
    s/([`'")])(\s|$)/ $1 /g;

    # But some German words legitimately start with apostrophes,
    # such as 'n (ein) or 's (es), so we reassemble those.
    s/\' +(n|s|ne|nem)\b/\'$1/g;

    # An entirely empty line is probably a paragraph break,
    # so we insert a marker for the sentence delimiter.
    s/^\s*$/$marker/g;

    my @line = split /\s+/;
    for(@line) {
      push @words, $_ unless /^\s*$/;
    }
  }
  close(IN);

  # Now apply transformations that need context to either side. This
  # stage transforms the token stream so that each item corresponds to
  # exactly one word in the final lattice. We also decide where the
  # sentences boundaries will be and insert the special marker $marker
  # for the next stage.
  my @tokens = ();
  my $num_re = qr '^[-+]?[0-9.]+$';
  my $abbrev_re = qr "^($abbrevs)$";

  my $single_quotes = 0;
  my $double_quotes = 0;

 WORD:
  for(my $i = 0; $i <= $#words; $i++) {
    my $this = $words[$i];
    my $last = $i ? $words[$i-1] : '';
    my $next = $i<$#words ? $words[$i+1] : '';

    # find next proper, i.e. non-punctuation word
    my $j = $i;
    $j++ while $j <= $#words && $words[$j] !~ /[a-zA-Z]/;
    my $next_proper_word = $j <= $#words ? $words[$j] : '';

    # When more than one bit of punctuation occurs at the end of a
    # sentence, add all following bits in front of the EOS marker.
    sub push_punctuation {
      while($i < $#words) {

        # add " to the previous sentence if we are within quoted
        # material, otherwise to the next one.
        if($words[$i+1] eq '"' && $double_quotes == 1) {
          push @tokens, '"';
          $i++;
          $double_quotes = 0;
        }
        elsif($words[$i+1] eq '\'' && $single_quotes == 1) {
          push @tokens, '\'';
          $i++;
          $single_quotes = 0;
        }
        elsif($words[$i+1] =~ /^[.)]$/) {
          push @tokens, $words[$i+1];
          $i++;
        } else {
          last;
        }
      }
    }

    if($this =~ /^[^\"]*\"[^\"]*$/) {
      $double_quotes = 1 - $double_quotes;
    }
    elsif($this eq "'") {
      $single_quotes = 1 - $single_quotes;
    }
    elsif($this eq "`") {
      $single_quotes = 1 - $single_quotes;
    }

    if ($this eq '.') {

      # reassemble mis-tokenized numbers:
      # '1', '.', '000', '.', '000' --> '1.000.000'
      if($last =~ $num_re && $next =~ /^[0-9]{3}\b/) {
        my $prev = pop @tokens;
        push @tokens, "$prev.$next";
        $i++;
        next WORD;
      }

      # combine abbreviation and stop: 'Dr', '.' --> 'Dr.'
      elsif($last =~ $abbrev_re) {
        my $prev = pop @tokens;
        push @tokens, "$prev.";
        next WORD;
      }
      # FIXME: This can fail if a sentence ends with an abbreviation,
      # since the sentence full stop is suppressed in that case. Also,
      # a sentence can legitimately end in a number: `...auf Platz 9.'
      # Currently we err on the side of caution and do not end the
      # sentence.

      # ignore full stop when the next non-punctuation word is
      # lowercase (these are abbreviations in most cases)
      if($next_proper_word =~ /^[a-z]/) {
        push @tokens, '.';
        next WORD;
      }
      push @tokens, '.';
      push_punctuation;
      push @tokens, $marker;
    }

    # try to distinguish between ' used as quotation and ' used as
    # genitive morpheme. If the ' followed a sibilant without
    # whitespace, and the level of quotation is 0, then we assume
    # that it is a genitive morpheme and recombine it with the
    # previous word.
    elsif($this eq 'APOSTROPHEORGEN') {
      if($single_quotes) {
        $single_quotes = 0;
        push @tokens, "'";
        next WORD;
      } else {
        my $w = pop @tokens;
        push @tokens, "$w'";
        next WORD;
      }
    }

    # reassemble compound names:
    # `Nordrhein- Westfalen' --> `Nordrhein-Westfalen'
    elsif($last =~ /.-$/ && $this =~ /^[A-Z]/) {
      my $prev = pop @tokens;
      push @tokens, "$prev$this";
      next;
    }

    # reassemble compounds with quotes in them
    #
    # Formations like `XML-"Sprache"' are often mistokenized because
    # the lateral " is taken to be a single token. We reassemble them
    # here if a word with -" in them is adjacent to a single ".
    elsif($this =~ /\"-/ && $last eq '"') {
      pop @tokens;
      push @tokens, "\"$this";
    }
    elsif($this eq '"' && $last =~ /-\"/) {
      pop @tokens;
      push @tokens, "$last\"";
    }

    # Split off clitic 's:
    # 'gibt's' --> 'gibt' ''s'
    elsif($this =~ /^(Weil|[a-z].+)('[sn])$/i) {
      push @tokens, $1;
      push @tokens, $2;
    }
    # Of course there are many more words that can carry the clitic s;
    # we need a method here of distinguishing `Barron's' from `soll's'.

    # Generate end of sentence marker.
    else {
      push @tokens, $this;
      if($this =~ /^([?!:;]|\.\.\.)$/ &&
         $next_proper_word =~ /^[A-Z]/) {
        push_punctuation;
        push @tokens, $marker;
      }
    }

    # The token `15"-Monitor' does not contain a quotation mark,
    # therefore reset the count.
    if($this =~ /\d\"-/ && $double_quotes) {
      $double_quotes = 0;
    }
  }
  push @tokens, $marker;

  # Now send the delimited bunches of words to the output routine.
  my @sentence = ();
  for(@tokens) {
    if($_ eq $marker) {
      write_lattice @sentence;
      @sentence = ();
    } else {
      push @sentence, $_;
    }
  }
}

