# Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## AllParses - manage parser results.
##
## 
## \author Dietmar Fnning, Kilian A. Foth (see also AUTHORS and THANKS for more)
## $Id: allparses.tcl,v 1.58 2004/04/01 15:33:54 foth Exp $
# ----------------------------------------------------------------------------
class AllParses {
   inherit DataBrowser


   # public methods
   public method getParseForAnno {anno}; ## \type TclString
   public method selectIdsOfNet {netId}; ## \type TclString
   public method selectIdsOfLattice {netId}; ## \type TclString
   public method addParse {parse}; ## \type Parse
   public method addParsesOfNet {netId}; ## \type TclString
   public method showparse {args}; ## \type TclList
   public method deleteparse {args}; ## \type TclList
   public method verifyparse {args}; ## \type TclList
   public method getAllparseIds {};  ## \type TclList
   public method getAllAnnoIds {};   ## \type TclList
   public method handlePartialResult {parse_pointer}; ## \type Parse
   public method handleICinteraction {}
   public method reset {}

   public method getCData {id};## \type TclString
   public method getParseForLattice {id};## \type TclString

   public method refreshrow {row parse}; ## \type TclNumber, Parse

   # protected methods
   protected method init_data {}

   constructor {args} {}; ## \type TclList

   # public variables
   public variable autoshow "valid"

   # Should we display parses of different sentences
   # in separate windows?
   public method setManyWindows {x}
   private variable manyWindows 0
  
   # private methods
   private method treebutton_action {}
   private method deletebutton_action {}
   private method verifybutton_action {}

   # private variables
   private variable allparses; ## \type TclArray

};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body AllParses::constructor {args} {
    set _idColumnIndex 0
    set table $itk_component(table)
    $table configure -cols 8 -colstretchmode last \
 
   
   itk_component add treebutton {
      button $itk_component(childsite).show \
                -anchor w \
                -text "Tree" \
                -command [code $this treebutton_action]
   } { }

   itk_component add verifybutton {
      button $itk_component(childsite).verify \
                -anchor w \
                -text "Verify" \
                -command [code $this verifybutton_action]
   } { }

   itk_component add deletebutton {
      button $itk_component(childsite).delete\
               -anchor w \
               -text "Delete" \
               -command [code $this deletebutton_action]
   } { }

   # packing
   pack $itk_component(deletebutton) \
	$itk_component(treebutton) \
	$itk_component(verifybutton) \
	-side left -padx 3 


   # get the initial data
   init_data
   
   # register the helpmessages
   .cdgmain help sethelpstr \
      $itk_component(deletebutton) "Deletes the selected parse" \
      $itk_component(treebutton) "Graphical display of the selected parse" \
      $itk_component(verifybutton) "Compare the selected parse agains the annotation" \
      $itk_component(table) "Click to select"

   eval itk_initialize $args
}

## ----------------------------------------------------------------------------
## refresh a row with a specified parse.
## ----------------------------------------------------------------------------
body AllParses::refreshrow {row parse} {

  set table $itk_component(table)
  $table setCell row $row,0 [list \
    [$parse getId] \
    [$parse getLatticeId] \
    [format "%4.3e" [$parse getScore]] \
    [listSize [$parse getLevels]] \
    [listSize [$parse getViolations]] \
    [$parse getDate] \
    [$parse getUserName] \
    [$parse getComment]]

}


## ----------------------------------------------------------------------------
## initialize with all available Parses
## ----------------------------------------------------------------------------
body AllParses::init_data {} {

  set table $itk_component(table)
  $table configure -state normal
  $table clear all
  set rmax [$table cget -rows]
  $table clear all
  $table delete rows 1 [$table cget -rows]
  $table configure -rows $rmax
  $table tag config default
  $table tag config colored -bg gray80 
  $table tag config lefties -anchor w
  $table tag col lefties 12
  $table tag col lefties 6
  $table width 0 10 1 11 2 11 3 7 4 9 5 22 6 20 7 20

  $table setCell row 0,0 {
       "" "Lattice" "Score" "Levels" "Violations" 
       "Date" "User" "Comment" }
               
  # check all itcl objects of class `Parse'
  # check whether the underlying C parse exists.
  # if needed remove the Parse and any attached ParseTrees 
  array unset allparses
  foreach parse [itcl::find objects -isa Parse] {
    
    set id [$parse getId]
    if {[parseFind $id] != "NULL"} {
      set allparses($id) $parse
    } else {
      set id "[$parse getSearchStrategy]:[$parse getLatticeId]"
      foreach viso [itcl::find objects -isa VisParses] {
 	if {[$viso cget -id] != $id} {
  	  continue
	}
  	catch {$viso removeParse $parse}
      }
      itcl::delete object $parse
    }
  }
 
  # refresh each row of the table
  _setCount [llength [array names allparses]]
  set r 1
  foreach id [lsort -command ::smartCompare [array names allparses]] {
    refreshrow $r $allparses($id)
    incr r
  }

  $table configure -state disabled
}


## ----------------------------------------------------------------------------
## actions to take place on pressing the detailsbutton
## ----------------------------------------------------------------------------
body AllParses::treebutton_action {} {
   eval showparse $_selection
}

## ----------------------------------------------------------------------------
## actions to take place on pressing the detailsbutton
## ----------------------------------------------------------------------------
body AllParses::verifybutton_action {} {
   eval verifyparse $_selection
}

## ----------------------------------------------------------------------------
## action taking place when pressing the deletebutton
## ----------------------------------------------------------------------------
body AllParses::deletebutton_action {} {

  eval deleteparse $_selection
  set _selection ""
}

## ----------------------------------------------------------------------------
## delete given parses
## ----------------------------------------------------------------------------
body AllParses::deleteparse {args} {

  if {$args == {} } {
    ::cmd::Puts "ERROR: please select a parse"
    return
  }

  foreach id $args {
    set parse [getCData $id]
    foreach viso [itcl::find objects -isa VisParses] {
      catch {$viso removeParse $parse}
    }
    unset allparses($id)
    itcl::delete object $parse
  }

  init_data
  set _selection ""
}

## ----------------------------------------------------------------------------
## delete given parses
## ----------------------------------------------------------------------------
body AllParses::verifyparse {args} {

  if {$args == {} } {
    ::cmd::Puts "ERROR: please select a parse"
    return
  }

  foreach id $args {
    ::cmd::Puts "INFO: verifying $id"
    ::cmd::Verify $id
  }
}

## ----------------------------------------------------------------------------
## get all parseIds
## ----------------------------------------------------------------------------
body AllParses::getAllparseIds {} {
   return [array names allparses]
}

## ----------------------------------------------------------------------------
## get names of all known annotations 
## ----------------------------------------------------------------------------
body AllParses::getAllAnnoIds {} {
  set result ""
  set h [InputStruct_annotations_get [inputCurrentGrammar_get]]
  set m [hashListOfKeys $h]
  for {set l $m} {$l != "NULL"} {set l [listNext $l]} {
    lappend result [pointer2string [listElement $l]]
  }
  listDelete $m
  return $result
}


## ----------------------------------------------------------------------------
## Return a parse created from the C annotation ANNO.
## ----------------------------------------------------------------------------
body AllParses::getParseForAnno {anno} {

  set cparse [parseFromAnno $anno]
  if {$cparse == "NULL"} {
    return "NULL"
  }

  set lat [findLattice [AnnoEntryStruct_lattice_get $anno]]
  if {$lat == "NULL"} {
    ::cmd::Puts "WARNING: no lattice `[AnnoEntryStruct_lattice_get $anno]' found. Parse will be abstract."
  } else {
    
    set lg [lgNew $lat]
    if {$lg != "NULL"} {
      parseDecorate $cparse $lg $anno
      parseEval $cparse
    }
  }
  return $allparses([addParse $cparse])
}

## ----------------------------------------------------------------------------
## Get parse of a given id by looking it up in allparses().
##
## If no such Parse exists, try to create one from the annotation
## with the same name and store it there, then return it.
## ----------------------------------------------------------------------------
body AllParses::getCData {id} {

  # maybe create parse and store it in allparses()
  if {![info exists allparses($id)]} {

    # look among Tcl Parse objects
    set found 0
    foreach parse [itcl::find objects -isa Parse] {
      if {$id == [$parse getId]} {
        incr found
        break
      }
    }

    # create new C Parse from annotation with the same name
    if {!$found} {
      set anno [findAnnoByName $id 1]
      if {$anno == "NULL"} {
        return "NULL"
      }
      return [getParseForAnno $anno] 
    } else {
      set allparses($id) $parse
      _setCount [llength [array names allparses]]
    }
  }
  
  # return the result
  return $allparses($id)
}

## ----------------------------------------------------------------------------
## Get a parse from an annotation of lattice ID.
##
## This is the more deperate version of getParseForAnno: instead of
## annotations with the name ID, it accepts annotations whose
## lattice's name is ID. You should have tried getParseForAnno first.
## ----------------------------------------------------------------------------
body AllParses::getParseForLattice {id} {

  set lat [.cdgmain wordgraphs getCData $id]
  if {$lat == "NULL"} {
    return "NULL"
  }
  set anno [findAnnoForLattice $lat 1]
  if {$anno == "NULL"} {
    return "NULL"
  }
  return [getParseForAnno $anno] 
}


## ----------------------------------------------------------------------------
## select parses of a specified netId.
## previous selections are cleared
## ----------------------------------------------------------------------------
body AllParses::selectIdsOfNet {netId} {
   set table $itk_component(table)

   set rmax [$table cget -rows]
   $table selection clear 0,0 end
   
   set newselection ""
   for {set row 0} {$row < $rmax} {incr row} {
      if {[$table getCell $row,2] == $netId} {
	 set id [$table getCell $row,0 ]
	 $table selection set $row,0
	 lappend newselection $id
      }
   }
   set _selection $newselection
}

## ----------------------------------------------------------------------------
## select parses of a specified Lattice
## previous selections are cleared
# ----------------------------------------------------------------------------
body AllParses::selectIdsOfLattice {latticeId} {
   set table $itk_component(table)

   set rmax [$table cget -rows]
   $table selection clear 0,0 end
   
   set newselection ""
   for {set row 0} {$row < $rmax} {incr row} {
      if {[$table getCell $row,1] == $latticeId} {
	 set id [$table getCell $row,0 ]
	 $table selection set $row,0
	 lappend newselection $id
      }
   }
   set _selection $newselection
}


## ----------------------------------------------------------------------------
## Display parses graphically.
##
## All parameters are interpreted as the names of parses. If no parse
## of the given name exists, one is constructed if possible,
## preferably from an annotation with the same name, or if that fails,
## from an annotation of the lattice of the same name.
##
## If a parameter has the form parse0:1,2,3 or similar, then not only
## is parse0 displayed, but the edges 1, 2, and 3 are immediately
## highlighted. 
## ----------------------------------------------------------------------------
body AllParses::showparse {args} {

   if {$args == {} } {
      ::cmd::Puts "ERROR: please select a parse"
      return
   }

   set allvis ""
   foreach id $args {
     
     # detect details specification: parse3:3,17
     set details ""
     regexp {^(.+):([0-9,]+)} $id dummy id details
     
     set parse [getCData $id]
     if {$parse == "NULL"} {
       set parse [getParseForLattice $id]
     }
     if {$parse == "NULL"} continue

     set method [$parse getSearchStrategy]
     set latticeId [$parse getLatticeId]
     set id "$method:$latticeId"

     # find or create a window to display $parse in
     set foundVis ""
      foreach vis [itcl::find objects -isa VisParses] {

        # display each sentence in one window
        if {$manyWindows == 1} {
          if {[$vis cget -title] == $id} {
            regexp "::(.*)" $vis dummy foundVis
            break
          }
        } else {

          # display all parses in the same window
	  regexp "::(.*)" $vis dummy foundVis
	  break
	}
      }

      if {$foundVis == ""} {
        set foundVis [VisParses ".#auto" \
                          -title "Xcdg - $id" \
                          -iconname $id ]
		      
      wm withdraw $foundVis

      }
      lappend allvis $foundVis
     $foundVis addParse $parse 
     $foundVis highlight $parse $details
     $foundVis view 0
   }

   foreach vis $allvis {
      after idle [wm deiconify $vis]
   }
}



## ----------------------------------------------------------------------------
## add all parses of a given net
## ----------------------------------------------------------------------------
body AllParses::addParsesOfNet {netId} {

   set net [.cdgmain networks getCData $netId]
   if {$net == "NULL"} {
      ::cmd::Puts "ERROR: net `$netId' not found"
      return "NULL"
   }
   set parses [ConstraintNetStruct_parses_get $net]
   set parseids ""

   for {set l $parses} {$l != "NULL"} {set l [listNext $l]} {
      set currentparse [listElement $l]
      lappend parseids [addParse $currentparse]
   }

   return $parseids
}

## ----------------------------------------------------------------------------
## add parse
## ----------------------------------------------------------------------------
body AllParses::addParse {currentparse} {

  # generate a tcl parse object representing the C parse
  set parse [Parse "::#auto"]
  $parse init $currentparse
  $parse register

  # put the Tcl Parse into the array allparses
  set id [$parse getId]
  set allparses($id) $parse
  _setCount [llength [array names allparses]]
  refreshid $id

  return $id
}


## ----------------------------------------------------------------------------
## This function is called whenever a solution method has yielded
## another partial result. 
## ----------------------------------------------------------------------------
body AllParses::handlePartialResult {parse_pointer} {
  
  global errorInfo
  if {[catch {

    set origin "[ParseStruct_searchStrategy_get $parse_pointer] [ParseStruct_latticeId_get $parse_pointer]"

    # when doing an IC run, the parses handed over by the frobbing
    # stage should not be displayed after all, because they already
    # appear in the IC window.

    if {$origin == "frobbing interactive"} {
      return
    }
      
    # obey autoshow flag 
    if {$autoshow == "none"} {
      return
    }

    set parse $allparses([addParse $parse_pointer])

    # only show parses in softland
    set score [$parse getScore]
    if {$score == 0.0 && $autoshow == "valid"} {
      return
    }

    # find the window to put it in...
    set foundVis ""
    set id "[$parse getSearchStrategy]:[$parse getLatticeId]"
    foreach vis [itcl::find objects -isa VisParses] {
      if {[$vis cget -title] == $id} {
	regexp "::(.*)" $vis dummy foundVis
	break
      }
    }

    # ...or get a new one
    if {$foundVis == ""} {
      set foundVis [VisParses ".#auto" -title "Xcdg - $id" -iconname $id]
    }

    $foundVis addParse $parse 
    $foundVis view end
    wm deiconify $foundVis
  } msg]}  {
    if {$msg != ""} {
      ::cmd::Puts "ERROR: $errorInfo"
    }
  }
}


## ----------------------------------------------------------------------------
## This function is called through the hook "IC interaction".
## It returns another word typed by the user, or "" to
## denote "stop processing". 
## ----------------------------------------------------------------------------
body AllParses::handleICinteraction {} {

  # What window does user input come from?
  # I think it is fair to assume that the interactive IC window 
  # should be unique.
  set foundVis ""
  foreach vis [itcl::find objects -isa VisParses] {
    if {[$vis cget -title] == "IC:interactive"} {
      regexp "::(.*)" $vis dummy foundVis
      break
    }
  }
  
  # If no interactive IC window exists, the user must have aborted
  # parsing. This is signified by an empty word.
  if {$foundVis == ""} {
    return ""
  }

  set result [$foundVis readWord]
  
  return $result

}


## ----------------------------------------------------------------------------
## Switch display of parses in one/many windows
## ----------------------------------------------------------------------------
body AllParses::setManyWindows {x} {
  set manyWindows $x
}

## ----------------------------------------------------------------------------
## Called when the CDG grammar changes and all nets, parses etc.
## become invalid. It deletes all Tcl parses, their VisParses, the
## corresponding ParseTrees, etc.
##
## If Tcl ever decides to provide graphical representations of
## constraint nets or lexicon entries, these must be cleared here as well.
## ----------------------------------------------------------------------------
body AllParses::reset {} {
  foreach p [itcl::find objects -isa Parse] {
    deleteparse [$p getId]
  }
}
