# Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## CdgBusy - dialog indicating work in progress.
## This class is used to display a dialog while a computational intensive operation
## is scheduled in the background. It stays open as long as the bacground computation has
## no finished. The user is given an "interrupt" button to signal the background computation
## that it might terminate earlier. Actually to let the user do that the background
## computation has to be written cooperatively as we have no real concurrency in tcl.
## This is accomplished by a frequent call to "update" somewhere in your computation
## loop.
##
## The busy dialog should also be closed in case of any errors. This is accomplished
## by overwriting the tcl commands "error", "catch" and "bgerror" in \ref error.tcl.
##
## \author Michael Daum (see also AUTHORS and THANKS for more)
## $Id: busy.tcl,v 1.22 2004/02/25 14:40:42 micha Exp $
## ----------------------------------------------------------------------------
class CdgBusy {
  inherit itk::Widget

  # public methods
  public method compute {text args}; ## \type TclString, TclList
  public method interrupt {args}; ## \type TclList
  # public variables
  public variable wait 500

  constructor {args} {}; ## \type TclList

  # private methods
  private method blink {}
  private method reset {}
  private method enterAction {w x y}
  private method leaveAction {}

  # private variables
  private variable computeText ""; ## \type TclString
  private variable oldCursor ""; ## \type TclString
  private variable ledState 0; ## \type Boolean
  private variable isActive 0; ## \type Boolean
  private variable blinkJob ""; ## \type AfterJob
};


## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body CdgBusy::constructor {args} {
  global env
  
  set isActive 0
  set ledState 0

  image create photo led_off -file $env(XCDG_PATH)/g_grey.gif
  image create photo led_on -file $env(XCDG_PATH)/g_red_on.gif

  itk_component add intButton {
   button $itk_interior.intButton\
       -anchor w \
       -image led_off \
       -command [code $this interrupt] \
       -state disabled \
       -relief flat
  } {
      keep -font -background -foreground
  }

  pack $itk_component(intButton) -side right
       	
  eval itk_initialize $args
}

## ----------------------------------------------------------------------------
## execute the command.
## This method first shows the dialog and the spawns an \c after _compute().
## \param text the message to be displayed in the dialog
## \param args the script to be computed.
## \returns the return value of the evaluated script
## ----------------------------------------------------------------------------
body CdgBusy::compute {text args} {
  global _errorFlag


  if {$isActive > 0} {
    cmd::Puts "ERROR: cannot execute \n$text"
    return 
  } 

  set _errorFlag 0
  cdgCtrlCAllowed_set 1
  .cdgmain help showstr $text
  set computeText $text
  bind $itk_interior <Enter> [code $this enterAction %W %X %Y]
  bind $itk_interior <Leave> [code $this leaveAction]

  set isActive 1
  set oldCursor [. cget -cursor]
  $itk_component(intButton) configure -state normal
  . configure -cursor watch
  update idletask

  set blinkJob [after idle [code $this blink]]
  update idletask
  set result [uplevel $args]
  reset
  return $result
}

## ----------------------------------------------------------------------------
## interrupt the current computation.
## ----------------------------------------------------------------------------
body CdgBusy::interrupt {args} {
  global _errorFlag 

  if {$isActive <= 0} {
    return
  }
   
  if {[cdgCtrlCAllowed_get]} {
    ::cmd::Puts "WARNING: Interrupt received, wait a moment ..."
    cdgCtrlCTrapped_set 1
    incr isActive -1
    if {$isActive <= 0 || $_errorFlag} {
       reset
    } 
  } else {
    ::cmd::Puts "WARNING: Interrupt suppressed, try later."
  }
}

## ----------------------------------------------------------------------------
## indicate computation
## ----------------------------------------------------------------------------
body CdgBusy::blink {} {
  if {$ledState} {
    $itk_component(intButton) configure -image led_off
    set ledState 0
  } else {
    $itk_component(intButton) configure -image led_on
    set ledState 1
  }

  set blinkJob [after 1000 [code $this blink]]
  update idletask
}

## ----------------------------------------------------------------------------
## reset the blinker to a normal state.
## ----------------------------------------------------------------------------
body CdgBusy::reset {} {
  set isActive 0
  . configure -cursor $oldCursor
  .cdgmain help showstr ""
  if {$blinkJob != ""} {
    after cancel $blinkJob
  }
  $itk_component(intButton) configure -state disabled -image led_off
  set ledState 0
  set blinkJob ""
  set oldCursor ""
  set computeText ""
  bind $itk_interior <Enter> ""
  .cdgmain balloon off
}

## ----------------------------------------------------------------------------
## display the busy balloon.
## This slot is connected to the <Enter> event.
## ----------------------------------------------------------------------------
body CdgBusy::enterAction {w x y} {
  .cdgmain balloon on $x $y $computeText
}

## ----------------------------------------------------------------------------
## disable the busy balloon.
## This slot is connected to the <Leave> event.
## ----------------------------------------------------------------------------
body CdgBusy::leaveAction {} {
  .cdgmain balloon off
}
