# Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# Author: Michael Daum
# $Id: mytable.tcl,v 1.16 2004/02/25 14:40:41 micha Exp $

## ----------------------------------------------------------------------------
## MyTable - OO wrapper for tktable.
## Some time aggo Jeffrey Hobbs, the maintainer of tktable, discontinued the
## 00 wrapper for this marvelous widget. So here is my version of it.
## Its main purpose is to delegate calls from MyTable to the embedded table
## component. 
##
## There are some value addons that have been done ontop, that is
## - the table is sorteable (MyTable::sortRows())
## - clearing table cells is easier (MyTable::erase())
## - provide an array to store the table content
## - vertical and horizontal scrollbars 
## - table rows are collored
## - an improved widget layout
## ----------------------------------------------------------------------------
class MyTable {
   inherit itk::Widget

   # new options
   itk_option define -vscrollmode vscrollMode ScrollMode static
   itk_option define -hscrollmode hscrollMode ScrollMode static
   itk_option define -selforeground selForeground SelForeground yellow
   itk_option define -selbackground selBackground SelBackground gray50
   itk_option define -state state State disabled
   itk_option define -rows rows Rows 20
   itk_option define -cols cols Cols 10
   itk_option define -outerborderwidth outerBorderWidth Width 2
   itk_option define -outerrelief outerRelief Relief sunk
   itk_option define -troughcolor troughColor Background gray

   # public methods
   public method activate {args}; ## \type TclList
   public method childsite {}
   public method clear {args}; ## \type TclList
   public method delete {args}; ## \type TclList
   public method erase {fromCell toCell}; ## \type CellSpec, CellSpec
   public method getCell {args}; ## \type TclList
   public method hset {args}; ## \type TclList
   public method icursor {args}; ## \type TclList
   public method index {args}; ## \type TclList
   public method insert {args}; ## \type TclList
   public method print {}
   public method see {args}; ## \type TclList
   public method selection {args}; ## \type TclList
   public method setCell {args}; ## \type TclList
   public method spans {args}; ## \type TclList
   public method tag {args}; ## \type TclList
   public method vset {args}; ## \type TclList
   public method width {args}; ## \type TclList
   public method xview {args}; ## \type TclList
   public method yview {args}; ## \type TclList
   public method curselection {args }; ## \type TclList
   public method sortRows {index args}; ## \type TclNumber, TclList

   constructor {args} {}; ## \type TclList


   # private methods
   private method resize {} 
   private method colorize {num} ; ## \type TclNumber

   ## storage for the table content.
   private variable _myArray
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body MyTable::constructor {args} {

   # define all widget-components
   itk_component add frame {
      frame $itk_interior.frame
   } {
     rename -borderwidth -outerborderwidth outerBorderWidth Width
     rename -relief -outerrelief outerRelief Relief
   }

   itk_component add table {
      table $itk_component(frame).table \
		-relief flat \
		-drawmode slow \
		-variable [scope _myArray] \
		-xscrollcommand [code $this hset] \
		-yscrollcommand [code $this vset] \
		-rowtagcommand [code $this colorize ] \
		-borderwidth 2 \
		-highlightthickness 0
   } {
      keep -state -background -cursor -browsecommand -selectioncommand\
	   -colstretchmode -rowstretchmode -rowseparator -colseparator \
	   -exportselection -selectmode -selecttype -titlecols \
	   -titlerows -rowtagcommand -coltagcommand -font -borderwidth \
	   -selecttitles
      rename -anchor -cellanchor cellAnchor Anchor
   }
   $itk_component(table) tag config title \
		-anchor c \
		-relief raised \
		-borderwidth 1
   $itk_component(table) tag config sel \
		-relief flat
   $itk_component(table) tag config active \
		-anchor w \
		-fg black \
		-bg gray
   

   itk_component add vbar {
      scrollbar $itk_component(frame).vbar \
		-orient vertical \
		-command [code $this yview] \
		-elementborderwidth -1 \
		-width 12 \
		-borderwidth 2
   } { }

   itk_component add hbar {
      scrollbar $itk_component(frame).hbar \
		-orient horizontal \
		-command [code $this xview] \
		-elementborderwidth -1 \
		-width 12 \
		-borderwidth 2
   } { }

   # packing + gridding
   pack $itk_component(frame) -fill both -expand 1
   grid $itk_component(table) -row 0 -column 0 -sticky news -padx 1
   grid $itk_component(vbar) -row 0 -column 1 -sticky ns -pady 1
   grid $itk_component(hbar) -row 1 -column 0 -sticky ew -padx 2 -pady 1

   grid columnconfig $itk_component(frame) 0 -weight 1
   grid rowconfig $itk_component(frame) 0 -weight 1

   # some bindings
   bind $itk_component(table) <Configure> [code $this resize]

   eval itk_initialize $args
}

## ----------------------------------------------------------------------------
## xview.
## ----------------------------------------------------------------------------
body MyTable::xview {args} {
   eval $itk_component(table) xview $args
}

## ----------------------------------------------------------------------------
## yview
## ----------------------------------------------------------------------------
body MyTable::yview {args} {
   eval $itk_component(table) yview $args
}

## ----------------------------------------------------------------------------
## delegate the set command to the vertical scrollbar.
## ----------------------------------------------------------------------------
body MyTable::vset {args} {
   eval $itk_component(vbar) set $args
}

## ----------------------------------------------------------------------------
## delegate the set command to the horizontal scrollbar.
## ----------------------------------------------------------------------------
body MyTable::hset {args} {
   eval $itk_component(hbar) set $args
}

## ----------------------------------------------------------------------------
## propagate table resizing to the scrollbars.
## ----------------------------------------------------------------------------
body MyTable::resize {} {
   if {$itk_option(-hscrollmode) == "dynamic"} {
      if {[string comp {0 1} [xview]]} {
	 grid $itk_component(hbar)
      } else {
	 grid remove $itk_component(hbar)
      }
   }

   if {$itk_option(-vscrollmode) == "dynamic"} {
      if {[string comp {0 1} [yview]]} {
	 grid $itk_component(vbar)
      } else {
	 remove $itk_component(vbar)
      }
   }
}
## ----------------------------------------------------------------------------
## rowtag callback of the table.
## This method is called whenever the tktable component needs a tag for a
## table row.
## ----------------------------------------------------------------------------
body MyTable::colorize {num} {
   if {$num>0 && $num%2} { 
      return "" 
   } else {
      return colored
   }
}

## ----------------------------------------------------------------------------
## print the stored data.
## This simply calls parray on the _myArray variable.
## ----------------------------------------------------------------------------
body MyTable::print {} {
    parray _myArray
}

## ----------------------------------------------------------------------------
## erase table cells.
## This method clears the content between two cell specifications by directly
## manipulating the _myArray. A cell specification has the format \a "row,col",
## where \a row and \a col are integers.
## \param fromCell the cell which is the first to be cleared
## \param toCell the cell which is the last to be cleared
## ----------------------------------------------------------------------------
body MyTable::erase {fromCell toCell} {
  
  scan $fromCell "%d,%d" fromRow fromCol
  scan $toCell "%d,%d" toRow toCol

  if {$fromRow > $toRow} {
    set temp $fromRow
    set fromRow $toRow
    set toRow $temp
  }
  if {$fromCol > $toCol} {
    set temp $fromCol
    set fromCol $toCol
    set toCol $temp
  }

  for {set i $fromRow} {$i <= $toRow} {incr i} {
    for {set j $fromCol} {$j <= $toCol} {incr j} {
      set _myArray($i,$j) ""
    }
  }
}

## ----------------------------------------------------------------------------
## clear
## ----------------------------------------------------------------------------
body MyTable::clear {args} {
  eval $itk_component(table) clear $args
}

## ----------------------------------------------------------------------------
## retrieve the table component.
## \returns the widget path to the tktable
## ----------------------------------------------------------------------------
body MyTable::childsite {} {
   return $itk_component(table)
}

## ----------------------------------------------------------------------------
## delegate the tag command to the table
## ----------------------------------------------------------------------------
body MyTable::tag {args} {
   eval $itk_component(table) tag $args
}

# ----------------------------------------------------------------------------
# delegate the see command to the table
# ----------------------------------------------------------------------------
body MyTable::see {args} {
   eval $itk_component(table) see $args
}

## ----------------------------------------------------------------------------
## delegate the selection command to the table.
## ----------------------------------------------------------------------------
body MyTable::selection {args} {
   eval $itk_component(table) selection $args
}

## ----------------------------------------------------------------------------
## delegate the set command to the table
## ----------------------------------------------------------------------------
body MyTable::setCell {args} {
  eval $itk_component(table) set $args
  
#  set _myArray($index) $value
#  if {$args != ""} {
#     eval setCell $args
#  }
}

## ----------------------------------------------------------------------------
## delegate the width command to the table
## ----------------------------------------------------------------------------
body MyTable::width {args} {
   eval $itk_component(table) width $args
}

## ----------------------------------------------------------------------------
## delegate the index command to the table
## ----------------------------------------------------------------------------
body MyTable::index {args} {
   eval $itk_component(table) index $args
}

## ----------------------------------------------------------------------------
## delegate the insert command to the table
## ----------------------------------------------------------------------------
body MyTable::insert {args} {
  eval $itk_component(table) insert $args
}

## ----------------------------------------------------------------------------
## delegate the delete command to the table
## ----------------------------------------------------------------------------
body MyTable::delete {args} {
   eval $itk_component(table) delete $args
}

## ----------------------------------------------------------------------------
## get a value of a table-cell
## ----------------------------------------------------------------------------
body MyTable::getCell {args} {
   eval $itk_component(table) get $args
}

## ----------------------------------------------------------------------------
## delegate the activate command to the table
## ----------------------------------------------------------------------------
body MyTable::activate {args} {
   eval $itk_component(table) activate $args
}

## ----------------------------------------------------------------------------
## delegate the icursor command to the table
## ----------------------------------------------------------------------------
body MyTable::icursor {args} {
   eval $itk_component(table) icursor $args
}

## ----------------------------------------------------------------------------
## delegate the spans command to the table
## ----------------------------------------------------------------------------
body MyTable::spans {args} {
   eval $itk_component(table) spans $args
}

## ----------------------------------------------------------------------------
## delegate the curselection command to the table
## ----------------------------------------------------------------------------
body MyTable::curselection {args} {
   eval $itk_component(table) curselection $args
}

## ----------------------------------------------------------------------------
## sort the table rows in the given column.
## This method sorts the table along the column given by the parameter \a index.
## The sorting itself is the done by a \c lsort command 
## \param index the column which along which the table is sorted
## \param args the arguments for the \c lsort command
## ----------------------------------------------------------------------------
body MyTable::sortRows {index args} {
  set table $itk_component(table)
  set rmax [$table cget -rows]
  set cmax [$table cget -cols]
  set noTitleRows [$table cget -titlerows]

  # build a list of sublist representing the array
  set restList ""
  set tableList ""
  for {set rowNo $noTitleRows} {$rowNo < $rmax} {incr rowNo} {
    set row [$table get $rowNo,0 $rowNo,$cmax]
    set cell [lindex $row $index]
    if {$cell == "" || $cell == "{}"} {
      lappend restList $row
    } else {
      lappend tableList $row
    }
  }

  if {$tableList != ""} {
    set tableList [eval lsort $args -index $index \$tableList]
  }

  # restore the rows
  erase $noTitleRows,0 $rmax,$cmax
  set rowNo $noTitleRows

  foreach row $tableList {
    $table set row $rowNo,0 $row
    incr rowNo
  }
  foreach row $restList {
    $table set row $rowNo,0 $row
    incr rowNo
  }
}

## ----------------------------------------------------------------------------
## option -troughcolor.
## set the troughcolor of the scollbars
## ----------------------------------------------------------------------------
configbody MyTable::troughcolor {
  $itk_component(hbar) configure -troughcolor $itk_option(-troughcolor)
  $itk_component(vbar) configure -troughcolor $itk_option(-troughcolor)
}

## ----------------------------------------------------------------------------
## option -vscrollmode.
## Enable/disable display and mode of vertical scrollbar.
## ----------------------------------------------------------------------------
configbody MyTable::vscrollmode {
    switch $itk_option(-vscrollmode) {
	static	{ grid $itk_component(vbar) }
	dynamic	-
	none	{ grid remove $itk_component(vbar) }
	default {
	    error "bad vscrollmode option\
		    \"$itk_option(-vscrollmode)\": should\
		    be static, dynamic, or none"
	}
    }
}

## ----------------------------------------------------------------------------
## option -hscrollmode.
## Enable/disable display and mode of horizontal scrollbar.
## ----------------------------------------------------------------------------
configbody MyTable::hscrollmode {
    switch $itk_option(-hscrollmode) {
	static	{ grid $itk_component(hbar) }
	dynamic	-
	none	{ grid remove $itk_component(hbar) }
	default {
	    error "bad hscrollmode option\
		    \"$itk_option(-hscrollmode)\": should\
		    be static, dynamic, or none"
	}
    }
}

## ----------------------------------------------------------------------------
## option -selforeground.
## set the color of the selected row
## ----------------------------------------------------------------------------
configbody MyTable::selforeground {
   $itk_component(table) tag config sel \
		-foreground $itk_option(-selforeground)
}

## ----------------------------------------------------------------------------
## option -selbackground.
## set the color of the selected row
## ----------------------------------------------------------------------------
configbody MyTable::selbackground {
   $itk_component(table) tag config sel \
		-background $itk_option(-selbackground)
}

## ----------------------------------------------------------------------------
## option -rows.
## configure the number of rows of the table
## ----------------------------------------------------------------------------
configbody MyTable::rows {
   $itk_component(table) configure -rows $itk_option(-rows)
   resize
}

## ----------------------------------------------------------------------------
## option -cols.
## configure the number of cols of the table
## ----------------------------------------------------------------------------
configbody MyTable::cols {
   $itk_component(table) configure -cols $itk_option(-cols)
   resize
}

