# Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## ParseTree - visualization of one parse.
##
## ParseTree objects are the central means to give you a nice
## dependency tree drawing. They are only thought to be used as
## components of a VisParses object being the GUI surrounding a set of
## visualized ParseTree objects. Taking the embedded Parse object into
## account, these tree classes implement the tree editor in XCDG.
##
## \author Michael Daum, Kilian A. Foth, Dietmar Fnning 
## (see also AUTHORS and THANKS for more)
##
## $Id: parsetree.tcl,v 1.107 2004/02/25 14:40:41 micha Exp $
## ----------------------------------------------------------------------------

class ParseTree {
   inherit ::itk::Widget 

   # new options

   ## itk_option: color of a node of the dependency tree
   itk_option define -nodecolor nodeColor Color "black"

   ## itk_option: color of edges up to the parents
   itk_option define -edgecolor edgeColor Color "black"

   ## itk_option: color of vertical lines
   itk_option define -vlinecolor vlineColor Color "gray" 

   ## itk_option: font-property for the words in the sentence
   itk_option define -wordfont wordFont Font "" 
   
   ## itk_option: font-property of the dependency labels
   itk_option define -labelfont labelFont Font "" 

   ## itk_option: color of the words in the sentence
   itk_option define -wordcolor wordColor Color "black"

   ## itk_option: color of the dependency labels
   itk_option define -labelcolor wordColor Color "black"

   ## itk_option: decoration of a level component
   itk_option define -levelrelief levelRelief Relief "flat"

   ## itk_option: decoration of a level component
   itk_option define -levelborderwidth levelBorderwidth Borderwidth 1

   ## the embedded Parse object
   private variable _parse "" ;## \type ParseTree

   ## who is my daddy
   private variable _visparses ""; ## \type VisParses
  
   ## array holding the middle points of the words.
   ## initialized by drawText; indexed by timepoints, e.g. 0
   private variable _textMidPos; ## \type TclArray

   ## holds all items on a canvas belonging to one timepoint.
   ## Example: _itemsOfTimepoint(SYN,3) == 4 27 26 24 25
   private variable _itemsOfTimepoint; ## \type TclArray

   ## array holding the timepoint which a specified id belongs to.
   ## Example: _timepointOfId(SYN,3) = 0
   private variable _timepointOfId; ## \type TclArray

   ## array storing arcs by level and timepoint.
   ## Example: _arcOfTimepoint(3,OBL1) = 47 48
   private variable _arcOfTimepoint; ## \type TclArray

   ## holds ids of canvas items to be highlighted when touching a violation.
   ## Example: _itemsOfViolation(1) == 23 SYN 22 SYN
   private variable _itemsOfViolation; ## \type TclArray

   ## array storing the canvas of a level.
   private variable _levelOfCanvas; ## \type TclArray
 
   ## array indicating whether a level is already drawn.
   private variable _isDrawn; ## \type TclArray

   private variable _dragItem ""
   private variable _activeItem ""
   private variable _activeBinding ""
   private variable _dragStartX 0
   private variable _dragStartY 0
   private variable _zoomstep 0
   private variable _oldColor ""

   private variable _lastxmark 0
   private variable _lastymark 0

   ## Command-History-Object for undo-implementation
   private variable _commandHistory ""



   public method drawAll {}
   public method writeToFile {fileName}; ## \type TclString
   public method mirror {}
   public method verify {}
   public method showcycles {}
   public method breakcycles {}
   public method zoom {direction}; ## \type TclNumber
   public method setOrientation {orient}; ## \type TclString
   public method getOrientation {}
   public method parse {args}; ## \type TclList
   public method mainlevelname {}; ## \type TclString
   public method highlight {which}

   ## Methods used to implement undo-functionality (called from VisParses)
   public method undo {}
   public method redo {}
   public method addUndoListener {listener}; ## \type TclString
   public method canUndo {}
   public method canRedo {}


   constructor {p vp args} {}; ## \type Parse, VisParses, TclList
   destructor {}
 
   ## Methods used to implement undo-functionality (Internal methods)
   private method undoEdge {modifier modifiee level} ; # \type TclNumber, TclNumber, TclString
   private method undoLabel {modifier label level}; # \type TclNumber, TclString, TclString
   private method undoLexeme {from description}; # \type TclNumber, TclString

   private method _zoom {step}; ## \type TclNumber
   private method allEnter {w}; ## \type Canvas
   private method allLeave {w}; ## \type Canvas
   private method edgeEnter {w}; ## \type Canvas
   private method edgeLeave {w}; ## \type Canvas
   private method allSybillings {id1 id2 level}; ## \type TclNumber, TclNumber, TclString
   private method canvasToScreen {canvas x y}; ## \type Canvas, TclNumber, TclNumber
   private method createArc {canvas x1 y1 x2 y2 text bindNo}; ## \type Canvas, TclNumber, TclNumber, TclNumber, TclNumber, TclNumber, TclNumber
   private method drawArcs {}
   private method drawLevel {level}; ## \type TclString
   private method drawText {canvas x y}; ## \type Widget, TclNumber, TclNumber
   private method drawViolas {}
   private method canvasClick {w x y}; ## \type Canvas, TclNumber, TclNumber
   private method edgeClick {w x y}; ## \type Canvas, TclNumber, TclNumber
   private method edgeDrop {w x y}; ## \type Canvas, TclNumber, TclNumber
   private method edgeDrag {w x y}; ## \type Canvas, TclNumber, TclNumber
   private method edgeDragPan {w x y}; ## \type Canvas, TclNumber, TclNumber
   private method edgeMoveCancel {w}
   private method fixCoords {w x y}; ## \type Canvas, TclNumber, TclNumber
   private method labelClick {w}; ## \type Canvas
   private method labelRightClick {w}; ## \type Canvas
   private method labelSelect {label}; ## \type TclString
   private method makeUserInterface {w}; ## \type Canvas
   private method moveEdge {canvas x1 y1 x2 y2}; ## \type Canvas, TclNumber, TclNumber, TclNumber, TclNumber
   private method moveLabel {canvas label_id x1 y1 x2 y2}; ## \type Canvas, TclString, TclNumber, TclNumber, TclNumber, TclNumber
   private method multiCanvas {args}; ## \type TclList
   private method multiMark {x y}; ## \type TclNumber, TclNumber
   private method multiDragto {x y}; ## \type TclNumber, TclNumber
   private method noCrossing {id bindings}; ## \type TclString, TclList
   private method paintTabs {}
   private method reDrawLevel {level}; ## \type TclString
   private method registerItems {level}; ## \type TclString
   private method rgValue {score}; ## \type TclNumber
   private method showPopMenu {w x y items}; ## \type Canvas, TclNumber, TclNumber, TclList
   private method violaBrowse {mode xcoord ycoord}; ## \type TclString, TclNumber, TclNumber
   private method wordClick {w}; ## \type Canvas
   private method wordRightClick {w}; ## \type Canvas
   private method wordEnter {w x y}; ## \type Canvas, TclNumber, TclNumber
   private method wordLeave {w}; ## \type Canvas
   private method wordSelect {w from lexeme}; ## \type Canvas TclNumber, LexemNode
   private method compareLexeme {a b}; ## \type LexiconItem, LexiconItem
   private method closestNode {w x y};
   private method findLabel {w e};
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body ParseTree::constructor {p vp args} {

  set _parse $p
  set _visparses $vp

  ## instantiating command-history-object
  set _commandHistory [CommandHistory $this.commandHistory]


  set levels [$_parse getLevels]
  foreach level $levels {
    set _isDrawn($level) 0
  }

  # split pane
  itk_component add pane {
     iwidgets::panedwindow $itk_interior.pane \
       -height 10c -width 10c  -orient horizontal
  } {}
  set pane $itk_component(pane)
  $pane add "Parse"
  set pane1 [$pane childsite "Parse"]
  $pane reset

  if {![$_parse isAbstract]} {
    $pane add "Conflicts"
    set pane2 [$pane childsite "Conflicts"]
  }

  # tabular containing the levels
  itk_component add tbn {
     iwidgets::Tabnotebook $pane1.tbn
  } {}

  if {![$_parse isAbstract]} {

    # conflicts
    itk_component add conflicts {
      MyTable $pane2.conflicts \
          -exportselection 0 \
          -cols 6 \
          -state disabled \
          -colstretchmode unset \
          -selecttype row \
          -font {Helvetica -12 bold} \
          -background gray90 \
          -titlerows 0  \
          -outerborderwidth 1
    } {}
    bind [$itk_component(conflicts) component table] <ButtonPress-1> \
        [code $this violaBrowse 1 %x %y]
    bind [$itk_component(conflicts) component table] <ButtonPress-3> \
        [code $this violaBrowse 2 %x %y]
  }

   # popup menu
   itk_component add popmenu {
      menu $itk_interior.popmenu
   } {
      keep -background -cursor
      rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
   }

   # build a scrolled canvas for each level 
   set tbn $itk_component(tbn)
   foreach level $levels {

     # add page
     set page [$tbn add  \
       	-label $level \
       	-command [code $this drawLevel $level]]

     # add canvases
     itk_component add $level {
        iwidgets::scrolledcanvas $page.cs 
     } {
        keep -cursor -background
        rename -relief -levelrelief levelRelief Relief
        rename -borderwidth -levelborderwidth levelBorderwidth Borderwidth
     }

     # advise the horizontal scrolling commands
     # so that each scrollbar scrolls all levels.
     set hbar [$itk_component($level) component horizsb]
     $hbar configure -command [code $this multiCanvas xview]

     # fast scanning
     bind [$itk_component($level) childsite] <ButtonPress-2> [code $this multiMark %x %y] 
     bind [$itk_component($level) childsite] <B2-Motion> [code $this multiDragto %x %y] 

     pack $itk_component($level) -fill both -side top -expand 1
     set _levelOfCanvas([$itk_component($level) childsite]) $level
   }

   # packing
  if {[$_parse isAbstract]} {
    pack $pane $pane1 $tbn -expand 1 -fill both
  } else {
    pack $pane $pane1 $pane2 $tbn -expand 1 -fill both
  }
   if {! [$p isAbstract]} {
     pack $itk_component(conflicts) -expand 1 -fill both
   }

   # initialize
   eval itk_initialize $args

   drawViolas
   paintTabs

  # view main level initially
  set mainlevel [mainlevelname]
  set index [lsearch [$_parse getLevels] $mainlevel]
  if {$index == -1} {
    $tbn view 0
  } else {
    $tbn view $index
  }
  if {![$_parse isAbstract]} {
    $pane fraction 80 20 
  }
}


## ----------------------------------------------------------------------
## destructor
## ----------------------------------------------------------------------
body ParseTree::destructor {} {
    delete object $_commandHistory
}



## ----------------------------------------------------------------------------
## part of the initialization process.
## ----------------------------------------------------------------------------
body ParseTree::makeUserInterface {w} {

  # touching a word pops up a tooltip giving its description
  $w bind word <Any-Enter> [code $this wordEnter %W %X %Y]
  $w bind word <Any-Leave> [code $this wordLeave %W ]

  # clicking on a word shows its menu of alternatives
  $w bind word <ButtonPress-1> [code $this wordClick %W]
  $w bind word <ButtonPress-3> [code $this wordRightClick %W]
  
  # touching a label highlights it
  $w bind label <Any-Enter> [code $this allEnter %W ]
  $w bind label <Any-Leave> [code $this allLeave %W ]

  # clicking on a label shows its menu of alternatives
  $w bind label <ButtonPress-1> [code $this labelClick %W]
  $w bind label <ButtonPress-3> [code $this labelRightClick %W]

  # touching an edge highlights it
  $w bind edge <Any-Enter> [code $this edgeEnter %W ]
  $w bind edge <Any-Leave> [code $this edgeLeave %W ]

  # right-clicking on an edge shows its label menu
  $w bind edge <ButtonPress-3> [code $this labelClick %W]

  # dragging an edge moves it around
  $w bind edge <ButtonPress-1> [code $this edgeClick %W %x %y]
  $w bind edge <B1-Motion> [code $this edgeDrag %W %x %y]
  $w bind edge <ButtonRelease-1> [code $this edgeDrop %W %x %y]

  # clicking on the canvas pretends the nearest edge was clicked
  bind $w <ButtonPress-1> [code $this canvasClick %W %x %y]
  bind $w <B1-Motion> [code $this edgeDrag %W %x %y]
  bind $w <B1-B2-Motion> [code $this edgeDragPan %W %x %y]
  bind $w <ButtonRelease-1> [code $this edgeDrop %W %x %y]

  # typing ESC cancels an edge move action
  bind $w <Escape> [code $this edgeMoveCancel %W]

}



## ----------------------------------------------------------------------------
## command bound to Enter-event.
## ----------------------------------------------------------------------------
body ParseTree::wordEnter {w x y} {

  allEnter $w

   set level $_levelOfCanvas($w)
   set _activeItem [$w find withtag current]
   set timepointId $_timepointOfId($level,$_activeItem)
   scan $timepointId "%i%c%i" wordposition useless1 useless2


   # index for listNthElement starts with 1 !
   set lexiconId [$_parse getCurrentReading $wordposition]
   .cdgmain balloon on $x $y $lexiconId
}

## ----------------------------------------------------------------------------
## command bound to Leave-event.
## ----------------------------------------------------------------------------
body ParseTree::wordLeave {w} {
  allLeave $w
  .cdgmain balloon off
}

## ----------------------------------------------------------------------------
## command bound to Enter-event
## ----------------------------------------------------------------------------
body ParseTree::allEnter {w} {
  set _oldColor [$w itemcget current -fill]
  set cur [$w find withtag current]
  $w itemconfigure current -fill red
}

## ----------------------------------------------------------------------------
## command bound to Leave-event
## ----------------------------------------------------------------------------
body ParseTree::allLeave {w} {
  if {$_oldColor == ""} {
    set _oldColor black
  } 
  $w itemconfigure current -fill $_oldColor
  set _oldColor ""
}

## ----------------------------------------------------------------------------
## Executed when an edge is touched.
##
## This is different from allEnter because touching an edge should
## also highlight its label.
## ----------------------------------------------------------------------------
body ParseTree::edgeEnter {w} {

  # paint edge
  allEnter $w

  # paint label
  set level $_levelOfCanvas($w)
  set edge [$w find withtag current]
  set t $_timepointOfId($level,$edge)
  foreach i $_itemsOfTimepoint($level,$t) {
    if {[lsearch -exact [$w gettags $i] "label"]>=0} {
      set label $i
      break
    }
  }
  $w itemconfigure $label -fill red

}

## ----------------------------------------------------------------------------
## Executed when an edge is left by the mouse pointer.
## ----------------------------------------------------------------------------
body ParseTree::edgeLeave {w} {

  # paint edge
  allLeave $w

  # paint label
  set level $_levelOfCanvas($w)
  set edge [$w find withtag current]
  set t $_timepointOfId($level,$edge)
  foreach i $_itemsOfTimepoint($level,$t) {
    if {[lsearch -exact [$w gettags $i] "label"]>=0} {
      set label $i
      break
    }
  }
  $w itemconfigure $label -fill black

}

## ----------------------------------------------------------------------------
## command bound to violation browsing.
## ----------------------------------------------------------------------------
body ParseTree::violaBrowse {mode xcoord ycoord} {

  # get the violation 
  set index [$itk_component(conflicts) index @$xcoord,$ycoord]

  scan $index "%d,%d" viola_no col
  if {$viola_no < 0} {
    return
  }
  set viola [lindex [$_parse getViolations] $viola_no]
  set bId [lindex $viola $mode]
  if {$bId < 0} {
    return
  }
  set binding [$_parse getBindingById $bId]

  # highlight
  foreach level [$_parse getLevels] {
    set canvas [$itk_component($level) childsite]
    $canvas itemconfigure arc -fill black
    $canvas itemconfigure edge -fill black
    $canvas itemconfigure word -fill black
    $canvas itemconfigure label -fill black
  }

  foreach {no level} $_itemsOfViolation($viola_no) {
    drawLevel $level
    set canvas [$itk_component($level) childsite]
    $canvas itemconfigure $no -fill red
  }

  # set the selection on right click 
  $itk_component(conflicts) selection clear all
  $itk_component(conflicts) selection set $viola_no,0

  # focus the violation
  scan $binding "%s %s %s %s %s" modifier_id modifiee_id label bindNo level
  $itk_component(tbn) view [lsearch [$_parse getLevels] $level]
  set canvas $itk_component($level)

  if {$mode == 1} {
    set ids [lrange $_itemsOfViolation($viola_no) 0 7]
  } else {
    set ids [lrange $_itemsOfViolation($viola_no) 8 end]
  }

  set bbox [eval $canvas bbox $ids]
  set minx [lindex $bbox 0]
  set miny [lindex $bbox 1]
  if {$minx == ""} {
    set minx 0
    set miny 0
  }

  set bbox [$canvas bbox all]
  set maxx [lindex $bbox 2]
  set maxy [lindex $bbox 3]
  set fraqx [expr $minx / ( 0.0 + $maxx ) - 0.01]
  set fraqy [expr $miny / ( 0.0 + $maxy ) - 0.1]

  # don't scroll if there is nothing to scroll
  if {"[$canvas xview]" != "0 1"} {
    multiCanvas xview moveto $fraqx 
  }    
  if {"[$canvas yview]" != "0 1"} {
    multiCanvas yview moveto $fraqy 
  }
}

## ----------------------------------------------------------------------------
## compare two pointers to LexemeNode by their description.
## ----------------------------------------------------------------------------
body ParseTree::compareLexeme {a b} {
  set x [LexiconItemStruct_description_get [LexemNodeStruct_lexem_get $a]]
  set y [LexiconItemStruct_description_get [LexemNodeStruct_lexem_get $b]]
  return [string compare $x $y]

}

## ----------------------------------------------------------------------------
## show menu of available lexicon-entries when clicking on a word.
## ----------------------------------------------------------------------------
body ParseTree::wordClick {w} {

  .cdgmain balloon off

  # calculate position of the menu
  scan [$w bbox current] "%d %d %d %d" x1 y1 x2 y2
   
  # collect needed data
  set _activeItem [$w find withtag current]
  set lexemes [$_parse getLexemes]

  # which word is active now?
  # we must only offer alternatives that are from the same timespan.
  set level $_levelOfCanvas($w)
  set timepoint $_timepointOfId($level,$_activeItem)
  set word [$_parse getWord $timepoint]
  set from [WordStruct_from_get $word]
  set to [WordStruct_to_get $word]

  set items ""
  foreach l [lsort -command compareLexeme $lexemes] {
    set le [LexemNodeStruct_lexem_get $l]
    set d [LexiconItemStruct_description_get $le]

    # require exactly the same timespan
    set a [LexemNodeStruct_arc_get $l]
    set f [ArcStruct_from_get $a]
    set t [ArcStruct_to_get $a]
    if {$f != $from} continue
    if {$t != $to} continue
    lappend items [list $d [code $this wordSelect $w $timepoint $l]]
  }
 
  showPopMenu $w $x1 $y2 $items
}

## ----------------------------------------------------------------------------
## action taking place when a lexeme is chosen from a wordlist.
## ----------------------------------------------------------------------------
body ParseTree::wordSelect {w from lexeme} {

  set oldDescription [$_parse getCurrentReading $from]
   
  set le [LexemNodeStruct_lexem_get $lexeme]
  set description [LexiconItemStruct_description_get $le]
  set reading [LexiconItemStruct_word_get $le]
  $_parse swapWord $from $description

  ## Saving action to history
  if { $description != $oldDescription} {
      $_commandHistory add [code $this undoLexeme $from $description]\
	  [code $this undoLexeme $from $oldDescription]
  }

  drawViolas
  paintTabs

  # changing a different lexeme may change the reading, 
  # therefore we must redraw the word on the canvas
  set level $_levelOfCanvas($w)
  foreach word $_itemsOfTimepoint($level,$from) {
    set tags [$w gettags $word]
    if {[regexp "word" $tags]} {
      $w itemconfigure $word -text $reading
      break
    }
  }
}


## ----------------------------------------------------------------------------
## show menu of available level-labels when clicking on a label.
## ----------------------------------------------------------------------------
body ParseTree::labelClick {w} {
  scan [$w bbox current] "%d %d %d %d" x1 y1 x2 y2
  set level $_levelOfCanvas($w)
  set _activeItem [$w find withtag current]
  set timepoint $_timepointOfId($level,$_activeItem)

  # The current item may not actually be the label in question,
  # but its corresponding edge if it received a right click.
  # Pretend that the event occurred at the label.
  foreach item $_itemsOfTimepoint($level,$timepoint) {
    if {[lsearch -exact [$w gettags $item] "label"]>=0} {
      set _activeItem $item
      break
    }
  }

  set _activeBinding [$_parse getBindingAt $level $timepoint]
  set labels [.cdgmain levels getLabels $level]
  if {$labels == ""} {
    set labels [$_parse getLabels]
  }

  set items ""
  foreach label [lsort $labels] {
    lappend items [list $label [code $this labelSelect $label]]
  }
  showPopMenu $w $x1 $y2 $items

}

## ----------------------------------------------------------------------------
## Auto-correct the current label
## ----------------------------------------------------------------------------
body ParseTree::labelRightClick {w} {

  set level $_levelOfCanvas($w)
  set canvas [$itk_component($level) childsite]
  set _activeItem [$w find withtag current]
  set timepoint $_timepointOfId($level,$_activeItem)
  set label [$_parse optimizeLabel $level $timepoint]
  set _activeBinding [$_parse getBindingAt $level $timepoint]

  if {"NULL" == "$label"} return;
  labelSelect $label
   
}

## ----------------------------------------------------------------------------
## Auto-correct the current word
## ----------------------------------------------------------------------------
body ParseTree::wordRightClick {w} {

  # remove tooltip about to be obsoleted
  .cdgmain balloon off

  # collect params
  scan [$w bbox current] "%d %d %d %d" x1 y1 x2 y2
  set level $_levelOfCanvas($w)
  set canvas [$itk_component($level) childsite]
  set _activeItem [$w find withtag current]
  set timepoint $_timepointOfId($level,$_activeItem)

  # find description of best word
  set result [$_parse optimizeWord $timepoint]
  if {"NULL" != "$result"} {

    # find corresponding LexemNode
    set word [$_parse getWord $timepoint]
    set from [WordStruct_from_get $word]
    set to [WordStruct_to_get $word]
    set lexemes [$_parse getLexemes]
    foreach l [lsort -command compareLexeme $lexemes] {
      set le [LexemNodeStruct_lexem_get $l]
      set d [LexiconItemStruct_description_get $le]
      set a [LexemNodeStruct_arc_get $l]
      set f [ArcStruct_from_get $a]
      set t [ArcStruct_to_get $a]
      if {$f != $from} continue
      if {$t != $to} continue
      if {![string compare $d $result]} {
        wordSelect $w $timepoint $l
        .cdgmain balloon on $x1 $y1 $result
        return
      }
    }
  }
}

## ----------------------------------------------------------------------------
## set the label of the _activeBinding.
## ----------------------------------------------------------------------------
body ParseTree::labelSelect {label} {

  scan $_activeBinding "%s %s %s %s %s" \
      modifier modifiee oldlabel binding_id level

  if {$label == $oldlabel} {
    return
  }

  $_commandHistory add [code $this undoLabel $modifier $label $level]\
      [code $this undoLabel $modifier $oldlabel $level]


  set canvas [$itk_component($level) childsite]
  $canvas itemconfigure $_activeItem -text $label
  set _activeItem ""
  set _activeBinding ""
  $_parse swapLabel $level $modifier $label

  drawViolas
  paintTabs

}


## ----------------------------------------------------------------------------
## command bound to B1-event on the canvas background.
## ----------------------------------------------------------------------------
body ParseTree::canvasClick {w x y} {

  set cur [$w find withtag current]
  if {"" != $cur} return

   scan [fixCoords $w $x $y] "%d %d" x y 

   # fond closest edge to the click
   set record 999999
   foreach e [$w find withtag edge] {
     scan [$w coords $e] "%f %f %f %f" cx1 cy1 cx2 cy2

     # compute distance of edge to click
     set d [point_segment_distance $x $y $cx1 $cy1 $cx2 $cy2]
     if {$d < $record} {
       set record $d
       set theedge $e
     }
   }

   scan [$w coords $theedge] "%f %f %f %f" x1 y1 x2 y2
   set _dragItem $theedge
   set _dragStartX $x2
   set _dragStartY $y2
   moveEdge $w $x1 $y1 $x $y

}

## ----------------------------------------------------------------------------
## command bound to B1-event on an edge.
## ----------------------------------------------------------------------------
body ParseTree::edgeClick {w x y} {
    scan [$w coords current] "%f %f %f %f" x1 y1 x2 y2
    set _dragItem [$w find withtag current]
    set _dragStartX $x2
    set _dragStartY $y2

    $w addtag undo withtag current

    scan [fixCoords $w $x $y] "%d %d" new_x new_y 
   
    moveEdge $w $x1 $y1 $new_x $new_y
}


## ----------------------------------------------------------------------------
## Find closest circle on the canvas.
##
## One should think that this should be possible with a normal
## `canvas find' command, but it isn't; you can only search for the
## closest item, OR for all circles.
## ----------------------------------------------------------------------------
body ParseTree::closestNode {w x y} {

  set result ""
  set items [$w find withtag node]
  set record 99999

  # find coordinates of the lower end of the moving edge; 
  # selecting this node would create a direct cycle, which look ugly
  # and is very likely wrong anyway.
  set level $_levelOfCanvas($w)
  set self $_timepointOfId($level,$_dragItem)
  foreach item $items {
    if { $self == $_timepointOfId($level,$item)} continue
    scan [$w coords $item] "%f %f" x1 y1
    set distance [expr sqrt(($x1 - $x) * ($x1 - $x) + ($y1 - $y) * ($y1 - $y))]
    if {$distance < $record} {
      set record $distance
      set result $item
    }
  }
  
  return $result

}

## ----------------------------------------------------------------------------
## Command bound to B1-Motion-event on a edge.
##
## Moving edges around on the canvas is implemented by setting the
## private variable _dragItem to the edge that is clicked on, moving
## this edge when the mouse is dragged, and resetting the variable
## when the mouse is released.
##
## Ordinarily one would use Tcl's automatically managed canvas tag
## `current' for this; but an edge may be selected not by clicking on
## it, but by clicking on the canvas background in its vicinity, and
## Tcl is not smart enough to assign the `current' tag to an item that
## was moved under the mouse pointer after the click occurred. It also
## doesn't allow setting the `current' tag manually, so we have to
## duplicate its functionality. 
## ----------------------------------------------------------------------------
body ParseTree::edgeDrag {w x y} {
  if {"" == $_dragItem} return
   scan [$w coords $_dragItem] "%f %f %f %f" x1 y1 x2 y2
   scan [fixCoords $w $x $y] "%d %d" new_x new_y
   moveEdge $w $x1 $y1 $new_x $new_y
   set releaseItem [closestNode $w $new_x $new_y]
   $w itemconfigure node -fill black
   $w itemconfigure $releaseItem -fill green
}

## ----------------------------------------------------------------------
## Handler for the unusual event that the user wants to pan while
## holding an edge; this is actually useful if an edge must be moved
## between two attachment points that do not fit on the canvas
## simultaneously.
## ----------------------------------------------------------------------
body ParseTree::edgeDragPan {w x y} {
  edgeDrag $w $x $y
  multiDragto $x $y
}


## ----------------------------------------------------------------------
## Cancel an ongoing move operation, i.e. return the edge to where it
## was before it was picked up. Completed moves are not undone, for
## that you need undo/redo.
## ----------------------------------------------------------------------
body ParseTree::edgeMoveCancel {w} {
  if {"" == $_dragItem} return
  scan [$w coords $_dragItem] "%f %f %f %f" x1 y1 x2 y2
  moveEdge $w $x1 $y1 $_dragStartX $_dragStartY
  $w itemconfigure $_dragItem -fill black
  $w itemconfigure node -fill black
  set _dragItem ""
}


## ----------------------------------------------------------------------------
## command bound to B1-Release-event on a edge.
## ----------------------------------------------------------------------------
body ParseTree::edgeDrop {w x y} {
  if {$_dragItem == "" } return

  scan [fixCoords $w $x $y] "%d %d" new_x new_y
  set releaseItem [closestNode $w $new_x $new_y]
  scan [$w coords $_dragItem] "%f %f %f %f" x1 y1 x2 y2
  scan [$w coords $releaseItem] "%f %f" x2 y2
  if {$releaseItem != "" && !($x2 == $_dragStartX && $y2 == $_dragStartY)} {
  
      # this is a correct manipulation
      moveEdge $w $x1 $y1 [expr $x2 + 2] [expr $y2 + 2]
      set level $_levelOfCanvas($w)
      set modifier $_timepointOfId($level,$_dragItem)
      set modifiee $_timepointOfId($level,$releaseItem)
      set binding [$_parse getBindingAt $level $modifier]
      scan $binding "%s %s %s %s %s" modifier oldmodifiee label where level

      set oldBinding [$_parse getBindings $level]

      $_parse shiftEdge $level $modifier $modifiee
      reDrawLevel $level

      ## if bindings changed, save commands to history
      if { $oldBinding != [$_parse getBindings $level]} {
	  $_commandHistory add\
	      [code "$this undoEdge $modifier $modifiee $level"]\
	      [code "$this undoEdge $modifier $oldmodifiee $level"]
      } 

      # update arcs on main level if necessary
      drawArcs
      drawViolas
      paintTabs

  } else {

      moveEdge $w $x1 $y1 $_dragStartX $_dragStartY
      $w itemconfigure $_dragItem -fill black
  }
  set _dragItem ""
}

## ----------------------------------------------------------------------------
## draw violations in the list.
## ---------------------------------------------------------------------------
body ParseTree::drawViolas {} {

  if {[$_parse isAbstract]} return

  set table $itk_component(conflicts)
  $table configure -state normal
  $table clear all
  $table delete rows 0 [$table cget -rows]
  $table tag config sel -fg "" -bg "" -relief raised -bd 1
  $table tag config justRight -anchor e
  $table tag config title -anchor w -background gray90 -foreground black -relief flat
  $table tag col justRight 2 4
  $table width 0 25 1 15 2 10 4 10

  # write the violation rows
  set no 0
  set violas [$_parse getViolations]
  $table configure -rows [expr [llength $violas] +1]
  foreach viola $violas {
    # build viola_text
    set rowString [list [lindex $viola 0]] ;# name
    lappend rowString [format "%4.3e" [lindex $viola 3]] ;# score
    set b1Id [lindex $viola 1]
    set b2Id [lindex $viola 2]

    # find the corresponding binding's id:
    set binding [$_parse getBindingById $b1Id]
    scan $binding "%s %s %s %s %s" modifier_id1 modifiee_id1 label bindNo1 level1

    set wordstruct [$_parse getWord $modifier_id1]
    set word1 [WordStruct_word_get $wordstruct]
    if {$modifiee_id1 == -1} {
	set word2 NIL
    } else {
      set wordstruct [$_parse getWord $modifiee_id1]
      set word2 [WordStruct_word_get $wordstruct]
    }
    lappend rowString "$level1:" "$word1-$word2"

    # more work for binary violations
    set bindNo2 ""
    if {$b2Id != -1} {

      # find the corresponding binding's id:
      set binding [$_parse getBindingById $b2Id]
      scan $binding "%s %s %s %s %s" modifier_id2 modifiee_id2 \
	  label bindNo2 level2
      set wordstruct [$_parse getWord $modifier_id2]
      set word1 [WordStruct_word_get $wordstruct]
      if {$modifiee_id2 == -1} {
	set word2 NIL
      } else {
	set wordstruct [$_parse getWord $modifiee_id2]
        set word2 [WordStruct_word_get $wordstruct]
      }
      lappend rowString "$level2:" "$word1-$word2"
    }

    # draw violation into the list
    $table setCell row $no,0 $rowString

    set _itemsOfViolation($no) ""
    incr no
  }

  set levels [$_parse getLevels]
  foreach level $levels {
    registerItems $level
  }
    
  $table configure -state disabled

  # notify daddy
  $_visparses getBadness
}

## ----------------------------------------------------------------------------
## draw the sentence to the canvas.
## ----------------------------------------------------------------------------
body ParseTree::drawText {canvas x y} {

   set xx $x
   set level $_levelOfCanvas($canvas)

   set vector [$_parse getWords]
   for {set i 0} {$i < [vectorSize $vector]} {incr i} {
     set wordstruct [vectorElement $vector $i]
     set word [WordStruct_word_get $wordstruct]

     set word_id [$canvas create text $xx $y \
       -text $word \
       -anchor nw \
       -font $itk_option(-wordfont) \
       -fill $itk_option(-wordcolor) \
                      -tags {word text}]

     scan [$canvas bbox $word_id] "%d %d %d %d" x1 y1 x2 y2
     set width [expr $x2 - $x1]
     set _textMidPos($i) [expr $width / 2.0 + $xx]
     set xx [expr $xx + $width + 20]

     # register all ids into _timepointOfId and _itemsOfTimepoint
     set _timepointOfId($level,$word_id) $i
     lappend _itemsOfTimepoint($level,$i) $word_id
   }
}

## ----------------------------------------------------------------------------
## move an existing edge and its label to a new position.
## ----------------------------------------------------------------------------
body ParseTree::moveLabel {w label_id x1 y1 x2 y2} {
   set dx [expr ($x2 - $x1) / 2.0]
   set dy [expr ($y2 - $y1) / 2.0]

   set best_cross  4 
   #how bad is our best solution

   if {$dx == 0.0 && $dy == 0.0} {
      $w coords $label_id $x1 $x2
   } else {
      #the search starting point - a little down the road
      set mx [expr $x1 + ($dx * 0.9)]
      set my [expr $y1 + ($dy * 0.9)]
      
      if {[expr abs($dx)] > [expr abs($dy)]} {
         #seek allong the x dimension
         set delta [expr $dy/$dx]
         if {$delta < 0.0} {
            set anchor1 se
            set anchor2 nw
         } else {
            set anchor1 sw
            set anchor2 ne
         }
         #preferences in worst case 
         set best_anchor $anchor1
         set best_ix [expr -($dx * 0.7)]

         $w itemconfigure $label_id -anchor $anchor1   
         for {set ix 0} {$ix < [expr abs($dx)]} {incr ix} {
            #up
            $w coords $label_id [expr $mx + $ix] [expr $my + ($ix * $delta)]
            set cross [llength [eval $w find overlap [$w bbox $label_id]]]
            if { $cross < $best_cross} {
               set best_anchor $anchor1
               set best_ix $ix
               if {$cross < 3} {
                  break 
               }
               set best_cross $cross
            }
            #another anchor
            $w itemconfigure $label_id -anchor $anchor2
            set cross [llength [eval $w find overlap [$w bbox $label_id]]]
            if { $cross < $best_cross} {
               set best_anchor $anchor2
               set best_ix $ix
               if {$cross < 3} {
                  break 
               }
               set best_cross $cross
            }
            #down
            $w coords $label_id [expr $mx - $ix] [expr $my - ($ix * $delta)]
            set cross [llength [eval $w find overlap [$w bbox $label_id]]]
            if { $cross < $best_cross} {
               set best_anchor $anchor2
               set best_ix [expr -$ix]
               if {$cross < 3} {
                  break 
               }
               set best_cross $cross
            }
            #another anchor
            $w itemconfigure $label_id -anchor $anchor1
            set cross [llength [eval $w find overlap [$w bbox $label_id]]]
            if { $cross < $best_cross} {
               set best_anchor $anchor1
               set best_ix [expr -$ix]
               if {$cross < 3} {
                  break 
               }
               set best_cross $cross
            }
         }
         $w coords $label_id [expr $mx + $best_ix] [expr $my + ($best_ix * $delta)]
         $w itemconfigure $label_id -anchor $best_anchor
      } else { 
         #seek allong the y dimension

         set delta [expr $dx/$dy]
         if {$delta < 0} {
            set anchor1 se
            set anchor2 nw
         } else {
            set anchor1 sw
            set anchor2 ne
         }

         set best_anchor $anchor1
         set best_iy [expr -($dy * 0.7)]

         $w itemconfigure $label_id -anchor $anchor1
         for {set iy 0} {$iy < [expr abs($dy)]} {incr iy} {
            #up
            $w coords $label_id [expr $mx + ($iy * $delta)] [expr $my + $iy]
            set cross [llength [eval $w find overlap [$w bbox $label_id]]]
            if { $cross < $best_cross} {
               set best_anchor $anchor1
               set best_iy $iy
               if {$cross < 3} {
                  break 
               }
               set best_cross $cross
            }
            $w itemconfigure $label_id -anchor $anchor2
            set cross [llength [eval $w find overlap [$w bbox $label_id]]]
            if { $cross < $best_cross} {
               set best_anchor $anchor2
               set best_iy $iy
               if {$cross < 3} {
                  break 
               }
               set best_cross $cross
            }
            #down
            $w coords $label_id [expr $mx - ($iy * $delta)] [expr $my - $iy]
            set cross [llength [eval $w find overlap [$w bbox $label_id]]]
            if { $cross < $best_cross} {
               set best_anchor $anchor2
               set best_iy [expr -$iy]
               if {$cross < 3} {
                  break 
               }
               set best_cross $cross
            }
            $w itemconfigure $label_id -anchor $anchor1
            set cross [llength [eval $w find overlap [$w bbox $label_id]]]
            if { $cross < $best_cross} {
               set best_anchor $anchor1
               set best_iy [expr -$iy]
               if {$cross < 3} {
                  break 
               }
               set best_cross $cross
            }
         }
         #set the best solution
         $w coords $label_id [expr $mx + ($best_iy * $delta)] [expr $my + $best_iy]
         $w itemconfigure $label_id -anchor $best_anchor
      }
   }
}   

## ----------------------------------------------------------------------------
## Find the label corresponding to an edge
## ----------------------------------------------------------------------------
body ParseTree::findLabel {w e} {
   set level $_levelOfCanvas($w)
   set timepoint $_timepointOfId($level,$e)
   set ids $_itemsOfTimepoint($level,$timepoint)
   set found 0
   foreach label $ids {
      if {[lsearch -exact [$w gettags $label] "label"]>=0} {
         return $label
      }
   }
}


## ----------------------------------------------------------------------------
## move an existing edge and its label to a new position.
## ----------------------------------------------------------------------------
body ParseTree::moveEdge {w x1 y1 x2 y2} { 
   $w coords $_dragItem $x1 $y1 $x2 $y2
   set level $_levelOfCanvas($w)
   set timepoint $_timepointOfId($level,$_dragItem)
   set ids $_itemsOfTimepoint($level,$timepoint)
   set found 0
   foreach label_id $ids {
      if {[lsearch -exact [$w gettags $label_id] "label"]>=0} {
         moveLabel $w $label_id $x1 $y1 $x2 $y2
         break
      }
   }
}

## ----------------------------------------------------------------------------
## checks if all nodes between id1 and id2 are modifier of id2.
## ----------------------------------------------------------------------------
body ParseTree::allSybillings {id1 id2 level} {
  set children [$_parse getModifiers $level $id2]

  if {$id2 < $id1} {
    set tmp $id1
    set id1 $id2
    set id2 $tmp
  }

  for {set i [expr $id1 + 1]} {$i < $id2} {incr i} {
    if {[lsearch -exact $children $i] < 0} {
      return 0
    }
  }

  return 1
}

## ----------------------------------------------------------------------------
## seeks an edge crossing with this id.
## ----------------------------------------------------------------------------
body ParseTree::noCrossing {id bindings} {
   foreach binding $bindings {
      set modifiee_id [lindex $binding 1]
      if {$modifiee_id == -1} continue
      set modifier_id [lindex $binding 0]
      if {$modifiee_id > $id && $modifier_id < $id} {
         return 0 }
      if {$modifier_id > $id && $modifiee_id < $id} {
         return 0 }
   }
   return 1
}


## ----------------------------------------------------------------------------
## force draw a specified level on its page.
## ----------------------------------------------------------------------------
body ParseTree::reDrawLevel {level} {
   set _isDrawn($level) 0
   drawLevel $level
}

## ----------------------------------------------------------------------------
## draw a specified level on its page.
## ----------------------------------------------------------------------------
body ParseTree::drawLevel {level} {

  # skip already drawn levels
  if { $_isDrawn($level) } {
    return
  } else {
    set _isDrawn($level) 1
  }

  # delete old drawing
  set canvas [$itk_component($level) childsite]
  array unset _timepointOfId "$level,*"
  array unset _itemsOfTimepoint "$level,*"

  $canvas delete all
  makeUserInterface $canvas
  set lapsTime [clock clicks -milliseconds]
  set heightInLinks [$_parse getHeight $level]
  set heightInPixels [expr $heightInLinks * 40.0 + 40]
  set lowerrow 0
  set middlerow -20
  set upperrow -60

  # draw the words
  drawText $canvas 10 [expr $heightInPixels + 5.0]

  # draw bindings of level
  set bindings [$_parse getBindings $level]

   foreach binding $bindings {
     scan $binding "%s %s %s %s %s" \
    	modifier_id modifiee_id label dummy dummy
     set label [join $label]

     if {$heightInLinks == -1} {
      # draw cyclical structures as bipartite graph

       set modifier_x $_textMidPos($modifier_id)
       set modifier_y $lowerrow

       # draw lower node
       set node_id [$canvas create oval \
                        [expr $modifier_x-2] [expr $modifier_y-2] \
                        [expr $modifier_x+2] [expr $modifier_y+2] \
                        -fill $itk_option(-nodecolor) \
                        -tags node]
       $canvas raise $node_id
       set _timepointOfId($level,$node_id) $modifier_id
       lappend _itemsOfTimepoint($level,$modifier_id) $node_id

       if {$modifiee_id == -1 } {

         # draw NIL node
         set nil_x $modifier_x
         set nil_y $middlerow
         set node_id [$canvas create oval \
                          [expr $nil_x-2] [expr $nil_y-2] \
                          [expr $nil_x+2] [expr $nil_y+2] \
                          -fill $itk_option(-nodecolor) \
                          -tags node ]
         $canvas raise $node_id
         set _timepointOfId($level,$node_id) -1
         lappend _itemsOfTimepoint($level,-1) $node_id
         set modifiee_x $nil_x
         set modifiee_y $nil_y
       } else {

         # draw upper node
         set modifiee_y $upperrow
         set modifiee_x $_textMidPos($modifiee_id)
         set node_id [$canvas create oval \
                          [expr $modifiee_x-2] [expr $modifiee_y-2] \
                          [expr $modifiee_x+2] [expr $modifiee_y+2] \
                          -fill $itk_option(-nodecolor) \
                          -tags node ]
         $canvas raise $node_id
         set _timepointOfId($level,$node_id) $modifiee_id
         lappend _itemsOfTimepoint($level,$modifiee_id) $node_id
       } 
       # Herein lurks a subtle bug. If none of the bindings of this level
       # are NIL bindings, then no NIL node is ever drawn, and no mouse
       # gesture can create a NIL binding. But this can only happen if
       # *all* of the bindings are part of cycles. Unlikely.
       
       # draw edge
       set edge_id [$canvas create line 0 0 0 0 \
                        -fill $itk_option(-edgecolor) \
                        -tags edge ]
       
       $canvas lower $edge_id
       set label_id [$canvas create text 0 0 \
                         -text "$label" \
                         -anchor sw \
                         -font $itk_option(-labelfont) \
                         -fill $itk_option(-labelcolor) \
                         -tags {label text}]
       $canvas lower $label_id
       set _timepointOfId($level,$label_id) $modifier_id
       lappend _itemsOfTimepoint($level,$modifier_id) $label_id
       set edges($label_id) $edge_id
       
       # register items and ids in _itemsOfTimepoint and _timepointOfId
       set _timepointOfId($level,$edge_id) $modifier_id
       lappend _itemsOfTimepoint($level,$modifier_id) $edge_id
       
       set _dragItem $edge_id
       moveEdge $canvas $modifier_x $modifier_y $modifiee_x $modifiee_y 
       set _dragItem ""

    } else {
      # draw non-cyclical structure

      set modifier_height [$_parse getWordHeight $level $modifier_id] 
      set modifier_x $_textMidPos($modifier_id)
      set modifier_y [expr $modifier_height * 40.0 + 40.0]
      
     if {$modifiee_id != -1 } {
        # non root arc
        if {[$_parse isLeafNode $level $modifier_id] &&
	    [allSybillings $modifier_id $modifiee_id $level] } {
           #set the node on the bottom
           set modifier_y $heightInPixels
        }
           
        set modifiee_height [$_parse getWordHeight $level $modifiee_id] 
        set modifiee_x $_textMidPos($modifiee_id)
        set modifiee_y [expr $modifiee_height * 40.0 + 40.0]
    
     } else {
	# root arc
        set modifiee_x $modifier_x
        if {[$_parse isLeafNode $level $modifier_id]} {
           #set the nil node on the bottom
           set modifier_y $heightInPixels
        }
        set modifiee_y [expr $modifier_y - 20.0]

       # draw nil-node
       set node_id [$canvas create oval \
	 	   [expr $modifiee_x-2] [expr $modifiee_y-2] \
		   [expr $modifiee_x+2] [expr $modifiee_y+2] \
		     -fill $itk_option(-nodecolor) \
		     -tags node ]
       $canvas raise $node_id

        # register all ids into _timepointOfId and _itemsOfTimepoint
        set _timepointOfId($level,$node_id) $modifiee_id
        lappend _itemsOfTimepoint($level,$modifiee_id) $node_id
      }

     # draw edge between node and word
     set vline_id [$canvas create line \
		  $modifier_x $modifier_y \
 		  $modifier_x $heightInPixels\
	 	  -fill $itk_option(-vlinecolor) \
		  -tags vline]

     $canvas lower $vline_id

     # draw node
     set node_id [$canvas create oval \
        	 [expr $modifier_x-2] [expr $modifier_y-2] \
		 [expr $modifier_x+2] [expr $modifier_y+2] \
		 -fill $itk_option(-nodecolor) \
		 -tags node]
     $canvas raise $node_id

     # draw edge between modifier and modifiee
      set edge_id [$canvas create line 0 0 0 0 \
                          -fill $itk_option(-edgecolor) \
                          -tags edge ]
         

      $canvas lower $edge_id
      set label_id [$canvas create text 0 0 \
                        -text "$label" \
                        -anchor sw \
                        -font $itk_option(-labelfont) \
                        -fill $itk_option(-labelcolor) \
                        -tags {label text}]
      $canvas lower $label_id
      set _timepointOfId($level,$label_id) $modifier_id
      lappend _itemsOfTimepoint($level,$modifier_id) $label_id
      set edges($label_id) $edge_id

      # register items and ids in _itemsOfTimepoint and _timepointOfId
      set _timepointOfId($level,$edge_id) $modifier_id
      set _timepointOfId($level,$vline_id) $modifier_id
      set _timepointOfId($level,$node_id) $modifier_id
      lappend _itemsOfTimepoint($level,$modifier_id) $edge_id
      lappend _itemsOfTimepoint($level,$modifier_id) $vline_id
      lappend _itemsOfTimepoint($level,$modifier_id) $node_id
      

      set _dragItem $edge_id 
      moveEdge $canvas $modifier_x $modifier_y $modifiee_x $modifiee_y 
      set _dragItem ""
    }
  }

  # move the labels to the correct position
  foreach label_id [array names edges] {
    eval moveLabel $canvas $label_id [$canvas coords $edges($label_id)]
  }

  if {$level == [mainlevelname]} {
    drawArcs 
  }

  $itk_component($level) configure -autoresize 1
  $itk_component($level) yview moveto 0

  registerItems $level
}

## ----------------------------------------------------------------------------
## Compute the array _itemsOfViolation for this level.
## ----------------------------------------------------------------------------
body ParseTree::registerItems {level} {
  if {!$_isDrawn($level)} {
    return
  }

  set mainlevel [mainlevelname]
  set no 0
  set canvas [$itk_component($level) childsite]

  foreach viola [$_parse getViolations] {
    set b1Id [lindex $viola 1]
    set b2Id [lindex $viola 2]

    # find the corresponding binding's id:
    set binding [$_parse getBindingById $b1Id]
    scan $binding "%s %s %s %s %s" \
         modifier_id1 modifiee_id1 label bindNo1 level1

    if {$level == $level1} {
      # register words, edges and labels for the first LV
      if {$modifiee_id1 != -1} {
	foreach id $_itemsOfTimepoint($level,$modifiee_id1) {
	  set id_tags [$canvas gettags $id]
	  if {[regexp "word" $id_tags]} {
	    lappend _itemsOfViolation($no) $id $level
	    break
	  }
	}
      }
      foreach id $_itemsOfTimepoint($level,$modifier_id1) {
        set id_tags [$canvas gettags $id]
        if {[regexp "edge|label|word" $id_tags]} {
          lappend _itemsOfViolation($no) $id $level
        }
      }
    }

    if {$level == [mainlevelname]} {
      # register arcs and labels for the first LV
      catch {
	foreach id $_arcOfTimepoint($level1,$modifier_id1) {
	  lappend _itemsOfViolation($no) $id $level
	}
      }
    }

    # more work for binary violations
    if {$b2Id != -1} {

      # find the corresponding binding's id:
      set binding [$_parse getBindingById $b2Id]
      scan $binding "%s %s %s %s %s" modifier_id2 modifiee_id2 \
          label bindNo2 level2

      if {$level == $level2} {
        # register edges and labels for the first LV
	if {$modifiee_id2 != -1} {
	  foreach id $_itemsOfTimepoint($level,$modifiee_id2) {
	    set id_tags [$canvas gettags $id]
	    if {[regexp "word" $id_tags]} {
	      lappend _itemsOfViolation($no) $id $level
	      break
	    }
	  }
	}
	foreach id $_itemsOfTimepoint($level,$modifier_id2) {
          set id_tags [$canvas gettags $id]
          if {[regexp "edge|label|word" $id_tags]} {
            lappend _itemsOfViolation($no) $id $level
          }
        }
      }

      if {$level == [mainlevelname]} {
      # register arcs and labels for the second LV
	catch {
	  foreach id $_arcOfTimepoint($level2,$modifier_id2) {
	    lappend _itemsOfViolation($no) $id $level
	  }
	}
      }
    }
    incr no
  }
}

## ----------------------------------------------------------------------------
## draw everything into the levels.
## ----------------------------------------------------------------------------
body ParseTree::drawAll {} {
  set _zoomstep 0
  foreach level [$_parse getLevels] {
    reDrawLevel $level
  }
  $itk_component(conflicts) selection clear all
  paintTabs
}

## ----------------------------------------------------------------------------
## colorize the tabs in the tabnotebook.
## This is done according to the worst violation on that tab.
## ----------------------------------------------------------------------------
body ParseTree::paintTabs {} {

  if {[$_parse isAbstract]} return

  set level_index [$itk_component(tbn) index select]
  if {$level_index < 0} {
    set level_index 0
  }

  set tabset [$itk_component(tbn) component tabset]
  $tabset configure -foreground black -selectforeground black

  # find the worst score of a violation for each level
  foreach level [$_parse getLevels] {
    set levelPenalty($level) 1.0
  }

  foreach viola [$_parse getViolations] {
    set b1Id [lindex $viola 1]
    set b2Id [lindex $viola 2]
    set score [lindex $viola 3]
    set binding [$_parse getBindingById $b1Id]
    scan $binding "%s %s %s %s %s" foo bar baz quux level1

    if {$score < $levelPenalty($level1)} {
      set levelPenalty($level1) $score
    }
    if {$b2Id != -1} {

      # find the corresponding binding's id:
      set binding [$_parse getBindingById $b2Id]
      scan $binding "%s %s %s %s %s" foo bar baz quux level2
      if {$score < $levelPenalty($level2)} {
        set levelPenalty($level2) $score
      }
    }
  }

  # paint the tabs in the tabnotebook green or red accordingly
  foreach level [$_parse getLevels] {

    # get a severity value from parse
    set severity $levelPenalty($level)
    set color [rgValue $severity]

    $tabset tabconfigure [$tabset index $level] \
        -foreground $color \
        -selectforeground $color 
  }
  $itk_component(tbn) view $level_index
}

## ----------------------------------------------------------------------------
## take a score and return an appropriate red/green color for it.
## ----------------------------------------------------------------------------
body ParseTree::rgValue {score} {

  # non-linear mapping to intensity values because the intensities
  # below 0.4 are pretty much indistinguishable
  # (try to find a closed form for this if you want)
  if {$score == 1.0} {
    set intensity 0.0
  } elseif {$score > 0.999} {
    set intensity 0.1 
  } elseif {$score > 0.99} {
    set intensity 0.2 
  } elseif {$score > 0.9} {
    set intensity 0.3 
  } elseif {$score > 0.8} {
    set intensity 0.4 
  } elseif {$score > 0.4} {
    set intensity 0.5 
  } elseif {$score > 0.2} {
    set intensity 0.6 
  } elseif {$score > 0.1} {
    set intensity 0.7 
  } elseif {$score > 0.01} {
    set intensity 0.8 
  } elseif {$score > 0.001} {
    set intensity 0.9 
  } else {
    set intensity 1.0
  }

  set redvalue [format "%.0f" [expr 4096 * $intensity]]
  if {$redvalue > 4095} { set redvalue 4095 }
  set greenvalue [format "%.0f" [expr (4095 - $redvalue) / 1.5]]
  set color [format "#%03x%03x%03x" $redvalue $greenvalue 0]
  return $color
}

## ----------------------------------------------------------------------------
## saves the Parse to file as an annotation in cdgp input format
## ----------------------------------------------------------------------------
body ParseTree::writeToFile {fileName} {

  writeAnnotation [$_parse toAnno] $fileName

}

## ----------------------------------------------------------------------------
## option -labelfont
## ----------------------------------------------------------------------------
configbody ParseTree::labelfont {

   foreach level [$_parse getLevels] {
     if {$_isDrawn($level)} {
       set canvas [$itk_component($level) childsite]
       $canvas itemconfigure label -font $itk_option(-labelfont)
     }
   }
}


## ----------------------------------------------------------------------------
## option -wordfont
## ----------------------------------------------------------------------------
configbody ParseTree::wordfont {

    foreach level [$_parse getLevels] {
     if {$_isDrawn($level)} {
       set canvas [$itk_component($level) childsite]
       $canvas itemconfigure word -font $itk_option(-wordfont)
       $canvas itemconfigure title -font $itk_option(-wordfont)
       $canvas itemconfigure viola -font $itk_option(-wordfont)
     }
   }
}

## ----------------------------------------------------------------------------
## compute the screen x and y coords of canvas-coords x and y.
## ----------------------------------------------------------------------------
body ParseTree::canvasToScreen {canvas x y} {

   set rootx [winfo rootx $canvas]
   set rooty [winfo rooty $canvas]
   scan [$canvas cget -scrollregion] "%d %d %d %d" min_x min_y max_x max_y
   set left_off [lindex [$canvas xview] 0]
   set up_off [lindex [$canvas yview] 0]
   scan [expr $rootx + $x - $min_x - ($max_x - $min_x) * $left_off] "%d" X
   scan [expr $rooty + $y - $min_y - ($max_y - $min_y) * $up_off] "%d" Y

   return [list $X $Y]
}

## ----------------------------------------------------------------------------
## Translate window-based into canvas-based co-ordinates.
##
## If a scrolledcanvas is scrolled to the right, away from the home
## position, and a mouse click occurs in its upper left corner, then
## the bind event will still receive the coordinates (1,1) through its
## %x and %y parameters, even though the canvas pixel the the mouse is
## touching is something like (320,1). (Actually, because of the way
## the scrolledcanvas widget is written, an offset occurs even in the
## home position.) This routine accounts for that offset. Therefore it
## must be called whenever coordinates received through `bind' are to
## be applied to things on the canvas. 
## ----------------------------------------------------------------------------
body ParseTree::fixCoords {w x y} {

   scan [$w cget -scrollregion] "%d %d %d %d" min_x min_y max_x max_y
   set left_off [lindex [$w xview] 0]
   set up_off [lindex [$w yview] 0]
   scan [expr $x + $min_x + ($max_x - $min_x) * $left_off] "%d" new_x
   scan [expr $y + $min_y + ($max_y - $min_y) * $up_off] "%d" new_y

   return [list $new_x $new_y]
}

## ----------------------------------------------------------------------------
## build a popup menu and show the provided items.
## Items is a list of pairs consisting of a label and the asociated command
## ----------------------------------------------------------------------------
body ParseTree::showPopMenu {w x y items} {

  set popmenu $itk_component(popmenu)
  $popmenu delete 0 end

  foreach item $items {
    $popmenu add command \
	-label " [lindex $item 0] " \
	-command [lindex $item 1]
  }
  update idletask

  set screenHeight [winfo screenheight $w]
  set reqHeight [winfo reqheight $popmenu]
  scan [canvasToScreen $w $x $y] "%d %d" screenX screenY
  set diffHeight [expr $screenHeight - $reqHeight -$screenY]
  set noItems [llength $items]

  if {$diffHeight<0} {

    # distribute items on some cascaded menus
    set heightPerItem [expr $reqHeight / $noItems]
    set maxItems [expr ($screenHeight -$screenY) / $heightPerItem - 2]
    set noCascades [expr ceil($noItems / double($maxItems))]
    $popmenu delete $maxItems end
    set lastCascade $popmenu
    set lastItem $maxItems

    for {set i 1} {$i < $noCascades} {incr i} {
      if {![winfo exists $lastCascade.$i]} {
	set newCascade [menu $lastCascade.$i ]
      } else {
	set newCascade $lastCascade.$i
	$newCascade delete 0 end
      }
      $lastCascade add separator
      $lastCascade add cascade \
	  -label " ... " \
	  -menu $newCascade

      for {set j 0} \
	  {$j < $maxItems && [expr $lastItem + $j] < $noItems} \
	  {incr j} {
	set item [lindex $items [expr $lastItem + $j]]
	$newCascade add command \
	    -label " [lindex $item 0] " \
	    -command [lindex $item 1]
      }
      incr lastItem $j
      set lastCascade $newCascade
    }
    update idletask
  }
  eval tk_popup $popmenu $screenX $screenY
}


## ----------------------------------------------------------------------------
## draws all bindings not on the main level as arcs under the main tree.
## ----------------------------------------------------------------------------
body ParseTree::drawArcs {} {

  set mainlevel [mainlevelname]
  set canvas [$itk_component($mainlevel) childsite]

  # retract all arcs
  $canvas delete arc
  array unset _arcOfTimepoint

  set heightInLinks [$_parse getHeight $mainlevel]
  set start [expr $heightInLinks * 40.0 + 80]
  set nrWords [vectorSize [$_parse getWords]]

  # collect bindings in array arcs which are worth drawing
  set arcs ""
  foreach level [$_parse getLevels] {

    # skip main level bindings
    if {$level != $mainlevel} {

      foreach binding [$_parse getBindings $level] {
	scan $binding "%s %s %s %s %s" from to label bindNo foo

	# skip NULL bindings
	if {$to != "-1" } {
	  lappend arcs [list $from $to $label $bindNo $level]
	}
      }
    }
  }

  # Now figure out where to draw the arcs so that they don't overlap.
  # We do it like this: the space below the sentence is segmented into
  # rows and columns, starting with row 0 directly under the sentence
  # and column 0 under the first word. Now when an arc is drawn from a
  # to b, it gets drawn into the first row r with unused columns a
  # through b and then marks cells ($r,$a) through ($r,b) as used.
  
  # init cell array
  set noArcs [llength $arcs]
  for {set row 0} {$row < $noArcs} {incr row} {
    for {set column 0} {$column < $nrWords} {incr column} {
      set cell($row,$column) 0
    }
  }

  # draw all arcs
  foreach arc $arcs {

    # get the arc
    set from [lindex $arc 0]
    set to [lindex $arc 1]
    set label [lindex $arc 2]
    set bindno [lindex $arc 3]
    set level [lindex $arc 4]

    # a = min(from,to), z = max(from,to)
    if {$from > $to} { 
      set a $to
      set z $from 
    } else {
      set a $from
      set z $to
    }
    
    # find first partially free row
    for {set row 0} {$row < $noArcs} {incr row} {
      set rowOK 1
      for {set column $a} {$column < $z} {incr column} {
        if {$cell($row,$column) == 1} {
          set rowOK 0
          break
        }
      }
      if {$rowOK == 1} {
	break
      }
    }

    # mark cells as used
    for {set column $a} {$column < $z} {incr column} {
      set cell($row,$column) 1
    }
   
    set x1 $_textMidPos($from)
    set x2 $_textMidPos($to)
    set y1 [expr $start + $row * 40]
    set y2 [expr $start + (1 + $row) * 40]

    # shift arrowheads so that they never overlap
    if {$from < $to} {
      set x2 [expr $x2 - 10]
    } else {
      set x2 [expr $x2 + 10]
    }

    set text "$level:$label"
    if {$label == "{}"} {
      set text $level
    }

    set items [createArc $canvas $x1 $y1 $x2 $y2 $text $bindNo]
 
    # register items
    foreach item $items {
      set _timepointOfId($level,$item) $from
      lappend _arcOfTimepoint($level,$from) $item
    }

  }
}

## ----------------------------------------------------------------------------
## creates a new arc with the specified parameters.
## ----------------------------------------------------------------------------
body ParseTree::createArc {canvas x1 y1 x2 y2 text bindNo} {

  # arc
  set xmid [expr ($x2 + $x1) / 2]
  set arc [$canvas create line $x1 $y1 $xmid $y2 $x2 $y1 \
             -arrow last \
             -smooth 1 \
             -tags [list arc binding$bindNo]]
         
             #-arrowshape "10 10 6"\

  # label
  set textx [expr $x1 + ($x2 - $x1) / 2]
  set texty [expr $y2 - 5]
  set label [$canvas create text $textx $texty\
                 -text $text\
                 -font $itk_option(-wordfont)\
                 -tags [list text arc binding$bindNo]]

  return "$arc $label"

}

## ----------------------------------------------------------------------------
## scan all levels of the tree horizontally.
## \param x the x position of the pointer
## \param y the x position of the pointer
## ----------------------------------------------------------------------------
body ParseTree::multiMark {x y} {
  set _lastxmark $x
  set _lastymark $y
  foreach level [$_parse getLevels] {
    $itk_component($level) scan mark $x $y
  }
}

## ----------------------------------------------------------------------------
## scan all levels of the tree horizontally.
## \param x the x position of the pointer
## \param y the x position of the pointer
## ----------------------------------------------------------------------------
body ParseTree::multiDragto {x y} {

  set mainlevel [mainlevelname]
  set canvas [$itk_component($mainlevel) childsite]
  if {"[$canvas xview]" == "0 1"} {
    set x $_lastxmark
  }
  if {"[$canvas yview]" == "0 1"} {
    set y $_lastymark
  }

  # Restrict the actual Y coordinate used for scrolling.
  #
  # Since all levels of analysis model the same word sequence (the
  # path criterion), all canvases in a ParseTree must be the same
  # width. But the tree structures created by them can differ; usually
  # the main level with its glorious syntax tree is much higher than
  # any auxiliary levels, where only isolated edges appear. Therefore
  # we cannot use the Y coordinate extracted from the click event
  # uncritically for vertical panning on all canvases. Instead, we
  # check whether each canvas is already completely visible, and if it
  # is, we do not pan vertically at all.
  foreach level [$_parse getLevels] {
    set c $itk_component($level)
    set eff_y $y
    if {"[$c yview]" == "0 1"} {
      set eff_y 0
    }
    $c scan dragto $x $eff_y
  }
}

## ----------------------------------------------------------------------------
## call all level canvases of the tree.
## ----------------------------------------------------------------------------
body ParseTree::multiCanvas {args} {

  foreach level [$_parse getLevels] {
    drawLevel $level
    if {"$args" != ""} {
      eval $itk_component($level) $args
    }
  }
}

## ----------------------------------------------------------------------------
## apply parseMirror.
## ----------------------------------------------------------------------------
body ParseTree::mirror {} {
  $_parse mirror
  
  drawViolas

  foreach level [$_parse getLevels] {
    reDrawLevel $level
  }
  paintTabs

}

## ----------------------------------------------------------------------------
## highlight differences to annotation
## ----------------------------------------------------------------------------
body ParseTree::verify {} {

  # downlight everything
  foreach level [$_parse getLevels] {
    set canvas [$itk_component($level) childsite]
    $canvas itemconfigure arc -fill black
    $canvas itemconfigure edge -fill black
    $canvas itemconfigure word -fill black
    $canvas itemconfigure label -fill black
  }

  set pv [$_parse verify]
  if {$pv == ""} return

  set wrongwords  [ParseVerificationStruct_lexErrors_get $pv]
  set wronglabels [ParseVerificationStruct_labelErrors_get $pv]
  set wrongedges  [ParseVerificationStruct_structErrors_get $pv]

  set mainlevel [mainlevelname]
  set canvas [$itk_component($mainlevel) childsite]

  # highlight wrong words
  foreach id [$canvas find withtag word] {
    set wordno $_timepointOfId($mainlevel,$id)
    if {[vectorElement $wrongwords $wordno] != "NULL"} {
      $canvas itemconfigure $id -fill red
    }
  }

  # highlight wrong labels
  foreach id [$canvas find withtag label] {
    set t $_timepointOfId($mainlevel,$id)
    set i [$_parse indexOf $mainlevel $t]
    if {[vectorElement $wronglabels $i] != "NULL"} {
      $canvas itemconfigure $id -fill red
    }
  }

  # highlight wrong edges
  foreach id [$canvas find withtag edge] {
    set t $_timepointOfId($mainlevel,$id)
    set i [$_parse indexOf $mainlevel $t]
    if {[vectorElement $wrongedges $i] != "NULL"} {
      $canvas itemconfigure $id -fill red
    }
  }
  
  parseVerificationDelete $pv
}

## ----------------------------------------------------------------------------
## highlight all edges participating in cycles on the current level.
## ----------------------------------------------------------------------------
body ParseTree::showcycles {} {

  # process currently shown level
  set level [lindex [$_parse getLevels] [$itk_component(tbn) view]]

  # search bindings for structure
  set bindings [$_parse getBindings $level]
  foreach binding $bindings {
    set down [lindex $binding 0]
    set up [lindex $binding 1]
    set parent($down) $up
  }

  # search structure for cycles
  set max [array size parent]
  for {set i 0} {$i < [array size parent]} {incr i} {
    set cyclic($i) 0
  }

  for {set i 0} {$i < [array size parent]} {incr i} {

    # already got this?
    if { 1 == $cyclic($i) } continue
    
    set counter 0
    set j $i
    while {$j != -1} {
      incr counter
      # found cycle that doesn't involve i?
      # let him live, we'll get to him later
      if {$counter > $max} break

      set j $parent($j)
      # found cycle that involves i?
      if {$j == $i} {
        set cyclic($i) 1
        break
      }
    }
  }

  set canvas [$itk_component($level) childsite]
  set items ""
  foreach id [$canvas find withtag edge] {
    if {1 == $cyclic($_timepointOfId($level,$id))} {
      $canvas itemconfigure $id -fill red
    } else {
      $canvas itemconfigure $id -fill black
    }
  }
}

## ----------------------------------------------------------------------------
## Break cycles in the tree on the current level.
## 
## This may be necessary because a cycle can't be broken
## by drag & drop if there isn't already a root binding.
## ----------------------------------------------------------------------------
body ParseTree::breakcycles {} {

  # process currently shown level
  set level [lindex [$_parse getLevels] [$itk_component(tbn) view]]

  # search bindings for structure
  set bindings [$_parse getBindings $level]
  foreach binding $bindings {
    set down [lindex $binding 0]
    set up [lindex $binding 1]
    set parent($down) $up
  }

  # search structure for cycles
  set max [array size parent]
  for {set i 0} {$i < [array size parent]} {incr i} {
    set cyclic($i) 0
  }

  for {set i 0} {$i < [array size parent]} {incr i} {

    set counter 0
    set j $i
    while {$j != -1} {
      set j $parent($j)
      if {$j != -1} {
        set old $parent($j)
      }
      incr counter
      if {$counter > $max || $j == $i} {

        # found cycle?
        $_parse shiftEdge $level $j -1

        $_commandHistory add\
            [code "$this undoEdge $j -1 $level"]\
            [code "$this undoEdge $j $old $level"]

        reDrawLevel $level
        drawViolas
        paintTabs
        breakcycles
        return

      }
    }
  }

}


## ----------------------------------------------------------------------------
## increase or decrease zoom level of all canvases.
## ----------------------------------------------------------------------------
body ParseTree::zoom {direction} {

  # Only five zoom levels are allowed: -2/-1/0/1/2
  set newzoom [expr $_zoomstep + $direction]
  if {$newzoom > 2} {
    set newzoom 2
  }
  if {$newzoom < -2} {
    set newzoom -2
  }

  _zoom $newzoom

}


## ----------------------------------------------------------------------------
## set zoom level for one canvas.
## ----------------------------------------------------------------------------
body ParseTree::_zoom {step} {

  set old_zoomstep $_zoomstep
  set _zoomstep $step

  if {$old_zoomstep == $_zoomstep} {
    return
  }
  
  # If we got here, each level actually needs to be redrawn.
  # It would be nicer just to rescale the canvases and leave
  # them drawn, but drawLevel{} uses the actual size of the
  # text items to group them implicitly, and that calculation
  # becomes invalid when the font size changes, so it looks better
  # to force a redraw here. 
  # This also breaks the time-saving delayed drawing of levels.
  foreach level [$_parse getLevels] {

    reDrawLevel $level

    # scale canvas
    set canvas [$itk_component($level) childsite]
    array set factors {-2 0.5 -1 0.75 0 1 1 1.33 2 1.75}
    set factor $factors($_zoomstep)
    $canvas scale all 0 0 $factor $factor
    
    # scale text items
    array set fontsizes {-2 5 -1 9 0 10 1 14 2 16}
    set size $fontsizes($_zoomstep)
    $canvas itemconfigure text -font [list Helvetica $size bold]
    $canvas configure -scrollregion [$canvas bbox all]

  }
}

## ----------------------------------------------------------------------------
## delegate to the _parse.
## ----------------------------------------------------------------------------
body ParseTree::parse {args} {
  eval $_parse $args
}

## ----------------------------------------------------------------------------
## set the orientation of the panedwindow and reset to default fraction.
## ----------------------------------------------------------------------------
body ParseTree::setOrientation {orient} {
  if {[$_parse isAbstract]} return
  $itk_component(pane) configure -orient $orient 
  $itk_component(pane) fraction 80 20
}

## ----------------------------------------------------------------------------
## get the orientation of the panedwindow.
## ----------------------------------------------------------------------------
body ParseTree::getOrientation {} {
  return [$itk_component(pane) cget -orient]
}

## ----------------------------------------------------------------------------
## Return he name of the main level.
##
## The main level is the one that gets all other edges drawn under it
## as little arcs. If the grammar does not define one, return the name
## of the first level in the parse.
## ----------------------------------------------------------------------------
body ParseTree::mainlevelname {} {

  set result [.cdgmain levels getMainlevelId]
  if {$result != ""} { return $result }

  return [lindex [$_parse getLevels] 0]

}

## ----------------------------------------------------------------------------
## Highlight specified edges in red color.
## ----------------------------------------------------------------------------
body ParseTree::highlight {which} {

  set levels [$_parse getLevels]
  set nolevels [llength $levels]
  set nowords [$_parse getWidth]

  regsub -all "," $which " " which
  foreach w [split $which] {
    set l [expr $w % $nolevels]
    set t [expr $w / $nolevels]

    # catch improper index
    if {$t > $nowords} continue
    
    # to be highlighted, the edge must first be drawn
    set level [lindex $levels $l]
    drawLevel $level
    set canvas $itk_component($level)
    set ids $_itemsOfTimepoint($level,$t)
    foreach edge $ids {
      if {[lsearch -exact [$canvas gettags $edge] "edge"]>=0} {
        $canvas itemconfigure $edge -fill red
      }
    }
  }
}


## ----------------------------------------------------------------------
## Change reading for given vertice to description
## @param from vertice to change
## @param description new reading
## ----------------------------------------------------------------------
body ParseTree::undoLexeme {from description} {
    $_parse swapWord $from $description
    drawAll
    drawViolas
}


## ----------------------------------------------------------------------
## Change modifier's label at given level
## ---------------------------------------------------------------------- 
body ParseTree::undoLabel {modifier label level} {
  $_parse swapLabel $level $modifier $label
  drawAll
  drawViolas
}



## ----------------------------------------------------------------------
## Shift edge (modifier) to modifiee on given level
## Used to undo edge-moving actions
## ----------------------------------------------------------------------
body ParseTree::undoEdge {modifier modifiee level} {
    $_parse shiftEdge $level $modifier $modifiee
    # update arcs on main level if necessary
    drawAll
    drawViolas
}


## ----------------------------------------------------------------------
## Add a listener to be informed of changes to the command-history
## @param listener Object-ID implementing method update {}
## ----------------------------------------------------------------------
body ParseTree::addUndoListener {listener} {
    if { $_commandHistory != "" } {
	$_commandHistory addListener $listener
    }
}


## ----------------------------------------------------------------------
## Calls undo on command history 
## ----------------------------------------------------------------------
body ParseTree::undo {} {
    if {[$_commandHistory canUndo] == 1} {
	$_commandHistory undo
    }
}

## ----------------------------------------------------------------------
## Calls redo on command history
## ----------------------------------------------------------------------
body ParseTree::redo {} {
    if {[$_commandHistory canRedo] == 1} {
	$_commandHistory redo
    }
}


## ----------------------------------------------------------------------
## Predicate-method rooted to command history
## @return 1 => undo can be executed, 0 => undo can not be executed
## ----------------------------------------------------------------------
body ParseTree::canUndo {} {
    return [$_commandHistory canUndo]
}

## ----------------------------------------------------------------------
## Predicate-method rooted to command history
## @return 1 => undo can be executed, 0 => undo can not be executed
## ----------------------------------------------------------------------
body ParseTree::canRedo {} {
    return [$_commandHistory canRedo]
}
