# Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## VisParses - the grafical tree editor.
## 
## A VisParses object instanciate one toplevel window that embedds a set
## of ParseTree objects to be displayed. There are two common situations that
## determine the set of parses: (1) displaying annotations, (2) displaying
## parser solutions. In the first case one VisParses object can fetch all
## annotations loaded into the system. In the second case one VisParses object
## displays all solutions a specific parsing flavour has found for one wordgraph.
## So in the latter case another parser cannot reuse the same VisParses object
## to visualize its solution record also, but needs a new VisParses object of
## its own, giving you two separate toplevel windows. I think this is the best
## compromise here between lots of toplevel windows and embedded objects in
## separate tabs of one window.
##
## As you see below we don't provide a separate destructor to destruct
## embedded ParseTree objects but add ParseTree objects directly as itk-components
## to the VisParses widget. So they get destructed automatically.
##
## \author Michael Daum, Kilian A. Foth, Dietmar Fnning 
## (see also AUTHORS and THANKS for more)
##
## $Id visparses.tcl $
## ----------------------------------------------------------------------------
class VisParses {
   inherit itk::Toplevel

   itk_option define -iconname iconName IconName "xcdg"
   itk_option define -interactive interactive Interactive 0

   public method addParse {parse}; ## \type ParseTree
   public method highlight {parse which}
   public method removeParse {parse}; ## \type ParseTree
   public method view {no}; ## \type TclNumber
   public method center {}
   public method readWord {}
   public method getBadness {}
   public method update {}

   constructor {args} {}; ## \type TclList


   private method _print {}
   private method _loadAction {}
   private method _saveAction {}
   private method _saveAsAction {}
   private method _mirrorAction {}
   private method _verifyAction {}
   private method _showcycleAction {}
   private method _breakcycleAction {}
   private method _deleteParseAction {}
   private method _nextAction {}
   private method _previousAction {}
   private method _refresh {}
   private method _undo {} ; # mapped to edit-menu
   private method _redo {} ; # mapped to edit-menu
   private method _computeViolas {}
   private method _labelOfIndex {index}; # \type TclNumber
   private method _spaceHandler {}
   private method _BackspaceHandler {}
   private method _zoomIn {}
   private method _zoomOut {}
   private method _setOrientation {}
   private method _getOrientation {}
   private method _updateVisual {}

   private variable _noPages 0
   private variable _userInput ""
   private variable _currentOrient "vertical"

  ## array mapping Parses to page-numbers
   private variable _pageOfParse  ;## \type TclArray
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body VisParses::constructor {args} {
  
  global queue env
  set queue ""

   wm withdraw $itk_interior

   itk_component add menuframe {
      frame $itk_interior.menuframe
   } {
      keep -background -cursor
      rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
   }

   # file menu
   itk_component add filebutton {
      menubutton $itk_component(menuframe).filebutton \
		-text "File" \
		-menu $itk_component(menuframe).filebutton.menu
   } {
      keep -background -cursor
      rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
   }

   itk_component add filemenu {
      menu $itk_component(filebutton).menu 
   } {
      keep -background -cursor
      rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
   }


   $itk_component(filemenu) add command \
		-label "Load ..." \
		-underline 0 \
		-command [code $this _loadAction]
   $itk_component(filemenu) add command \
		-label "Save" \
		-underline 0 \
                -command [code $this _saveAction]
   $itk_component(filemenu) add command \
		-label "Save as..." \
		-underline 0 \
                -command [code $this _saveAsAction]
   $itk_component(filemenu) add command \
		-label "Print ..." \
		-underline 0 \
		-command [code $this _print]
   $itk_component(filemenu) add separator
   $itk_component(filemenu) add command \
		-label "Close" \
		-underline 0 \
		-command [list destroy $itk_interior]
   $itk_component(filemenu) add command \
		-label "Quit" \
		-underline 0 \
		-command ::cmd::Quit

   # editmenu
   itk_component add editbutton {
      menubutton $itk_component(menuframe).editbutton \
		-text "Edit" \
		-menu $itk_component(menuframe).editbutton.menu
   } {
      keep -background -cursor
      rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
   }

   itk_component add editmenu {
      menu $itk_component(editbutton).menu 
   } {
      keep -background -cursor
      rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
   }

  ## Configure Edit-Menu with Undo-/Redo-Entries
  $itk_component(editmenu) add command \
                -label "Undo" \
      -command [code $this _undo]\
      -accelerator <Ctrl-z>


  $itk_component(editmenu) add command \
                -label "Redo" \
      -command [code $this _redo]\
      -accelerator <Ctrl-r>

  $itk_component(editmenu) entryconfigure 0 -state disabled
  $itk_component(editmenu) entryconfigure 1 -state disabled

   #$itk_component(editmenu) add separator

   #$itk_component(editmenu) add command \
   		-label "New parse" \
   		-command ".cdgmain not_yet $itk_interior"
   #$itk_component(editmenu) add command \
   		-label "Delete parse" \
		-command [code $this _deleteParseAction]
   #$itk_component(editmenu) add separator
   #$itk_component(editmenu) add command \
		-label "New level" \
		-command ".cdgmain not_yet $itk_interior"
   #$itk_component(editmenu) add command \
		-label "Delete level" \
		-command ".cdgmain not_yet $itk_interior"
   #$itk_component(editmenu) add separator
   #$itk_component(editmenu) add command \
		-label "New Word" \
		-command ".cdgmain not_yet $itk_interior"
   #$itk_component(editmenu) add command \
		-label "Delete Word" \
		-command ".cdgmain not_yet $itk_interior"
   #$itk_component(editmenu) add separator

   # viewmenu
   itk_component add viewbutton {
      menubutton $itk_component(menuframe).viewbutton \
		-text "View" \
		-menu $itk_component(menuframe).viewbutton.menu
   } {
      keep -background -cursor
      rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
   }

   itk_component add viewmenu {
      menu $itk_component(viewbutton).menu 
   } {
      keep -background -cursor
      rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
   }

   itk_component add zoommenu {
      menu $itk_component(viewmenu).zoommenu 
   } {
      keep -background -cursor
      rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
   }

   $itk_component(viewmenu) add command \
		-label "Refresh" \
		-command [code $this _refresh]
   $itk_component(viewmenu) add separator
   $itk_component(viewmenu) add command \
		-label "Zoom in"\
      -command [code $this _zoomIn]
   $itk_component(viewmenu) add command \
		-label "Zoom out"\
               -command [code $this _zoomOut]
  $itk_component(viewmenu) add separator
  $itk_component(viewmenu) add radiobutton \
    -label "Horizontal" \
    -variable [scope _currentOrient] \
    -value "horizontal" \
    -command [code $this _setOrientation]
  $itk_component(viewmenu) add radiobutton \
    -label "Vertical" \
    -variable [scope _currentOrient] \
    -value "vertical" \
    -command [code $this _setOrientation]

   # parsemenu
   itk_component add parsebutton {
      menubutton $itk_component(menuframe).parsebutton \
		-text "Parse" \
		-menu $itk_component(menuframe).parsebutton.menu
   } {
      keep -background -cursor
      rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
   }
   itk_component add parsemenu {
      menu $itk_component(parsebutton).menu 
   } {
      keep -background -cursor
      rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
   }
   $itk_component(parsemenu) add command \
		-label "Mirror" \
		-command [code $this _mirrorAction]
   $itk_component(parsemenu) add command \
		-label "Show cycles" \
		-command [code $this _showcycleAction]
   $itk_component(parsemenu) add command \
		-label "Break cycles" \
		-command [code $this _breakcycleAction]
   $itk_component(parsemenu) add command \
		-label "Verify" \
		-command [code $this _verifyAction]
   $itk_component(parsemenu) add separator
   $itk_component(parsemenu) add command \
		-label "Next" \
		-command [code $this _nextAction]
   $itk_component(parsemenu) add command \
		-label "Previous" \
		-command [code $this _previousAction]
  # toolbar
  itk_component add toolbar {
    frame $itk_interior.toolbar \
      -borderwidth 0 \
      -relief flat
  } {}
  set toolbar $itk_component(toolbar)

  # prev button
  set allImages [image names]
  if {[lsearch -exact $allImages prevImage] < 0} {
    image create photo prevImage -file $env(XCDG_PATH)/back.gif
  }
  itk_component add prevButton {
    button $itk_component(toolbar).prevButton \
      -image prevImage \
      -command [code $this _previousAction] \
      -borderwidth 0 \
      -relief flat \
      -activebackground gray90 \
  } {}
  bind $itk_component(prevButton) <Enter> {%W configure -relief raised}
  bind $itk_component(prevButton) <Leave> {%W configure -relief flat}

  # next button
  if {[lsearch -exact $allImages nextImage] < 0} {
    image create photo nextImage -file $env(XCDG_PATH)/forward.gif
  }
  itk_component add nextButton {
    button $itk_component(toolbar).nextButton \
      -image nextImage \
      -command [code $this _nextAction] \
      -borderwidth 0 \
      -relief flat \
      -activebackground gray90 \
  } {}
  bind $itk_component(nextButton) <Enter> {%W configure -relief raised}
  bind $itk_component(nextButton) <Leave> {%W configure -relief flat}

  # save button
  if {[lsearch -exact $allImages saveImage] < 0} {
    image create photo saveImage -file $env(XCDG_PATH)/floppy.gif
  }
  itk_component add saveButton {
    button $itk_component(toolbar).saveButton \
      -image saveImage \
      -command [code $this _saveAction] \
      -borderwidth 0 \
      -relief flat \
      -activebackground gray90 \
  } {}
  bind $itk_component(saveButton) <Enter> {%W configure -relief raised}
  bind $itk_component(saveButton) <Leave> {%W configure -relief flat}

  # mirror button
  if {[lsearch -exact $allImages mirrorImage] < 0} {
    image create photo mirrorImage -file $env(XCDG_PATH)/mirror.gif
  }
  itk_component add mirrorButton {
    button $itk_component(toolbar).mirrorButton \
      -image mirrorImage \
      -command [code $this _mirrorAction] \
      -borderwidth 0 \
      -relief flat \
      -activebackground gray90 \
  } {}
  bind $itk_component(mirrorButton) <Enter> {%W configure -relief raised}
  bind $itk_component(mirrorButton) <Leave> {%W configure -relief flat}

  # zoom in button
  if {[lsearch -exact $allImages zoomInImage] < 0} {
    image create photo zoomInImage -file $env(XCDG_PATH)/zoom_in.gif
  }
  itk_component add zoomInButton {
    button $itk_component(toolbar).zoomInButton \
      -image zoomInImage \
      -command [code $this _zoomIn] \
      -borderwidth 0 \
      -relief flat \
      -activebackground gray90 \
  } {}
  bind $itk_component(zoomInButton) <Enter> {%W configure -relief raised}
  bind $itk_component(zoomInButton) <Leave> {%W configure -relief flat}

  # zoom out button
  if {[lsearch -exact $allImages zoomOutImage] < 0} {
    image create photo zoomOutImage -file $env(XCDG_PATH)/zoom_out.gif
  }
  itk_component add zoomOutButton {
    button $itk_component(toolbar).zoomOutButton \
      -image zoomOutImage \
      -command [code $this _zoomOut] \
      -borderwidth 0 \
      -relief flat \
      -activebackground gray90 \
  } {}
  bind $itk_component(zoomOutButton) <Enter> {%W configure -relief raised}
  bind $itk_component(zoomOutButton) <Leave> {%W configure -relief flat}

  # id label 
  itk_component add idLabel {
    label $itk_component(toolbar).idLabel \
      -text "Parse " \
      -font {Helvetica -12 bold}
  } {}

  # score label
  itk_component add scoreLabel {
    label $itk_component(toolbar).scoreLabel \
      -text "Score: 00/00/0.0" \
      -font {Helvetica -12 bold}
  } {}
    
  # the tabnotebook containing the ParseTree objects
  itk_component add tbn {
    iwidgets::tabnotebook $itk_interior.tbn -auto true
  } {
    keep -width -height
  }

  # text box for parsing of user input 
  itk_component add textbox {
     iwidgets::entryfield $itk_interior.entry \
         -textvariable [scope _userInput] \
         -labeltext "Type here:"         \
         -labelpos w                     \
       }
  set box [$itk_component(textbox) component entry]

  # Take control of the textbox
  bind $box <space> [code "$this _spaceHandler; break"]
  bind $box <BackSpace> [code "$this _BackspaceHandler; break"]

  bind $itk_interior <q> [list destroy $itk_interior]
  bind $itk_interior <c> [code $this _showcycleAction]
  bind $itk_interior <m> [code $this _mirrorAction]
  bind $itk_interior <v> [code $this _verifyAction]
  bind $itk_interior <n> [code $this _nextAction]
  bind $itk_interior <p> [code $this _previousAction]
  bind $itk_interior <s> [code $this _saveAction]
  bind $itk_interior <Control-z> [code $this _undo]
  bind $itk_interior <Control-r> [code $this _redo]

   # packing
  pack $itk_component(menuframe) -fill x -side top
  pack $itk_component(toolbar) -fill x -side top -padx 5
  pack $itk_component(prevButton) \
       $itk_component(nextButton) \
       $itk_component(saveButton) \
       $itk_component(mirrorButton) \
       $itk_component(zoomInButton) \
       $itk_component(zoomOutButton) \
       -side left -padx 3 
  pack $itk_component(scoreLabel) -side right -padx 3
  pack $itk_component(idLabel) -side right -padx 3

  pack $itk_component(filebutton) \
        $itk_component(editbutton) \
       	$itk_component(viewbutton) \
       	$itk_component(parsebutton) \
       	-side left -padx 5 -pady 5
  pack $itk_component(tbn) -fill both -side top -expand 1



  eval itk_initialize $args

  if {$itk_option(-interactive)} {
    pack $itk_component(textbox) -fill both -side top
  }
   
  wm geometry $itk_interior [wm geometry .]
  wm deiconify $itk_interior
}

## ----------------------------------------------------------------------------
## add a parse on a separate page.
## ----------------------------------------------------------------------------
body VisParses::addParse {parse} {

  set method [$parse getSearchStrategy]
  set lattice [$parse getLatticeId]
  set newid "$method:$lattice"
  configure -title $newid

   # check wether this Parse is already visible
   if {[lsearch -exact [array names _pageOfParse] $parse] != -1 } {
      return
   }

   set tbn $itk_component(tbn)
   incr _noPages
   set page [$tbn add \
    -label "$_noPages" \
    -command [code $this _updateVisual]]

   itk_component add parseTree$_noPages {
      ParseTree $page.parseTree$_noPages $parse $this
   } {
      keep -background -cursor
   }
   pack $page.parseTree$_noPages -fill both -side top -expand 1

  $itk_component(parseTree$_noPages) addUndoListener $this ;# register for updating menu buttons
   set _pageOfParse($parse) $_noPages
}

## ----------------------------------------------------------------------------
## make a ParseTree invisible.
## This does not destroy the underlying Parse.
## ----------------------------------------------------------------------------
body VisParses::removeParse {parse} {

   if {![info exists _pageOfParse($parse)]} {
      error "parse $parse is not visualized here"
   }
   set page $_pageOfParse($parse)
   unset _pageOfParse($parse)
   if {[array names _pageOfParse] == ""} {
      destroy $itk_interior
      return
   }
   set parsetree $itk_component(parseTree$page)
   itk_component delete $parsetree
   $itk_component(tbn) next
   $itk_component(tbn) delete [expr $page -1]
}

## ----------------------------------------------------------------------------
## delegate the view command to the \c tbn component
## ----------------------------------------------------------------------------
body VisParses::view {no} {
   $itk_component(tbn) view $no
}

## ----------------------------------------------------------------------------
## _print the visible level.
## ----------------------------------------------------------------------------
body VisParses::_print {} {

  set parseId "parseTree[_labelOfIndex select]"
  if {$parseId == "parseTree"} {
    return
  }
  set title [$this cget -title]
  regexp {:(.*)$} $title dummy id
  set fileName "[pwd]/$id.ps"

  if {![.cdgmain question "Print to file $fileName?" $itk_interior]} {
    return
  }
  
  set parseTree $itk_component($parseId)
  set mainlevel [$parseTree mainlevelname]
  set synCanvas [[$parseTree component $mainlevel] component canvas]

  scan "[$synCanvas bbox all]" "%f %f %f %f" x1 y1 x2 y2

  $synCanvas postscript -file $fileName -x $x1 -width $x2 -y $y1 -height $y2 \
    -pagewidth 26c -pageheight 18c -rotate 1

  ::cmd::Puts "INFO: written postscript to file '$fileName'\n";
}

## ----------------------------------------------------------------------------
## load a parse directly into the editor.
## \todo this is not implelented yet here, but in AllParses
## ----------------------------------------------------------------------------
body VisParses::_loadAction {} {
   .cdgmain not_yet $itk_interior
}

## ----------------------------------------------------------------------------
## writes the tree to file as an annotation.
## \todo The name of the file to write to is determined by parsing the
## \c -title property. Actually this should be done better by paying a round
## of new properties.
## ----------------------------------------------------------------------------
body VisParses::_saveAction {} {
  set parseId "parseTree[_labelOfIndex select]"
  if {$parseId == "parseTree"} {
    return
  }

  set title [$this cget -title]
  regexp {:(.*)$} $title dummy id
  set fileName "[pwd]/$id.cda"

  set parse $itk_component($parseId)
  $parse writeToFile [$parse parse getId]
}

## ----------------------------------------------------------------------------
## writes the tree to file under a new name.
## ----------------------------------------------------------------------------
body VisParses::_saveAsAction {} {
  set parseId "parseTree[_labelOfIndex select]"
  if {$parseId == "parseTree"} {
    return
  }

  set parse $itk_component($parseId)
  set id [tk_getSaveFile \
	    -parent $itk_component(hull) \
            -title "Save as new file" \
            -defaultextension .cda \
            -initialfile [$parse parse getId]]

  regsub ".cda$" $id "" id

  if {$id != ""} {  
    $parse writeToFile $id 
  }
}


## ----------------------------------------------------------------------------
## call parseMirror() eventually.
## ----------------------------------------------------------------------------
body VisParses::_mirrorAction {} {
  set parseId "parseTree[_labelOfIndex select]"
  if {$parseId == "parseTree"} {
    return
  }
  set parse $itk_component($parseId)
  $parse mirror
}

## ----------------------------------------------------------------------------
## visual version of the `verify' command
## ----------------------------------------------------------------------------
body VisParses::_verifyAction {} {
  set parseId "parseTree[_labelOfIndex select]"
  if {$parseId == "parseTree"} {
    return
  }
  set parse $itk_component($parseId)
  $parse verify
}

## ----------------------------------------------------------------------------
## Display cycles on a level in red ink.
## ----------------------------------------------------------------------------
body VisParses::_showcycleAction {} {
  set parseId "parseTree[_labelOfIndex select]"
  if {$parseId == "parseTree"} {
    return
  }
  set parse $itk_component($parseId)
  $parse showcycles
}

## ----------------------------------------------------------------------------
## Break cycles in a parse.
## ----------------------------------------------------------------------------
body VisParses::_breakcycleAction {} {
  set parseId "parseTree[_labelOfIndex select]"
  if {$parseId == "parseTree"} {
    return
  }
  set parse $itk_component($parseId)
  $parse breakcycles
}

## ----------------------------------------------------------------------------
## switch to the next tab in the tree editor.
## Switching to a next tab means that we select the next tree in a set of 
## trees loaded into the editor. In case of solution parses we simply
## switch to the next parse in the editor. In case of annotation parses
## we switch to the next annotation or generate new annotation parses for the
## next lattice in the set of annotated lattices.
## ----------------------------------------------------------------------------
body VisParses::_nextAction {} {

  if {[$itk_component(tbn) index select] == [$itk_component(tbn) index end] && \
      [regexp {anno2parse:(.*)$} $itk_option(-title) dummy latticeId]} {
    set nextLatticeId [.cdgmain wordgraphs getNext $latticeId] 
    if {$nextLatticeId == ""} {
      return
    }
    set lat [findLattice $nextLatticeId]
    if {$lat == "NULL"} {
      ::cmd::Puts "ERROR: no lattice `$nextLatticeId' found."
      return
    }
    for {set l [findAnnotations $lat 1]} {$l != "NULL" } {set l [listNext $l]} {
      set id [AnnoEntryStruct_id_get [listElement $l]]
      set parse [.cdgmain parses getCData $id]
      addParse $parse
    }
  }
  $itk_component(tbn) next
}

## ----------------------------------------------------------------------------
## display the previous annotation.
## ----------------------------------------------------------------------------
body VisParses::_previousAction {} {
  $itk_component(tbn) prev
}


## ----------------------------------------------------------------------------
## compute the violations of the selected ParseTree.
## ----------------------------------------------------------------------------
body VisParses::_computeViolas {} {
  set parseId "parseTree[_labelOfIndex select]"
  if {$parseId == "parseTree"} {
    return
  }
  $itk_component($parseId) computeViolas
}

## ----------------------------------------------------------------------------
## undo the last action of the selected page.
## ----------------------------------------------------------------------------
body VisParses::_undo {} {
  set parseId "parseTree[_labelOfIndex select]"
  if {$parseId == "parseTree"} {
    return
  }
  $itk_component($parseId) undo
}


## ----------------------------------------------------------------------------
## undo the last action of the selected page.
## ----------------------------------------------------------------------------
body VisParses::_redo {} {
  set parseId "parseTree[_labelOfIndex select]"
  if {$parseId == "parseTree"} {
    return
  }
  $itk_component($parseId) redo
}

## ----------------------------------------------------------------------------
## _refresh the display of the selected VisParses.
## ----------------------------------------------------------------------------
body VisParses::_refresh {} {

  set index [expr [$itk_component(tbn) index select] + 1]
  $itk_component(parseId$index) drawAll
  getBadness
}

## ----------------------------------------------------------------------------
## option -iconname.
## ----------------------------------------------------------------------------
configbody VisParses::iconname {
   wm iconname $itk_interior $itk_option(-iconname)
}

## ----------------------------------------------------------------------------
## center on the middle of the application.
## ----------------------------------------------------------------------------
body VisParses::center {} {
   scan [wm geometry .] "%dx%d+%d+%d" appwd appht appx appy
   set wd [expr $appwd  - 80]
   set ht [expr $appht - 80 ]
   set x [expr ($appwd - $wd) / 2 + $appx]
   set y [expr ($appht - $ht)  / 2 + $appy]
   wm geometry $itk_interior "$wd\x$ht+$x+$y"
}

## ----------------------------------------------------------------------------
## gets another word of typed input.
## Hangs until the word is supplied.
## ----------------------------------------------------------------------------
body VisParses::readWord {} {
  global queue

  # busy waiting for next input
  if {$queue == ""} {
    vwait queue
  }

  set word [lindex $queue 0]
  set queue [lreplace $queue 0 0]
  #puts "DEBUG: dequeueing `$word'"

  return $word
}

## ----------------------------------------------------------------------------
## undisplay the selected ParseTree.
## ----------------------------------------------------------------------------
body VisParses::_deleteParseAction {} {

  set page [_labelOfIndex select]
  if {$page == ""} {
    return
  }
  set found 0
  foreach parse [array names _pageOfParse] {
     if {$_pageOfParse($parse) == $page} {
       incr found
       break
     }
  }

  # this should not happen in any case :
  if {!$found} {
     error "internal error: can't find parse of page $page"
  }
  removeParse $parse
}

## ----------------------------------------------------------------------------
## returns the label of a page.
## ----------------------------------------------------------------------------
body VisParses::_labelOfIndex {index} {
  set tbn $itk_component(tbn)
  set pageIndex [$tbn index $index]
  if {$pageIndex < 0} {
    return ""
  }
  return [lindex [$tbn pageconfigure $pageIndex -label] 4]
}

## ----------------------------------------------------------------------------
## invoked by space in the textbox.
## ----------------------------------------------------------------------------
body VisParses::_spaceHandler {} {

  global queue
  set box $itk_component(textbox)

  # Spaces are very special, since they trigger actions in the
  # background. But that only happens if the space
  # comes right after a word, not after another space.
  if {[regexp {[ 	]$} $_userInput]} return

  # And space at the beginning of a line doesn't make sense.
  if {$_userInput == ""} return

  # get last space-delimited word from the text
  if {![regexp {([^ ]+)[ 	]*$} $_userInput foo word]} {
    puts "WARNING: `$_userInput' doesn't match the regexp."
  }

  # deliver word to the queue
  #puts "DEBUG: queueing `$word'"
  lappend queue $word
  
  # finally insert the space
  $box insert end " "

}

## ----------------------------------------------------------------------------
## invoked by Backspace in the textbox.
## ----------------------------------------------------------------------------
body VisParses::_BackspaceHandler {} {

  global queue
  set box $itk_component(textbox)

  # backspace on an empty line is inoperable.
  if {$_userInput == ""} return

  # If the last space after a word is erased, the previous word is
  # withdrawn from the parse (but it remains in the textbox for
  # further editing).
  if {[regexp {[^ ] $} $_userInput]} {

    # if the queue has a word in it (not a take-back request),
    # we simply remove the last element (so strictly speaking, it
    # isn't really a queue...)

    if {$queue != "" && ![regexp {\$\#\$} $queue]} {
      #puts "DEBUG: retracting `[lindex $queue end]'"
      set nextToLast [expr [llength $queue] - 2]
      set queue [lrange $queue 0 $nextToLast]
    } else {

      # a special symbol is queued to denote `take back a word'
      set word "$\#$"
      #puts "DEBUG: queueing `$word'"
      lappend queue $word

      # and we also switch to the previous tab in the notebook
      set tbn $itk_component(tbn) 
      $tbn configure -auto true
      set previoustab [expr [$tbn view] -1]
      if {$previoustab >= 0} {
        $tbn view $previoustab
      }
    }

  }

  # process the backspace
  $box delete [expr [string length $_userInput] - 1]

}

## ----------------------------------------------------------------------------
## Display all canvases of the selected parse smaller.
## ----------------------------------------------------------------------------
body VisParses::_zoomIn {} {
  set parseId "parseTree[_labelOfIndex select]"
  if {$parseId == "parseTree"} {
    return
  }
  set parse $itk_component($parseId)
  $parse zoom 1
}

## ----------------------------------------------------------------------------
## display all canvases of the selected parse bigger.
## ----------------------------------------------------------------------------
body VisParses::_zoomOut {} {
  set parseId "parseTree[_labelOfIndex select]"
  if {$parseId == "parseTree"} {
    return
  }
  set parse $itk_component($parseId)
  $parse zoom -1
}

## ----------------------------------------------------------------------------
## set the orientation of the split of the current parsetree.
## ----------------------------------------------------------------------------
body VisParses::_setOrientation {} {
  set parseId "parseTree[_labelOfIndex select]"
  if {$parseId == "parseTree"} {
    return
  }
  set parse $itk_component($parseId)
  $parse setOrientation $_currentOrient
}

## ----------------------------------------------------------------------------
## get the orientation of the split of the current parsetree.
## ----------------------------------------------------------------------------
body VisParses::_getOrientation {} {
  set parseId "parseTree[_labelOfIndex select]"
  if {$parseId == "parseTree"} {
    return
  }
  set parse $itk_component($parseId)
  set _currentOrient [$parse getOrientation]
  return $_currentOrient
}

## ----------------------------------------------------------------------------
## get the score of current parsetree
## ----------------------------------------------------------------------------
body VisParses::getBadness {} {
  set parseId "parseTree[_labelOfIndex select]"
  if {$parseId == "parseTree"} {
    return
  }
  set parse $itk_component($parseId)
  if {[scan [$parse parse getBadness] "%d %d %f" hard soft score] > 0} {
    set badnessString [format "Score: %d/%d/%4.3e" $hard $soft $score]
  } else {
    set badnessString "Score: n.a."
  }
  $itk_component(scoreLabel) configure -text $badnessString

  $itk_component(idLabel) configure -text "Parse: [$parse parse getId]"
}

## ----------------------------------------------------------------------------
## update the visual parts which depends on the selected parsetree
## ----------------------------------------------------------------------------
body VisParses::_updateVisual {} {
  set parseId "parseTree[_labelOfIndex select]"
  if {$parseId == "parseTree"} {
    return
  }
  set parse $itk_component($parseId)
  set method [$parse parse getSearchStrategy]
  set lattice [$parse parse getLatticeId]
  configure -title "$method:$lattice"
  _getOrientation
  getBadness
}

## ----------------------------------------------------------------------------
## Delegate a highlight action to the parsetree
## ----------------------------------------------------------------------------
body VisParses::highlight {parse which} {
  set parseId "parseTree$_pageOfParse($parse)"
  $itk_component($parseId) highlight $which
}

body VisParses::update {} {
    set parseId "parseTree[_labelOfIndex select]"
    if {$parseId == "parseTree"} {
	return
    }

    if { [$itk_component($parseId) canUndo] == 1} {
	$itk_component(editmenu) entryconfigure 0 -state normal
    } else {
	$itk_component(editmenu) entryconfigure 0 -state disabled
    }

    if { [$itk_component($parseId) canRedo] == 1} {
	$itk_component(editmenu) entryconfigure 1 -state normal
    } else {
	$itk_component(editmenu) entryconfigure 1 -state disabled
    }
}