# Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaMain - The root of the YADA application.
## This class is serves as a mediator between the different components of the
## application. Its main tasks are
## - construction of the main application window where 
## - container for all YadaDocuments
## - user navigation between loaded YadaDocuments
## - initialization of all needed components during system startup
## - information propagation between different components
## - authority to delegate application wide tasks to responsible components
##
## So in order to talk to one specific part of the application
## a method in YadaMain is available delegating a task to the correct component.
## The YadaMain is also used to query components of the application, i.e. YadaDocuments,
## only known at runtime by their symbolic name.
##
## All YadaDocuments loaded at system startup and those created during runtime
## are registered with YadaMain::registerDocument() and from there on known to
## the application so that the user can switch to it. All registered YadaDocuments
## are kept in the YadaMain::_registeredDocument hash. The YadaDocument::nextDocument
## and YadaDocument::prevDocument establish a double linked list of all known 
## documents where the head and the tail of the list are stored in YadaMain::_firstDocument
## and YadaMain::_lastDocument respectively. Showing a registered document is
## the achieved by YadaMain::activateDocument().
##
## \author Michael Daum 
##
## $Id: YadaMain.tcl,v 1.16 2004/03/25 13:58:23 ddreyer Exp $
## ----------------------------------------------------------------------------
class YadaMain {
  inherit itk::Widget

  # variables ----------------------------------------------------------------

  ## name of the currently visible document embedded into the main widget
  private variable _currentDocumentName	""

  ## currently visible YadaDocument
  private variable _currentDocument	""

  ## toolbar of the currently visible YadaDocument
  private variable _currentToolbar	""

  ## hash of all known YadaDocuments matching a symbolic name to its YadaDocument
  private variable _registeredDocument; ## \type TclArray

  ## head of the list of all YadaDocuments
  private variable _firstDocument ""

  ## tail of the list of all YadaDocuments
  private variable _lastDocument ""

  ## toggle to switch on and off the toolbar
  private variable _toolbarFlag 1

  ## toggle to switch on and off the status line
  private variable _statusFlag 1

  ## current status message. 
  ## Use printStatus() to change its contents.
  private variable _status ""


  # methods ------------------------------------------------------------------
  public method activateDocument {{document ""}}; ## \type YadaDocument
  public method registerDocument {document}; ## \type YadaDocument
  public method unregisterDocument {document}; ## \type YadaDocument
  public method newDocument {className docName}; ## \type TclString, TclString
  public method quit {}
  public method printStatus {message} ; ## \type TclString
  public method getDocument {name}; ## \type TclString
  public method getDocuments {args}; ## \type TclList
  public method getExperiment {experimentName}; ## \type TclString
  public method getExperiments {args}; ## \type TclList
  public method getActiveExperiments {args}; ## \type TclList
  public method getExperimentNames {args}; ## \type TclList
  public method getGrammar {grammarName}; ## \type TclString
  public method getGrammars {args}; ## \type TclList
  public method getGrammarNames {args}; ## \type TclList
  public method getGrammarOfExperiment {experimentName}; ## \type TclString
  public method getWordgraphsOfExperiment {experimentName}; ## \type TclString
  public method getSignificance {}
  public method getTimeTolerance {}
  public method getEditor {}
  public method getWeirdness {}
  public method getMachine {name}; ## \type TclString
  public method getMachines {args}; ## \type TclList
  public method getMachineNames {args}; ## \type TclList
  public method setToggle {name {value ""}}
  public method getToggle {name}; ## \type TclString
  public method document {docName args}; ## \type TclString,TclList
  public method init {}

  ## methods for getting unique file paths to save object state  
  public method getPersistanceFileNameForGrammar { grammarName }; ## \type TclString
  public method getPersistanceFileNameForExperiment { experimentName }; ## \type TclString
  public method getPersistanceFileNameForMachine { machineName }; ## \type TclString
  public method getPersistanceFileNameForMisc {}
  public method getPersistanceFileNameForRunner { runnerName  }; ## \type TclString

  constructor {args} {}; ## \type TclList
  destructor {}

  private method _about {}
};

## ----------------------------------------------------------------------------
## constructor.
## This constructor builds up all widgets and packs them into one megawidget.
## \param args are handed over to itk_initialize() to initialize this megawidget
## ----------------------------------------------------------------------------
body YadaMain::constructor {args} {

  #
  # the menu
  #
  itk_component add menuFrame {
     frame $itk_interior.menuFrame \
       	-borderwidth 2 \
       	-relief raised
  } {}

  # file menu
  itk_component add fileButton {
     menubutton $itk_component(menuFrame).fileButton \
       	-text "File" \
       	-underline 0 \
       	-menu $itk_component(menuFrame).fileButton.menu
  } {
     keep -background -cursor
     rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
  }

  # edit menu
  itk_component add editButton {
     menubutton $itk_component(menuFrame).editButton \
       	-text "Edit" \
       	-underline 0 
  } {
     keep -background -cursor
     rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
  }

  # view menu
  itk_component add viewButton {
     menubutton $itk_component(menuFrame).viewButton \
       	-text "View" \
       	-underline 0 
  } {
     keep -background -cursor
     rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
  }

  # Documents
  itk_component add documentButton {
     menubutton $itk_component(menuFrame).documentButton \
       	-text "Documents" \
       	-underline 0 \
       	-menu $itk_component(menuFrame).documentButton.menu
  } {
     keep -background -cursor
     rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
  }

  itk_component add documentMenu {
     menu $itk_component(documentButton).menu 
  } {
     keep -background -cursor
     rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
  }

  # help 
  itk_component add helpButton {
    menubutton $itk_component(menuFrame).helpButton \
	-text "Help" \
	-underline 0 \
	-menu $itk_component(menuFrame).helpButton.menu
  } {
     keep -background -cursor
     rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
  }

  itk_component add helpMenu {
    menu $itk_component(helpButton).menu
  } {
     keep -background -cursor
     rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
  }

  $itk_component(helpMenu) add command \
       	-label "Index" \
       	-underline 0 \
       	-command {printMessage "Under Construction!"}
  $itk_component(helpMenu) add command \
       	-label "Context" \
       	-underline 0 \
       	-command {printMessage "Under Construction!"}
  $itk_component(helpMenu) add command \
       	-label "About" \
       	-underline 0 \
       	-command [code $this _about]
  

  #
  # toolbar area
  #
  itk_component add toolbarArea {
    frame $itk_interior.toolbarArea \
	-borderwidth 2 \
	-relief flat
  }

  #
  # the workarea
  #
  itk_component add childsite {
    iwidgets::labeledframe $itk_interior.childsite \
	-labelvariable [scope _currentDocumentName]  \
	-labelpos nw \
	-labelfont {-adobe-helvetica-bold-r-normal--12-*-*-*-*-*-*-*}
  }

  # the status line
  itk_component add status {
    label $itk_interior.status \
	-textvariable [scope _status] \
	-borderwidth 2 \
	-relief sunk \
	-anchor w \
	-justify left \
	-font {-adobe-helvetica-medium-r-normal--12-*-*-*-*-*-*-*}
  } {}
  itk_component add padder {
    frame $itk_interior.padder
  }

  
  # the startup screen
  newDocument YadaStartUp Startup

  # 
  # packing
  #
  pack $itk_component(menuFrame) -side top -fill x
  pack $itk_component(fileButton) \
       $itk_component(editButton) \
       $itk_component(viewButton) \
       $itk_component(documentButton) \
       $itk_component(helpButton) -side left -padx 2 -pady 2
  pack $itk_component(toolbarArea) -side top -expand 0 -fill x -padx 2
  pack $itk_component(childsite) -side top -expand 1 -fill both
  pack $itk_component(status) -fill x -side top -side top -padx 7
  pack $itk_component(padder) -fill x -side top -side top -padx 7 -pady 3

  eval itk_initialize $args
  activateDocument 

}

## ----------------------------------------------------------------------------
## init
## ----------------------------------------------------------------------------
body YadaMain::init {} {
  #
  # build the initial session
  # TODO: implement load session / save session
  #
  set configDoc [newDocument YadaConfig Configure]
  $configDoc component tabno configure -font {-adobe-helvetica-bold-r-normal--12-*-*-*-*-*-*-*}
  
  #newDocument YadaRunner Runner


  # initialize all documents
  foreach name [array names _registeredDocument] {
    $_registeredDocument($name) init
  }

  printStatus "Welcome to Yet Another Data Analyser!"

  # destroy the startup dialog
  itcl::delete object [getDocument Startup]

  # initial document
  activateDocument
}

## ----------------------------------------------------------------------------
## destructor
## ----------------------------------------------------------------------------
body YadaMain::destructor {} {
}

## ----------------------------------------------------------------------------
## activateDocument
## ----------------------------------------------------------------------------
body YadaMain::activateDocument {{document ""}} {

  # without parameters we activate the first document
  if {$document == ""} {
    set document $_firstDocument
    
    if {$document == ""} {
      return
    }
  }

  set docName [$document cget -name]
  if {$docName == ""} {
    error "trying to register $document without a name"
  }


  # is the window already active
  if {$_currentDocument == $document} {
    return
  }

  # is the window registered
  if {![info exists _registeredDocument($docName)]} {
    error "window $docName not registered"
  }

  if {$_currentDocument != ""} {
    # call the deActivation handle of the old document
    $_currentDocument deActivationHandle

    # hide the current active window and toolbar
    pack forget [$_currentDocument component hull]
    if {$_currentToolbar != ""} {
      pack forget $_currentToolbar
    }
  } 

  # show the requested window
  set _currentDocumentName $docName
  set _currentDocument $document
  set _currentToolbar [$document component toolbar]
  set toolbarArea $itk_component(toolbarArea)
  pack [$_currentDocument component hull] \
    -fill both -expand 1 \
    -in [$itk_component(childsite) childsite] \
    -padx 5 -pady 5

  if {$_currentToolbar != ""} {
    pack $_currentToolbar -fill x -expand 0 -in $toolbarArea
  }

  # call the activation handle of the document
  $_currentDocument activationHandle
}


## ----------------------------------------------------------------------------
## registerDocument
## ----------------------------------------------------------------------------
body YadaMain::registerDocument {document} {

  # add to known documents
  set docName [$document cget -name]
  set _registeredDocument($docName) $document

  # add to menu
  $itk_component(documentMenu) add radiobutton \
       	-label "$docName" \
       	-variable [scope _currentDocumentName] \
       	-value "$docName" \
       	-command [code $this activateDocument $document]

  # link into application
  if {$_firstDocument == ""} {
    $document configure -prevDocument ""
    $document configure -nextDocument ""
    set _firstDocument $document
    set _lastDocument $document
  } else {
    $_lastDocument configure -nextDocument $document
    $document configure -prevDocument $_lastDocument
    $document configure -nextDocument ""
    set _lastDocument $document
  }
}

## ----------------------------------------------------------------------------
## unregisterDocument
## ----------------------------------------------------------------------------
body YadaMain::unregisterDocument {document} {
  set docName [$document cget -name]

  # remove from menu
  $itk_component(documentMenu) delete $docName

  # switch backwards if we just see that document
  if {$docName == $_currentDocumentName} {
    $document backward

    if {$docName == $_currentDocumentName} {
      $document forward
    } 
  } 

  unset _registeredDocument($docName)

  # fix the double linked list of documents
  set prevDocument [$document cget -prevDocument]
  set nextDocument [$document cget -nextDocument]
  
  if {$prevDocument != ""} {
    $prevDocument configure -nextDocument $nextDocument
  }
  if {$nextDocument != ""} {
    $nextDocument configure -prevDocument $prevDocument
  }

  if {$_firstDocument == $document} {
    set _firstDocument $nextDocument
  }

  if {$_lastDocument == $document} {
    set _lastDocument $prevDocument
  }
}

## ----------------------------------------------------------------------------
## quit
## ----------------------------------------------------------------------------
body YadaMain::quit {} {
  # TODO: implement soft landing
  exit
}

## ----------------------------------------------------------------------------
## newDocument
## ----------------------------------------------------------------------------
body YadaMain::newDocument {className docName} {

  # build a unique name
  set components [component]
  set noObjects 1
  set qualifiedName $docName
  while {[lsearch -exact $components $qualifiedName] >= 0} {
    incr noObjects
    set qualifiedName "$docName#$noObjects"
  }

  # check for unique documents
  if {$noObjects > 1} {
    if {$className == "YadaConfig"} {
      printMessage "ERROR: only one YadaConfig allowed." error
      return
    }
  }

  # add and build the new document
  set childsite [$itk_component(childsite) childsite]
  itk_component add $qualifiedName {
    $className $childsite.#auto -name $qualifiedName
  } {}

  return $itk_component($qualifiedName)
}

## ----------------------------------------------------------------------------
## getExperiment
## ----------------------------------------------------------------------------
body YadaMain::getExperiment {experimentName} {
  return [[$itk_component(Configure) component experiments] getItem $experimentName]
}

## ----------------------------------------------------------------------------
## getExperiments
## ----------------------------------------------------------------------------
body YadaMain::getExperiments {args} {
  set experimentsComponent [$itk_component(Configure) component experiments]
  return [eval $experimentsComponent getAllItems $args]
}

## ----------------------------------------------------------------------------
## getExperimentNames
## ----------------------------------------------------------------------------
body YadaMain::getExperimentNames {args} {
  set experimentsComponent [$itk_component(Configure) component experiments]
  return [eval $experimentsComponent getAllItemNames $args]
}

## ----------------------------------------------------------------------------
## getActiveExperiments
## ----------------------------------------------------------------------------
body YadaMain::getActiveExperiments {args} {
  set experimentsComponent [$itk_component(Configure) component experiments]
  return [eval $experimentsComponent getActiveExperiments $args]
}


## ----------------------------------------------------------------------------
## getGrammar
## ----------------------------------------------------------------------------
body YadaMain::getGrammar {grammarName} {
  set grammarsComponent [$itk_component(Configure) component grammars]
  return [$grammarsComponent getItem $grammarName]
}

## ----------------------------------------------------------------------------
## getGrammars
## ----------------------------------------------------------------------------
body YadaMain::getGrammars {args} {
  set grammarsComponent [$itk_component(Configure) component grammars]
  return [eval $grammarsComponent getAllItems $args]
}

## ----------------------------------------------------------------------------
## getGrammarNames
## ----------------------------------------------------------------------------
body YadaMain::getGrammarNames {args} {
  set grammarsComponent [$itk_component(Configure) component grammars]
  return [eval $grammarsComponent getAllItemNames $args]
}

## ----------------------------------------------------------------------------
## getMachine
## ----------------------------------------------------------------------------
body YadaMain::getMachine {macName} {
  return [[$itk_component(Configure) component machines] getItem $macName]
}

## ----------------------------------------------------------------------------
## getMachines
## ----------------------------------------------------------------------------
body YadaMain::getMachines {args} {
  set machinesComponent [$itk_component(Configure) component machines]
  return [eval $machinesComponent getAllItems $args]
}

## ----------------------------------------------------------------------------
## getMachineNames
## ----------------------------------------------------------------------------
body YadaMain::getMachineNames {args} {
  set machinesComponent [$itk_component(Configure) component machines]
  return [eval $machinesComponent getAllItemNames $args]
}

## ----------------------------------------------------------------------------
## getGrammarOfExperiment
## ----------------------------------------------------------------------------
body YadaMain::getGrammarOfExperiment {experimentName} {
  if {$experimentName == "" || $experimentName == "<none>"} {
    return ""
  }
  return [getGrammar [[getExperiment $experimentName] cget -grammarName]]
}

## ----------------------------------------------------------------------------
## getWordgraphsOfExperiment
## ----------------------------------------------------------------------------
body YadaMain::getWordgraphsOfExperiment {experimentName} {
  if {$experimentName == "" || $experimentName == "<none>"} {
    return ""
  }

  return [[getGrammarOfExperiment $experimentName] getSelection]
}

## ----------------------------------------------------------------------------
## getWeirdness
## ----------------------------------------------------------------------------
body YadaMain::getWeirdness {} {
  return [[$itk_component(Configure) component misc] cget -weirdness]
}

## ----------------------------------------------------------------------------
## getSignificance
## ----------------------------------------------------------------------------
body YadaMain::getSignificance {} {
  return [[$itk_component(Configure) component misc] cget -significance]
}

## ----------------------------------------------------------------------------
## getTimeTolerance
## ----------------------------------------------------------------------------
body YadaMain::getTimeTolerance {} {
  return [[$itk_component(Configure) component misc] cget -timeTolerance]
}

## ----------------------------------------------------------------------------
## getEditor
## ----------------------------------------------------------------------------
body YadaMain::getEditor {} {
  return [[$itk_component(Configure) component misc] cget -editor]
}

## ----------------------------------------------------------------------------
## getDocument
## ----------------------------------------------------------------------------
body YadaMain::getDocument {docName} {
  if {[info exists _registeredDocument($docName)]} {
    return $_registeredDocument($docName)
  } else {
    return ""
  }
}

## ----------------------------------------------------------------------------
## getDocuments
## ----------------------------------------------------------------------------
body YadaMain::getDocuments {args} {
  set documents ""

  if {[llength $args] == 0} {
    foreach docName [array names _registeredDocument] {
      lappend documents $_registeredDocument(docName)
    }
  } else {
    foreach docName [array names _registeredDocument] {
      foreach pattern $args {
	if {[string match $pattern $docName]} {
	  lappend documents $_registeredDocument($docName)
	  break
	}
      }
    }
  }

  return $documents
}

## ----------------------------------------------------------------------------
## printStatus
## ----------------------------------------------------------------------------
body YadaMain::printStatus {message} {
  set _status $message
  update 
}

## ----------------------------------------------------------------------------
## getToggle
## ----------------------------------------------------------------------------
body YadaMain::getToggle {name} {
  switch $name {
    "toolbar" {
      return $_toolbarFlag
    }
    "status" {
      return $_statusFlag
    }
    default {
      return 0
    }
  }
}

## ----------------------------------------------------------------------------
## setToggle
## ----------------------------------------------------------------------------
body YadaMain::setToggle {name {value ""}} {
  switch $name {
    "toolbar" {
      if {$value != 0 && $value != 1} {
	set _toolbarFlag [expr $_toolbarFlag?0:1]
      } else {
	set _toolbarFlag $value
      }
      if {$_toolbarFlag} {
	pack $itk_component(toolbarArea) \
	  -side top -expand 0 -fill x -padx 2 \
	  -after $itk_component(menuFrame)
      } else {
	pack forget $itk_component(toolbarArea)
      }
      return $_toolbarFlag
    }
    "status" {
      if {$value != 0 && $value != 1} {
	set _statusFlag [expr $_statusFlag?0:1]
      } else {
	set _statusFlag $value
      }
      if {$_statusFlag} {
	pack $itk_component(status) \
	  -fill x -side top -padx 7 \
	  -after $itk_component(childsite)
      } else {
	pack forget $itk_component(status)
      }
      return $_statusFlag
    }
  }
}

## ----------------------------------------------------------------------------
## delegate a task to a named YadaDocument.
## This method fetches the YadaDocument named \a name and passes over the
## arguments \a args as a command.
## \param docName the name of the YadaDocument
## \param args passed commands 
## \returns the result of the YadaDocument call
## ----------------------------------------------------------------------------
body YadaMain::document {docName args} {
  eval [getDocument $docName] $args
}

## ----------------------------------------------------------------------------
## _about
## ----------------------------------------------------------------------------
body YadaMain::_about {} {
  set about [YadaAbout .#auto]
  $about center .
  $about activate
  destroy $about
}

## ----------------------------------------------------------------------------
## Returns unique file path dependent on grammar name
## ----------------------------------------------------------------------------
body YadaMain::getPersistanceFileNameForGrammar { grammarName } {
  global env
  ## delete blanks in name
  return [file join $env(YADA_DATA) [string map {" " ""} $grammarName]_grammarconfig.xml]    
}

## ----------------------------------------------------------------------------
## Returns unique file path dependent on experiment name
## ----------------------------------------------------------------------------
body YadaMain::getPersistanceFileNameForExperiment { experimentName } {
  global env
  ## delete blanks in name
  return [file join $env(YADA_DATA) [string map {" " ""} $experimentName]_experimentconfig.xml] 
}

## ----------------------------------------------------------------------------
## Returns unique file path dependent on machine name
## ----------------------------------------------------------------------------
body YadaMain::getPersistanceFileNameForMachine { machineName } {
  global env
  ## delete blanks in name
  return [file join $env(YADA_DATA) [string map {" " ""} $machineName]_machineconfig.xml] 
}

## ----------------------------------------------------------------------------
## Returns unique file path for general configuration tab
## ----------------------------------------------------------------------------
body YadaMain::getPersistanceFileNameForMisc {} {
    return "GeneralsConfig.xml"
}

## ----------------------------------------------------------------------------
## Returns unique file path for runner documents
## ----------------------------------------------------------------------------
body YadaMain::getPersistanceFileNameForRunner { runnerName } {
  global env
  ## delete blanks in name
  return [file join $env(YADA_DATA) [string map {" " ""} $runnerName]_runner.xml]     
}