# Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## \file YadaProc.tcl
## Bunch of small helper functions.
## This module collects all helpers that don't fit in a proper class for themselves.
## Actually for those cases the YadaMain might be taken, but ...
##
## \author Michael Daum
##
## $Id YadaProc.tcl$
## ----------------------------------------------------------------------------

## ----------------------------------------------------------------------------
## convert a time in numerical format to a human readeable format.
## \param time time in milliseconds
## \returns a string representation
## ----------------------------------------------------------------------------
proc timeToString {time} {; ## \type TclNumber
  
  if {$time <= 0.0} {
    return "0 ms"
  }
  
  # calculate the units
  set time [expr double($time)]
  set w    [expr floor($time / 604800000.0)]
  set time [expr $time - $w * 604800000.0]
  set d    [expr floor($time / 86400000.0)]
  set time [expr $time - $d * 86400000.0]
  set h    [expr floor($time / 3600000.0)]
  set time [expr $time - $h * 3600000.0]
  set m    [expr floor($time / 60000.0)]
  set time [expr $time - $m * 60000.0]
  set s    [expr floor($time / 1000.0)]
  set ms   [expr $time - $s * 1000]

  # build the string 
  set isFirst 1
  set result ""
  foreach unit {w d h m s ms} {
    set value [subst \$$unit]
    if {$value > 0.0} {
      if {$isFirst} {
	set isFirst 0
      } else {
	append result " "
      }
      append result [format "%.0f %s" $value $unit]
    }
  }

  return $result
}

## ----------------------------------------------------------------------------
## get the index of an entry in a listbox.
## \param listBox the tcl command of the list box 
## \param entry an entry in the \a listBox
## \returns the index > 0 if the \a entry exists, -1 if not so
## ----------------------------------------------------------------------------
proc getIndexOfName {listBox entry} {; ## \type TclCommand, TclString
  set listOfEntries [$listBox get 0 end]
  set index [lsearch $listOfEntries $entry]
  return $index
}

## ----------------------------------------------------------------------------
## open a message dialog with a given text.
## \param message the notification text
## \param image an image name, see YadaImages for possible values
## ----------------------------------------------------------------------------
proc printMessage {message {image "info"}} {; ## \type TclString, TclString
  if {![winfo exists .msg]} {
    iwidgets::messagedialog .msg \
  } 
  .msg configure \
    -text $message \
    -title "Yada - Message"  \
    -image [YadaImages::get $image] \
    -modality application \
    -textpadx 10 \
    -master .

  .msg hide Cancel
  .msg center .
  .msg activate
  destroy .msg
}

## ----------------------------------------------------------------------------
## This helper constructs some helpfull YadaExperiment objects as long as you
## didn't provide some yourself.
## ----------------------------------------------------------------------------
proc getYadaDefaultExperiments {} {
  global env

  set scriptIntro {
    set edges off
    set debug off
    set profile on
    set searchresult off
    set progress off
    set eval off
    set cache off
    set timelimit 120000
    set taggerIgnoreImpossible 1
    set taggerNormalizeToOne 1
    set preprocessor /usr/bin/zcat
    set prepsuffix .gz
    set locale de_DE
    %load
  }

  set scriptOutro {
    printparse
    verify
  }

  set machine "HampsterCluster"
  set grammar "stellingen-flat"
  set grammarVersion "0.01"

  #
  # frobbings
  #
  foreach pressure { 0.0 0.1 } {
    YadaExperiment ::#auto \
      -name		"frobbing pressure $pressure" \
      -isActive		1 \
      -type		"frobbing" \
      -command		"nice $env(YADA_CDGP)" \
      -machineName	"$machine" \
      -grammarName	"$grammar" \
      -dataDir		"$env(YADA_DATA)/$machine/$grammar/v$grammarVersion/frobbing/pressure-$pressure" \
      -script "
	$scriptIntro
	newnet %wordgraph
	frobbing net0 combined 0.0 batch
	$scriptOutro"

    YadaExperiment ::#auto \
      -name		"frobbing pressure 0.1" \
      -isActive		1 \
      -type		"frobbing" \
      -command		"nice $env(YADA_CDGP)" \
      -machineName	"$machine" \
      -grammarName	"$grammar" \
      -dataDir		"$env(YADA_DATA)/$machine/$grammar/v$grammarVersion/frobbing/pressure-$pressure" \
      -script "
	$scriptIntro
	newnet %wordgraph
	frobbing net0 combined 0.0 batch
	$scriptOutro"
  }
  

  #
  # gls
  #
  YadaExperiment ::#auto \
    -name		"gls default" \
    -isActive		1 \
    -type		"gls" \
    -command		"nice $env(YADA_CDGP)" \
    -machineName	"$machine" \
    -grammarName	"$grammar" \
    -dataDir		"$env(YADA_DATA)/$machine/$grammar/v$grammarVersion/gls/default" \
    -script "
      $scriptIntro
      newnet %wordgraph
      gls
      $scriptOutro"

  #
  # netsearching
  #
  YadaExperiment ::#auto \
    -name		"netsearch default" \
    -isActive		1 \
    -type		"netsearch" \
    -command		"nice $env(YADA_CDGP)" \
    -machineName	"$machine" \
    -grammarName	"$grammar" \
    -dataDir		"$env(YADA_DATA)/$machine/$grammar/v$grammarVersion/netsearch/default" \
    -script		"
      $scriptIntro
      set cache on
      newnet %wordgraph
      netsearch
      $scriptOutro"
}

## ----------------------------------------------------------------------------
## This helper constructs some helpfull YadaGrammar objects as long as you
## didn't provide some yourself.
## ----------------------------------------------------------------------------
proc getYadaDefaultGrammars {} {
  global env

  # the stellingen corpus
  set grammar [YadaGrammar ::#auto -name "stellingen-flat"]
  $grammar setFiles [file join $env(YADA_GRAMMARS) stellingen stellingen-flat.cdg.gz]
  $grammar setSelection \
      n001k000-1 \
      n001k000-2 \
      n001k001-1 \
      n001k002-1 \
      n001k002-2 \
      n001k003-1 \
      n001k003-2 \
      n001k004-1 \
      n001k005-1 \
      n001k006-1 \
      n001k007-1 \
      n001k007-2 \
      n001k008-1 \
      n001k008-2 \
      n001k008-3 \
      n001k009-1 \
      n001k010-1 \
      n001k011-1 \
      n001k011-2 \
      n001k012-1 \
      n001k013-1 \
      n001k013-2 \
      n001k014-1 \
      n001k015-1 \
      n001k015-2 \
      n001k016-1 \
      n001k016-2 \
      n001k018-1 \
      n001k018-2 \
      n001k019-1 \
      n002k000-1 \
      n002k000-2 \
      n002k001-1 \
      n002k002-1 \
      n002k002-2 \
      n002k002-3 \
      n002k003-1 \
      n002k004-1 \
      n002k004-2 \
      n002k006-1 \
      n002k007-1 \
      n002k008-1 \
      n002k009-1 \
      n002k010-1 \
      n002k011-1 \
      n002k012-1 \
      n002k012-2 \
      n002k012-3 \
      n002k014-1 \
      n002k015-1 \
      n002k017-1 \
      n002k019-1 \
      n002k020-1 \
      n002k022-1 \
      n002k022-1b \
      n002k022-1c \
      n002k022-2 \
      n002k023-1 \
      n002k023-1b \
      n002k023-2 \
      n002k023-3 \
      n002k025-1 \
      n002k026-1 \
      n002k028-1 \
      n002k029-1 \
      n002k029-2 \
      n002k030-1 \
      n002k032-1 \
      n002k033-1 \
      n002k035-1 \
      n002k036-1 \
      n002k036-2 \
      n002k037-1 \
      n002k037-1a \
      n002k037-1b \
      n002k037-1c \
      n002k037-1d \
      n002k037-1e \
      n002k037-1f \
      n002k037-1g \
      n002k037-1h \
      n002k037-2 \
      n002k038-1 \
      n002k039-1 \
      n002k040-1 \
      n002k040-2 \
      n002k041-1 \
      n002k042-1 \
      n003k000-1 \
      n003k000-2 \
      n003k001-1 \
      n003k001-2 \
      n003k002-1 \
      n003k003-1 \
      n003k003-2 \
      n003k003-3 \
      n003k004-1 \
      n003k004-2 \
      n003k004-3 \
      n003k005-1 \
      n003k005-2 \
      n003k006-1 \
      n003k007-1 \
      n003k009-1 \
      n003k009-2 \
      n003k011-1 \
      n003k012-1 \
      n003k013-1 \
      n003k013-1b \
      n003k013-2 \
      n003k013-2b \
      n003k014-1 \
      n003k014-2 \
      n003k015-1 \
      n003k015-2 \
      n003k015-3 \
      n004k000-1 \
      n004k001-1 \
      n004k003-1 \
      n004k003-2 \
      n004k003-23 \
      n004k003-3 \
      n004k004-1 \
      n004k005-1 \
      n004k006-1 \
      n004k007-1 \
      n004k008-1 \
      n004k009-1 \
      n004k009-1b \
      n004k010-1 \
      n004k013-1 \
      n004k013-2 \
      n004k013-2b \
      n004k014-1 \
      n004k014-2 \
      n004k014-3 \
      n004k015-1 \
      n004k016-1 \
      n004k016-1b \
      n004k017-1 \
      n004k017-2 \
      n004k017-3 \
      n004k018-1 \
      n004k019-1 \
      n004k019-2 \
      n004k019-3 \
      n004k020-1 \
      n004k020-2 \
      n004k020-3 \
      n004k020-4 \
      n004k021-1 \
      n004k021-2 \
      n004k021-2b \
      n004k021-3 \
      n004k022-1 \
      n004k022-2 \
      n004k023-1 \
      n004k024-1 \
      n004k024-1b \
      n004k025-1 \
      n004k025-2 \
      n004k026-1 \
      n004k026-2 \
      n004k027-1 \
      n004k030-1 \
      n004k030-2 \
      n004k030-2b \
      n004k030-3 \
      n004k031-1 \
      n004k031-2 \
      n004k031-3 \
      n004k032-1 \
      n004k032-1b \
      n004k032-1c \
      n004k032-2 \
      n004k034-1 \
      n004k034-2 \
      n004k034-3 \
      n004k034-3b \
      n004k035-1 \
      n004k035-2 \
      n004k036-1 \
      n004k036-2 \
      n004k037-1 \
      n004k037-2 \
      n004k038-1 \
      n004k038-2 \
      n004k038-3 \
      n004k039-1 \
      n005k000-1 \
      n005k000-2 \
      n005k001-1 \
      n005k002-1 \
      n005k002-2 \
      n005k003-1 \
      n005k004-1 \
      n005k004-2 \
      n005k004-3 \
      n005k004-4 \
      n005k005-1 \
      n005k006-1 \
      n005k006-2 \
      n005k008-1 \
      n005k008-2 \
      n005k009-1 \
      n005k010-1 \
      n005k011-1 \
      n005k012-1 \
      n005k012-2 \
      n005k013-1 \
      n005k014-1 \
      n005k015-1 \
      n005k016-1 \
      n005k016-2 \
      n005k017-1 \
      n005k017-2 \
      n005k018-1 \
      n005k018-2 \
      n005k019-1 \
      n005k022-1 \
      n005k023-1 \
      n005k024-1 \
      n004k031-12 
}

## ----------------------------------------------------------------------------
## This helper constructs some helpfull YadaMachine objects as long as you
## didn't provide some yourself.
## ----------------------------------------------------------------------------
proc getYadaDefaultMachines {} {

  YadaMachine ::#auto \
    -name "HampsterCluster" \
    -access "ssh %node" \
    -nodes "nats42 nats43 nats44 nats45 nats46 nats47" \
    -cpus 10

  YadaMachine ::#auto \
    -name "MosixCluster" \
    -access "ssh %node mosrun -c -l -F -f -j1-6" \
    -nodes "nats42 nats43 nats44 nats45 nats46 nats47" \
    -cpus 6

  YadaMachine ::#auto \
    -name "nats0" \
    -access "ssh -x %node" \
    -nodes "nats0" \
    -cpus 1

}


## ----------------------------------------------------------------------------
## extract notifications.
## This helper extracts all notifications in a cdgp DomDocument. That is, it
## searches for \\<notify ...\\> nodes and gets the \a desc attribute from it.
## \param document the DomDocument that we are interested in
## \param severity optionally specify the \c type attribute of the notification
## that we want to extract
## \returns the TclList of all messages in all notifications
## ----------------------------------------------------------------------------
proc getAllNotifications {document {severity ""}} {; ## \type DomDocument, TclString

  set messages ""

  # get the cdgp logfile root element
  set cdgpNode [$document getElement "cdgp"]
  if {$cdgpNode != ""} {

    if {$severity != ""} {
      set nodes [$cdgpNode descendant all "notify" "type" $severity]
    } else {
      set nodes [$cdgpNode descendant all "notify"]
    }

    foreach node $nodes {
      lappend messages [$node getAttribute "desc"]
    }
  } 

  return $messages
}
