# Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaRunnableDocument - abstract document able to launch YadaJobs.
## This document offers all functionality to launch one or a whole bunch of
## YadaJobs. Most notably YadaRunner inherits this abstract class and implements
## the left virtual methods.
## \ingroup YadauRunnableDocuments
##
## \author Michael Daum 
##
## $Id: YadaRunnableDocument.tcl,v 1.15 2004/03/25 15:12:53 ddreyer Exp $
## ----------------------------------------------------------------------------
class YadaRunnableDocument {
  inherit YadaDocument

  # variables ----------------------------------------------------------------

  ## flag to indicate a received break signal
  protected variable _gotBreak 0

  ## flag to indicate that the document is currently busy
  protected variable _isRunning 0

  ## results of _analyseDocument().
  ## This variable is modified by _analyseDocument() in a sideeffect.
  ## Outside this method we only read _analyseResult.
  ## The filled values are:
  ## - wordgraph: the name of the wordgraph in the DomDocument
  ## - noValues: the number of levelvalues in a computed constraintnet
  ## - score: the score of a found parse
  ## - totalTime: the total amount of time needed to terminate the solution method
  ## - solutionTime: the time to find the parse
  ## - softTime: the time to find the first parse without any hard violations
  ## - tries: the number of tried items in the solution parse
  ## - structHits: the number of structural correct items in the solution parse
  ## - labelledHits: the number of structural correct and correctly labelled items
  ## - lexicalHits: the number of structural, labelled and lexically correct items
  ## - strictHits: the number of identical dependency edges
  protected variable _analyseResult; ## \type TclArray


  ## the wordgraph selected by a right mouse click on the table.
  protected variable _selectedContextWordgraph ""

  ## hash mapping wordgraph ids to 0 or 1.
  ## This indicates whether a wordgraph has been computed or not.
  protected variable _doneWordgraph; ## \type TclArray

  ## hash of registered YadaJobs.
  ## This structure hashes a YadaExperiment to all YadaJobs that have been constructed
  ## for it and therefor holds all jobs scheduled by this document. 
  ## \see _registerJob(). 
  private variable _currentJobs; ## \type TclArray

  ## profiler for job execution time.
  private variable _timer 0

  ## flag indicating whether the display should focus on new data.
  private variable _followDataFlag 1

  ## current sorting state.
  ## This variable stores the column number allong which the table is sorted.
  private variable _lastSorting 0

  ## traced process.
  ## This variable stores the YadaProcess that we currently trace.
  private variable _progressProcess ""

  ## flag indicating whether we clear the log window on each new computation.
  private variable _clearLogFlag 0

  ## flag indicating whether we display the table
  private variable _tableFlag 1

  ## flag indicating whether we display the log window
  private variable _consoleFlag 1

  ## flag indicating the orientation of the splitted window.
  ## Possible values are horizontal and vertical
  private variable _orientFlag "horizontal"

  # methods ------------------------------------------------------------------
  public method init {}; ## \virtual
  public method print {message} ; ## \type TclString
  public method loadUpdatedWordgraphs {args} ; ## \type TclList
  public method sortRows {colNo args} ;## \type TclNumber, TclList
  public method resortRows {}
  public method selectExperiment {{experimentName ""}} ;## \type TclString
  
  public method step {{wordgraph ""}} ;## \type TclString
  public method run {{experiment ""}}; ## \type YadaExperiment
  public method runall {} 
  public method askRunall {} 
  public method clear {}
  public method reload {}
  public method save {}

  public method breakJob {}
  public method unbreakJob {}
  public method interruptJobs {}
  public method termJob {}
  public method killJob {}

  constructor {args} {}; ## \type TclList

  public method markWordgraph {wordgraph}; ## \type TclString
  public method unmarkWordgraph {wordgraph}; ## \type TclString
  public method setToggle {name {value ""}}; ## \type TclString, TclRef

  protected method _getWordgraphs {}
  protected method _getActiveWordgraphs {}
  protected method _getSelectedWork {}
  protected method _getUndoneWork {}
  protected method _getDoneWork {}
  protected method _getSelectedExperiment {}

  protected method _getRowOfWordgraph {wordgraph}; ## \type TclString
  protected method _computeSpeedup {time1 time2}; ## \type TclNumber, TclNumber
  protected method _insertData {wordgraph args}; ## \type TclString, TclList
  protected method _getTypeOfColumn {colNo} ;## \purevirtual \type TclNumber
  protected method _analyseDocument {document experiment}; ## \type DomDocument, TclExperiment
  protected method _computeStatistics {} 
  protected method _jobConsumer {job}; ## \type YadaJob
  protected method _registerJob {job}; ## \type YadaJob
  protected method _unregisterJob {job}; ## \type YadaJob
  protected method _showContextMenu {x y xroot yroot}; ## \type TclNumber, TclNumber, TclNumber, TclNumber
  protected method _viewLogfile {{experiments ""}}; ## \type TclList
  protected method _viewXmlfile {{experiments ""}}; ## \type TclList
  protected method _clearRow {}
  protected method _deleteData {}
  protected method _traceProcess {}
  protected method _resetDataOf {wordgraphs}; ## \type TclList
  protected method _getXMLFileName { wordgraph }; ## \type TclString
  
  private method _computeETA {machine noJobs}; ## \type YadaMachine, TclNumber
  private method _escapeCommand {}
  private method _returnCommand {}
  private method _leaveTableCommand {}
  private method _breakButtonAction {}
  private method _runButtonAction {}
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaRunnableDocument::constructor {args} {

  #
  # edit menu
  #
  $itk_component(editMenu) add command \
       	-label "Clear" \
       	-underline 0 \
       	-command [code $this clear]
  $itk_component(editMenu) add command \
       	-label "Run" \
       	-underline 0 \
       	-command [code $this run]
  $itk_component(editMenu) add command \
       	-label "Run All" \
       	-underline 0 \
       	-command [code $this runall]
  $itk_component(editMenu) add command \
       	-label "Step" \
       	-underline 0 \
       	-command [code $this step]
  $itk_component(editMenu) add command \
       	-label "Break" \
       	-underline 0 \
       	-command [code $this breakProcess]
  $itk_component(editMenu) add command \
       	-label "Unbreak" \
       	-underline 0 \
       	-command [code $this unbreakProcess]
  $itk_component(editMenu) add command \
       	-label "Interrupt" \
       	-underline 0 \
       	-command [code $this interruptProcess]
  $itk_component(editMenu) add command \
       	-label "Terminate" \
       	-underline 0 \
       	-command [code $this termProcess]
  $itk_component(editMenu) add command \
       	-label "Kill" \
       	-underline 0 \
       	-command [code $this killProcess]
  
  #
  # view menu
  #
  itk_component add sortIncMenu {
     menu $itk_component(viewMenu).sortIncMenu 
  } {
     keep -background -cursor
     rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
  }

  itk_component add sortDecMenu {
     menu $itk_component(viewMenu).sortDecMenu
  } {
     keep -background -cursor
     rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
  }

  $itk_component(viewMenu) add checkbutton \
       	-label "Table" \
       	-variable [scope _tableFlag] \
       	-command [code $this setToggle "table"]
  $itk_component(viewMenu) add checkbutton \
       	-label "Console" \
       	-variable [scope _consoleFlag] \
       	-command [code $this setToggle "console"]
  $itk_component(viewMenu) add separator
  $itk_component(viewMenu) add radiobutton \
       	-label "Horizontal" \
       	-variable [scope _orientFlag] \
       	-value "horizontal" \
       	-command [code $this setToggle "orientation"]
  $itk_component(viewMenu) add radiobutton \
       	-label "Vertical" \
       	-variable [scope _orientFlag] \
       	-value "vertical" \
       	-command [code $this setToggle "orientation"]
  $itk_component(viewMenu) add separator
  $itk_component(viewMenu) add checkbutton \
	-label "Follow new data" \
	-variable [scope _followDataFlag] \
	-command [code $this setToggle "follow"]
  $itk_component(viewMenu) add checkbutton \
	-label "Clear new log" \
	-variable [scope _clearLogFlag] \
	-command [code $this setToggle "clear"]
  $itk_component(viewMenu) add separator
  $itk_component(viewMenu) add cascade \
	-label "Sort inc. by" \
	-underline 0 \
	-menu $itk_component(sortIncMenu)
  $itk_component(viewMenu) add cascade \
	-label "Sort dec. by" \
	-underline 0 \
	-menu $itk_component(sortDecMenu)

  #
  # toolbar
  #
  $itk_component(toolbar) add FlatButton reloadButton \
    -image [YadaImages::get reload] \
    -balloonstr "reload" \
    -command [code $this reload]

  $itk_component(toolbar) add FlatButton clearButton \
	-image [YadaImages::get erase] \
	-command [code $this clear] \
	-balloonstr "clear"

  $itk_component(toolbar) add FlatButton runButton \
	-image [YadaImages::get run] \
	-command [code $this _runButtonAction] \
	-balloonstr "run"

  $itk_component(toolbar) add FlatButton runAllButton \
	-image [YadaImages::get runall] \
	-command [code $this runall] \
	-balloonstr "run all"

  $itk_component(toolbar) add FlatButton stepButton \
	-image [YadaImages::get step] \
	-command [code $this step] \
	-balloonstr "step"

  $itk_component(toolbar) add FlatButton breakButton \
	-image [YadaImages::get giveup] \
	-command [code $this _breakButtonAction] \
	-balloonstr "break"

  $itk_component(toolbar) add FlatButton interruptButton \
	-image [YadaImages::get finish] \
	-command [code $this interruptJobs] \
	-balloonstr "interrupt"

  $itk_component(toolbar) add FlatButton termButton \
	-image [YadaImages::get stop] \
	-command [code $this termJob] \
	-balloonstr "terminate"

  $itk_component(toolbar) add FlatButton killButton \
	-image [YadaImages::get bomb] \
	-command [code $this killJob] \
	-balloonstr "kill"

  #
  # overview heading
  # 
  itk_component add heading {
    frame $itk_component(childsite).heading
  }


  #
  # the pane
  #
  itk_component add pane {
      iwidgets::panedwindow $itk_component(childsite).pane \
          -orient horizontal \
	  -sashindent -8 \
	  -thickness 2 \
	  -sashheight 8 \
	  -sashwidth 18
  } {
      keep -background -cursor
  }
  $itk_component(pane) add top -margin 0
  $itk_component(pane) add bottom -margin 0
  set top [$itk_component(pane) childsite top]
  set bottom [$itk_component(pane) childsite bottom]
  $itk_component(pane) fraction 74 26

  #
  # the logwindow
  #
  itk_component add log {
    YadaLog $bottom.log \
	    -hscrollmode dynamic \
	    -wrap none \
	    -borderwidth 2 \
	    -sbwidth 12 \
	    -troughcolor gray \
	    -textbackground gray90 \
	    -textfont {-misc-fixed-medium-normal-semicondensed-*-13-*-*-*-*-*-*-*}
  } { }

  $itk_component(log) autotag INFO \
      -foreground black \
      -background green3 
  $itk_component(log) autotag WARNING \
      -foreground black \
      -background yellow2 
  $itk_component(log) autotag ERROR \
      -foreground white \
      -background red3 
  $itk_component(log) autotag DEBUG \
      -foreground white \
      -background blue3 
  $itk_component(log) autotag PROFILE \
      -foreground black \
      -background lightblue 
  
  #
  # table
  #
  itk_component add table {
     MyTable $top.table \
       	-exportselection 1 \
	-selecttitles 1 \
	-rowseparator "\n" \
	-colseparator "\t" \
       	-rows 300 \
       	-cols 6 \
       	-state normal \
       	-cellanchor c \
       	-colstretchmode last \
       	-rowstretchmode none \
       	-selectmode extended \
       	-selecttype row \
       	-titlerows 1 \
       	-titlecols 1 \
	-borderwidth 3 \
	-outerborderwidth 2 \
	-outerrelief sunk \
	-selforeground "" \
	-selbackground "" \
	-font {-adobe-helvetica-medium-r-normal--12-*-*-*-*-*-*-*} \
        -troughcolor gray \
	-hscrollmode dynamic
  } { 
     keep -background -cursor 
  }


  #
  # the context menu
  #
  itk_component add contextMenu {
     menu $itk_interior.contextMenu 
  } {
     keep -background -cursor
     rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
  }
  $itk_component(contextMenu) add command \
       	-label "View Log" \
       	-underline 0 \
       	-command [code $this _viewLogfile]
  $itk_component(contextMenu) add command \
       	-label "View Xml" \
       	-underline 5 \
       	-command [code $this _viewXmlfile]
  $itk_component(contextMenu) add command \
       	-label "Trace" \
       	-underline 0 \
       	-command [code $this _traceProcess]
  $itk_component(contextMenu) add command \
       	-label "Clear" \
       	-underline 0 \
       	-command [code $this _clearRow]
  $itk_component(contextMenu) add command \
       	-label "Delete" \
       	-underline 0 \
       	-command [code $this _deleteData]

  # bindings
  set hull $itk_component(hull)
  set tableInTable [$itk_component(table) component table]
  bind $hull <Leave> [code $this _leaveTableCommand]
  bind $hull <FocusOut> [code $this _leaveTableCommand]
  bind $tableInTable <KeyPress-Escape> [code "$this _escapeCommand; break"]
  bind $tableInTable <KeyPress-Return> [code "$this _returnCommand; break"]
  bind $tableInTable <Button-3> [code "$this _showContextMenu %x %y %X %Y; break"]

  # packing
  pack $itk_component(heading) -fill x -anchor n 
  pack $itk_component(pane) -fill both -expand 1 -side top
  pack $itk_component(table) -fill both -expand 1 -padx 2 -pady 5
  pack $itk_component(log) -side top -expand 1 -fill both -padx 2 -pady 5

}

## ----------------------------------------------------------------------------
## _getRowOfWordgraph
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_getRowOfWordgraph {wordgraph} {

  set table $itk_component(table)
  
  set rmax [$table cget -rows]
  set ids [$table getCell 1,0 $rmax,0]
  set row [lsearch -exact $ids $wordgraph]
  incr row

  return $row
}

## ----------------------------------------------------------------------------
## _runButtonAction
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_runButtonAction {} {
  set _timer [clock seconds] 
  set _isRunning 1
  run
  while {$_isRunning} {
    vwait [scope _isRunning]
  }
  if {$_gotBreak} {
    printMessage "interrupted job(s)"
    unbreakJob
  } else {
    printMessage "finished job(s)"
  }
}

## ----------------------------------------------------------------------------
## _breakButtonAction
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_breakButtonAction {} {
  if {$_gotBreak} {
    unbreakJob
  } else {
    breakJob
  }
}

## ----------------------------------------------------------------------------
## breakJob
## ----------------------------------------------------------------------------
body YadaRunnableDocument::breakJob {} {
  print "INFO: break received\n"
  if {$_isRunning} {
    set _gotBreak 1
    $itk_component(toolbar) itemconfigure breakButton -background red3
  }

  # remove all queued and waiting jobs, let the running once finish
  after idle { printMessage "Please wait on step completition." }
  foreach experiment [array names _currentJobs] {
    foreach job $_currentJobs($experiment) {
      set experiment [$job cget -experiment]
      set machine [.main getMachine [$experiment cget -machineName]]
      if {![$job hasPassedState "dequeuing"]} {
	_unregisterJob $job
	$machine stopJob $job
      }
      update
    }
  }
  _computeStatistics
}

## ----------------------------------------------------------------------------
## interruptJobs
## ----------------------------------------------------------------------------
body YadaRunnableDocument::interruptJobs {} {
  breakJob
  foreach experiment [array names _currentJobs] {
    foreach job $_currentJobs($experiment) {
      set process [$job cget -process]
      if {$process != ""} {
	$process interrupt
      }
      update
    }
  }
}

## ----------------------------------------------------------------------------
## termJob
## ----------------------------------------------------------------------------
body YadaRunnableDocument::termJob {} {
  breakJob
  foreach experiment [array names _currentJobs] {
    foreach job $_currentJobs($experiment) {
      set process [$job cget -process]
      if {$process != ""} {
	$process term
      }
      update
    }
  }
}

## ----------------------------------------------------------------------------
## killJob
## ----------------------------------------------------------------------------
body YadaRunnableDocument::killJob {} {
  breakJob
  foreach experiment [array names _currentJobs] {
    foreach job $_currentJobs($experiment) {
      set process [$job cget -process]
      if {$process != ""} {
	$process killJob
      }
      update
    }
  }
}

## ----------------------------------------------------------------------------
## reload
## ----------------------------------------------------------------------------
body YadaRunnableDocument::reload {} {
  set selection [$itk_component(table) curselection]
  clear
  if {$selection != ""} {
    $itk_component(table) selection set [lindex $selection 0] [lindex $selection end]
  }
  loadUpdatedWordgraphs
  _computeStatistics
}


## ----------------------------------------------------------------------------
## clear
## ----------------------------------------------------------------------------
body YadaRunnableDocument::clear {} {
  set table $itk_component(table)
  set selection [$table curselection]
  set cmax [$table cget -cols]

  # only delete the seleted rows
  if {$selection != ""} {
    set minRow 9999
    set maxRow 0
    foreach index $selection {
      scan $index "%d,%d" row col
      $table erase $row,1 $row,$cmax
      $table clear tags $row,0 $row,$cmax
      set wordgraph [$table getCell $row,0]
      _resetDataOf $wordgraph
    }
  } else {
  
    # reset the whole statistics
    $table clear tags 
    $table delete rows -holddimensions 1 [$table cget -rows]

    # insert the wordgraphs
    set wordgraphs [_getWordgraphs]
    set rmax [llength $wordgraphs]
    set table $itk_component(table)
    if {$rmax < 20} {
      set rmax 20
    } else {
      incr rmax
    }
    $table configure -rows $rmax
    set row 1
    foreach wordgraph $wordgraphs {
      $table setCell $row,0 $wordgraph
      incr row
    }

    _resetDataOf "$wordgraphs"
    _getActiveWordgraphs
  }

  unbreakJob
  $itk_component(log) clear
  .main printStatus ""
}

## ----------------------------------------------------------------------------
## unbreakJob
## ----------------------------------------------------------------------------
body YadaRunnableDocument::unbreakJob {} {
  set _gotBreak 0
  $itk_component(toolbar) itemconfigure breakButton -background gray90
}

## ----------------------------------------------------------------------------
## _escapeCommand
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_escapeCommand {} {
  set table $itk_component(table)
  set lastActiveCell [$table tag cell active]
  scan $lastActiveCell "%d,%d" row col
  
  if {$col == 0} {
    $table selection clear all
  } else {
    $table activate -1,-1
  }
}

## ----------------------------------------------------------------------------
## _leaveTableCommand
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_leaveTableCommand {} {
  set table $itk_component(table)
  set lastActiveCell [$table tag cell active]
  if {$lastActiveCell == ""} {
    return
  }

  scan $lastActiveCell "%d,%d" row col
  $table activate -1,-1
}

## ----------------------------------------------------------------------------
## _returnCommand
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_returnCommand {} {
  set table $itk_component(table)
  set lastActiveCell [$table tag cell active]
  scan $lastActiveCell "%d,%d" row col
  set colMax [expr [$table cget -cols] -1]
  if {$col == $colMax} {
    incr row
    set col 1
  } else {
    incr col
  }

  $table activate $row,$col
  $table see $row,$col
}

## ----------------------------------------------------------------------------
## init
## ----------------------------------------------------------------------------
body YadaRunnableDocument::init {} {
  if {$_isInitialized} {
    return
  }
  chain
  clear

  $itk_component(sortIncMenu) delete 0 end
  $itk_component(sortDecMenu) delete 0 end

  set cmax [$itk_component(table) cget -cols]
  set headings [$itk_component(table) getCell 0,0 0,$cmax]
  set colNo 0
  foreach heading $headings {
    if {$heading != ""} {
      $itk_component(sortIncMenu) add command \
	-label "$heading" \
	-command [code $this sortRows $colNo -increasing]
      $itk_component(sortDecMenu) add command \
	-label "$heading" \
	-command [code $this sortRows $colNo -decreasing]
    }
    incr colNo
  }
}

## ----------------------------------------------------------------------------
## _computeSpeedup
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_computeSpeedup {time1 time2} {

  set timeTolerance [.main getTimeTolerance]

  if {[expr abs ($time1 - $time2) > $timeTolerance]} {
    return [expr (1.0 + $time2) / (1.0 + $time1)]
  } else {
    return 1.0
  }
}

## ----------------------------------------------------------------------------
## _computeStatistics
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_computeStatistics {} {

  set noJobs 0
  foreach experiment [array names _currentJobs] {
    incr noJobs [llength $_currentJobs($experiment)]
  }

  set experiments [array names _currentJobs]
  set noExperiments [expr [llength $experiments] -1]
#  if {$noExperiments < 0} {
#    return
#  }
  
  set timeStamp [clock seconds]
  set elapsedString [timeToString [expr ($timeStamp - $_timer) * 1000]]

  set experiment [lindex $experiments 0]
  if {$experiment != ""} {
    set machine [.main getMachine [$experiment cget -machineName]]
    set avgString [timeToString [$machine getTimePerJob]]
    set etaString [timeToString [_computeETA $machine $noJobs]]
    set noRunning [expr int([$machine getLoad] * [$machine cget -cpus])]
    incr noJobs -$noRunning
    .main printStatus "duration = $elapsedString : eta = $etaString : avg = $avgString : remaining experiments = $noExperiments : running jobs = $noRunning : remaining jobs = $noJobs"
  } else {
    .main printStatus "duration = $elapsedString"
  }
}

## ----------------------------------------------------------------------------
## _getWordgraphs: get all wordgraphs of a experiments grammar
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_getWordgraphs {} {

  set experiment [_getSelectedExperiment]
  if {$experiment == ""} {
    return ""
  }

  set grammar [.main getGrammar [$experiment cget -grammarName]]
  return [$grammar getSelection]
}


## ----------------------------------------------------------------------------
## _getSelectedExperiment
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_getSelectedExperiment {} {
  set experimentName [$itk_component(experimentBox) get]
  if {$experimentName == "" || $experimentName == "<none>"} {
    return ""
  }
  return [.main getExperiment $experimentName]
}

## ----------------------------------------------------------------------------
## _getSelectedWork
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_getSelectedWork {} {
  set table $itk_component(table)

  set cells [$table curselection]
  
  foreach index $cells {
    scan $index "%d,%d" row col
    set wordgraph [$table getCell $row,0]
    if {$wordgraph != ""} {
      set id($row,0) $wordgraph
    }
  }

  set selectedWordgraphs ""
  foreach index [array names id] {
    lappend selectedWordgraphs $id($index)
  }

  return $selectedWordgraphs
}

## ----------------------------------------------------------------------------
## _getDoneWork
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_getDoneWork {} {

  set result ""

  foreach wordgraph [array names _doneWordgraph] {
    if {$_doneWordgraph($wordgraph) != 0} {
      lappend result $wordgraph
    }
  }

  return $result
}

## ----------------------------------------------------------------------------
## _getUndoneWork
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_getUndoneWork {} {
  set result ""

  foreach wordgraph [array names _doneWordgraph] {
    if {$_doneWordgraph($wordgraph) == 0} {
      lappend result $wordgraph
    }
  }

  return $result
}

## ----------------------------------------------------------------------------
## unmarkWordgraph
## ----------------------------------------------------------------------------
body YadaRunnableDocument::unmarkWordgraph {wordgraph} {
  _computeStatistics
  set table $itk_component(table)
  set cmax [$table cget -cols]
  set rmax [$table cget -rows]

  set row [_getRowOfWordgraph $wordgraph]
  $table clear tags $row,0 
}

## ----------------------------------------------------------------------------
## markWordgraph
## ----------------------------------------------------------------------------
body YadaRunnableDocument::markWordgraph {wordgraph} {
  _computeStatistics
  if {$wordgraph == ""} {
    return
  }

  set table $itk_component(table)
  set cmax [$table cget -cols]
  set rmax [$table cget -rows]

  # indicate running wordgraph 
  set row [_getRowOfWordgraph $wordgraph]
  $table tag cell runner $row,0

  # clear the log window
  if {$_clearLogFlag} {
    $itk_component(log) clear
  }

  # scroll to the current wordgraph
  if {$_followDataFlag} {
    $table see $row,1
  }

  # possibly remove the given wordgraph from the selection
  $table selection clear $row,0 $row,$cmax
}


## ----------------------------------------------------------------------------
## print
## ----------------------------------------------------------------------------
body YadaRunnableDocument::print {message} {
  $itk_component(log) print $message
}

## ----------------------------------------------------------------------------
## loadUpdatedWordgraphs 
## 
## ----------------------------------------------------------------------------
body YadaRunnableDocument::loadUpdatedWordgraphs {args} {
  set experimentName [$itk_component(experimentBox) get]
  set experiment [.main getExperiment $experimentName]

  # reset scores of selected wordgraphs
  if {[llength $args] == 0} {
    set activeWordgraphs [_getActiveWordgraphs]
  } else {
    set activeWordgraphs $args
  }
  _resetDataOf "$activeWordgraphs"

  # load all xml files 
  set noLoadedFiles 0
  foreach wordgraph $activeWordgraphs {
  
    # check if the file is present
    set file [$experiment getXmlFileName $wordgraph]
    if {![file exists $file]} {
      set _doneWordgraph($wordgraph) 0
      continue
    }
    set _doneWordgraph($wordgraph) 1

    # say what we do
    print "INFO: loading `$file'\n"

    # analyse the file
    set fd [open "|zcat $file" r]
    set data [read $fd]
    close $fd
    if {[catch {
      set document [dom parse $data]
      _analyseDocument $document $experiment
    } errMsg]} {
      print "ERROR: invalid xml file\n"
      print "$errMsg\n$::errorInfo"
    }
    catch {
      $document delete
      set document ""
    }

    incr noLoadedFiles
    update
  }

  print "INFO: loaded $noLoadedFiles files\n"
  resortRows
  #_computeStatistics
  .main printStatus ""
}

## ----------------------------------------------------------------------------
## _getActiveWordgraphs
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_getActiveWordgraphs {} {

  set wordgraphs [lsort -dictionary [_getSelectedWork]]
  if {$wordgraphs == ""} {
    set wordgraphs [lsort -dictionary [_getUndoneWork]]
    if {$wordgraphs == ""} {
      set wordgraphs [lsort -dictionary [_getWordgraphs]]
    } else {
    }
  }

  return $wordgraphs
}

## ----------------------------------------------------------------------------
## _insertData
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_insertData {wordgraph args} {
  set table $itk_component(table)
  set cmax [$table cget -cols]

  # get place where we belong to
  set row [_getRowOfWordgraph $wordgraph]

  # fill it
  $table clear tags $row,1 $row,$cmax
  $table erase $row,1 $row,$cmax
  $table setCell row $row,1 $args

  # show it
  if {$_followDataFlag} {
    $table see $row,1
  }
}

## ----------------------------------------------------------------------------
## resortRows
## ----------------------------------------------------------------------------
body YadaRunnableDocument::resortRows {} {
  if {$_lastSorting != ""} {
    eval sortRows $_lastSorting
  }
}

## ----------------------------------------------------------------------------
## sortRows
## ----------------------------------------------------------------------------
body YadaRunnableDocument::sortRows {colNo args} {
  set type [_getTypeOfColumn $colNo]
  eval $itk_component(table) sortRows $colNo -$type $args
  set _lastSorting $colNo
  eval lappend _lastSorting $args
}

## ----------------------------------------------------------------------------
## compute the estimated time of arrival.
## This method tries to compute the left duration of given number of jobs
## that might run on a YadaMachine. This is done by querying the time-per-job
## statistics via YadaMachine::getTimePerJob(). This figure is being tracked
## by the YadaMachine itself and might change the more YadaJobs are executed
## by that machine.
## \param machine the YadaMachine which we are interested in
## \param noJobs the number of jobs that we still need to do
## \returns the estimation
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_computeETA {machine noJobs} {

  set timePerJob [$machine getTimePerJob]
  set noCpus [$machine cget -cpus]

  return [expr round($noJobs * $timePerJob / (0.0 + $noCpus))]
}


## ----------------------------------------------------------------------------
## analyse a given DomDocument produced with the help of a YadaExperiment.
## This method extracts all the information that we normaly are interested in
## in runnable documents. Classes (like YadaGlsStats) overload this function and
## extract extra information not analysed here. So as a sideeffect the resulting
## data is stored into the protected variable _analyseResult as documented
## above. In some cases where the DomDocument is illformed we still try to 
## extract as many information as possible. In case of any error you will notice
## that by a return code of 0. Error messages are give by print().
## The \a experiment is consulted especialy when no evaluation data is contained in 
## the \a document in order to get the number of tries in the statistics correctly
## an which is only provided by the used CDG grammar, especialy the contained
## annotation of the lattice.
## \param document the DomDocument resulting from a dom parse
## \param experiment the experiment with which we produced this document
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_analyseDocument {document experiment} {
  
  # initialize the _analyseResult
  if {[info exists _analyseResult]} {
    unset _analyseResult
  }
  array set _analyseResult {
    wordgraph		""
    noValues		0
    score		0.0
    totalTime		0
    solutionTime	0
    softTime		0
    tries		0
    structHits		0
    labelledHits	0
    lexicalHits		0
    strictHits		0
  }

  set isOk 1

  # get the root element
  set cdgpNode [$document documentElement]
  if {$cdgpNode == ""} {
    print "ERROR: no cdgp logfile\n"
    return 0
  } 

  # xpath analysis
  
  # get number of values
  set _analyseResult(noValues) [[$cdgpNode selectNodes \
    {descendant::netInfo[1]/number[@name="noValues"]}] @value ]

  # get name of wordgraph
  set _analyseResult(wordgraph) [[$cdgpNode selectNodes \
    {descendant::netInfo[1]/ref[@type="Lattice"]}] @name ]
  if {$_analyseResult(wordgraph) == ""} {
    print "ERROR: no net information found => newnet failed"
    return 0
  }

  # get the score
  set _analyseResult(score) [[$cdgpNode selectNodes \
    {descendant::parse[1]/number[@name="score"]}] @value]
  if {$_analyseResult(score) == ""} {
    set _analyseResult(score) 0.0
  }

  # get the parse verification node
  set parseVerificationNode [$cdgpNode descendant 1 "parseverification"]
  if {$parseVerificationNode != ""} {
    foreach aspect {tries structHits labelledHits lexicalHits strictHits} {
      set _analyseResult($aspect) [[$parseVerificationNode selectNodes \
	descendant::[string tolower $aspect]/number\[@name="SYN"\]] @value]
    }
  } else {
    # verification fallback: compute tries from annotation
    set isOk 0
    print "ERROR: no parse verification found\n"
    print "INFO: consulting annotation\n"
    set grammar [.main getGrammarOfExperiment [$experiment cget -name]]
    $grammar scan; # CAUTION: might be costy
    set yadaWordgraph [$grammar getWordgraph $_analyseResult(wordgraph)]
    set _analyseResult(tries) [$yadaWordgraph cget -noAnnoEdges]
    if {$_analyseResult(tries) == 0} {
      print "ERROR: no correct annotation for wordgraph '$_analyseResult(wordgraph)'"
    }
  }

  # parse the statistics 
  foreach aspect {totalTime solutionTime softTime} {
    catch {set _analyseResult($aspect) [[$cdgpNode selectNodes \
      descendant::statistics/number\[@name="$aspect"\]] @value ]}
    if {$_analyseResult($aspect) == ""} {
      set _analyseResult($aspect) 0
    }
  }

  # end of xpath analysis

  # fix solution and soft time in case of no score
  if {$_analyseResult(softTime) == 0} {
    set _analyseResult(softTime) $_analyseResult(solutionTime)
  }
  if {$_analyseResult(solutionTime) == 0} {
    set _analyseResult(softTime) $_analyseResult(totalTime)
    set _analyseResult(solutionTime) $_analyseResult(totalTime)
  }
  
  return $isOk
}


## ----------------------------------------------------------------------------
## step
## ----------------------------------------------------------------------------
body YadaRunnableDocument::step {{wordgraph ""}} {

  if {$wordgraph == ""} {
    set wordgraph [lindex [_getActiveWordgraphs] 0]
    if {$wordgraph == ""} {
      return 0
    }
  }

  # get the current experiment
  set experiment [_getSelectedExperiment]
  if {$experiment == ""} {
    return 1
  }

  # check if the job was already done
  set file [$experiment getXmlFileName $wordgraph]
  if {[file exists $file]} {
    loadUpdatedWordgraphs $wordgraph
    _computeStatistics
    return
  }
  
  # construct a job
  set machine [.main getMachine [$experiment cget -machineName]]
  set job [YadaJob ::#auto \
    -wordgraphName $wordgraph \
    -experiment $experiment \
    -runner $this \
    -machine $machine \
    -consumerCommand [code $this _jobConsumer]]
  _registerJob $job
  $machine queueJob $job

  # start the job
  $machine startJob

  # wait
  set _isRunning 1
  while {$_isRunning} {
    vwait [scope _isRunning]
  }

  # finish message
  if {$_gotBreak} {
    printMessage "interrupted job(s)"
    unbreakJob
  } else {
    printMessage "finished job(s)"
  }

  return 1
}

## ----------------------------------------------------------------------------
## run
## ----------------------------------------------------------------------------
body YadaRunnableDocument::run {{experiment ""}} {

  # get the current experiment
  if {$experiment == ""} {
    set experiment [_getSelectedExperiment]
    if {$experiment == ""} {
      return 1
    }
  } else {
    selectExperiment [$experiment cget -name]
  }

  # check and load
  .main printStatus "... checking experiment '[$experiment cget -name]'"
  $experiment check
  loadUpdatedWordgraphs
  _computeStatistics
  if {[$experiment cget -isComplete]} {
    return
  }
  
  # construct all jobs 
  set machine [.main getMachine [$experiment cget -machineName]]
  foreach wordgraph [_getActiveWordgraphs] {
    set job [YadaJob ::#auto \
      -wordgraphName $wordgraph \
      -experiment $experiment \
      -runner $this \
      -machine $machine \
      -consumerCommand [code $this _jobConsumer]]
    $machine queueJob $job
    _registerJob $job
  }

  $machine startJob

  # wait
  set _isRunning 1
  while {$_isRunning} {
    vwait [scope _isRunning]
  }

  # finish message
  if {$_gotBreak} {
    printMessage "interrupted job(s)"
    unbreakJob
  } else {
    printMessage "finished job(s)"
  }
}

## ----------------------------------------------------------------------------
## runall
## ----------------------------------------------------------------------------
body YadaRunnableDocument::runall {} {
  set experimentNames [[$itk_component(experimentBox) component list] get 0 end]
  
  # construct possible jobs
  set experiments ""
  foreach experimentName $experimentNames {
    set experiment [.main getExperiment $experimentName]
    if {$experimentName == "<none>" ||
	$experimentName == "" ||
	![$experiment cget -isActive]} {
	continue
    }

    # force experiment checking
    .main printStatus "... checking experiment `$experimentName'"
    if {[$experiment check]} {
      continue ;# is complete
    }

    # construct jobs
    set noJobs 0
    set grammar [.main getGrammar [$experiment cget -grammarName]]
    set machine [.main getMachine [$experiment cget -machineName]]
    foreach wordgraph [$grammar getSelection] {
      set xmlFile [$experiment getXmlFileName $wordgraph]
      # dont recompute existing data
      if {[file exists $xmlFile]} {
	continue
      }

      set job [YadaJob ::#auto \
	-wordgraphName $wordgraph \
	-experiment $experiment \
	-runner $this \
	-machine $machine \
	-consumerCommand [code $this _jobConsumer]]
      _registerJob $job
      incr noJobs
    }
    if {$noJobs > 0} {
      lappend experiments $experiment
      print "INFO: $experimentName needs $noJobs jobs\n"
    }
  }

  if {[llength $experiments] == 0} {
    printMessage "no jobs need to be done"
    return
  }

  
  # confirm batch of jobs
  if {![askRunall]} {
    foreach experiment [array names _currentJobs] {
      catch {
	foreach job $_currentJobs($experiment) {
	  _unregisterJob $job
	  itcl::delete object $job
	  update ;# this loop could be time consuming
	}
      }
    }
    _computeStatistics
    return
  }

  # do it
  set _timer [clock seconds] 
  foreach experiment $experiments {

    # select experiment
    selectExperiment  [$experiment cget -name]
    loadUpdatedWordgraphs
    _computeStatistics
    if {[$experiment cget -isComplete]} {
      foreach job $_currentJobs($experiment) {
	_unregisterJob $job
	itcl::delete object $job
      }
      _computeStatistics
      continue
    }

    # queue jobs of the current experiment
    foreach job $_currentJobs($experiment) {
      set machine [$job cget -machine]
      $machine queueJob $job
    }

    # start
    $machine startJob
    
    # wait
    set _isRunning 1
    while {$_isRunning} {
      vwait [scope _isRunning]
    }

    # escape
    if {$_gotBreak} {
      break
    }
  }

  # finish message
  if {$_gotBreak} {
    printMessage "interrupted job(s)"
    unbreakJob
  } else {
    printMessage "finished job(s)"
  }

}


## ----------------------------------------------------------------------------
## _jobConsumer
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_jobConsumer {job} {
  $job setState "consuming"
  set wordgraph [$job cget -wordgraphName]
  set experiment [$job cget -experiment]
  set machine [$job cget -machine]

  # analyse the new data
  set xmlFile [$experiment getXmlFileName $wordgraph]
  print "INFO: loading `$xmlFile'\n"
  set fd [open "|zcat $xmlFile" r]
  set data [read $fd]
  close $fd

  if {[catch {
    set document [dom parse $data]
    _analyseDocument $document $experiment
  } errMsg ]} {
    print "ERROR: invalid xml file\n"
    print "$errMsg\n$::errorInfo"
  }
  catch {
    $document delete
    set document ""
  }

  # remove the job from the list of current once
  _unregisterJob $job
  $machine stopJob $job

  # refresh the statistics
  unmarkWordgraph $wordgraph
  set _doneWordgraph($wordgraph) 1

  # start a new job
  if {!$_gotBreak} {
    if {[$machine startJob]} {
      return
    }
  }

  # detect finish
  if {$_isRunning} {
    if {![info exists _currentJobs($experiment)] ||
	[llength $_currentJobs($experiment)] == 0} {
      set _isRunning 0 ;# signal vwait
    }
  }
}

## ----------------------------------------------------------------------------
## _registerJob
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_registerJob {job} {
  set experiment [$job cget -experiment]
  lappend _currentJobs($experiment) $job
}

## ----------------------------------------------------------------------------
## _unregisterJob
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_unregisterJob {job} {
  
  set experiment [$job cget -experiment]
  if {![info exists _currentJobs($experiment)]} {
    return
  }
  set index [lsearch -exact $_currentJobs($experiment) $job]
  if {$index >= 0} {
    set _currentJobs($experiment) "[lreplace $_currentJobs($experiment) $index $index]"
    if {$_currentJobs($experiment) == ""} {
      unset _currentJobs($experiment)
    }
  }
}

## ----------------------------------------------------------------------------
## askRunall
## ----------------------------------------------------------------------------
body YadaRunnableDocument::askRunall {} {
  set experiments [array names _currentJobs]
  set noExperiments [llength $experiments]
  if {$noExperiments == 0} {
    return 0
  }

  # estimate job
  set eta 0
  set noJobs 0
  foreach experiment $experiments {
    set machine [.main getMachine [$experiment cget -machineName]]
    set noExperimentJobs [llength $_currentJobs($experiment)]
    incr noJobs $noExperimentJobs
    set eta [expr $eta + [_computeETA $machine $noExperimentJobs]]
  }
  set etaString [timeToString $eta]

  # display dialog
  set msg [iwidgets::messagedialog .msg \
    -text \
"Are you sure that you want to compute 
$noJobs jobs in $noExperiments experiments(s)? 
This will be a job of ca. $etaString!" \
    -title "Yada - Question"  \
    -image [YadaImages::get question] \
    -modality application \
    -master $itk_interior]
    
  $msg buttonconfigure OK -text "Yes"
  $msg buttonconfigure Cancel -text "No"
  $msg center .
  
  set result [$msg activate]

  destroy $msg
  if {!$result} {
    return 0
  }

  return 1
}

## ----------------------------------------------------------------------------
## selectExperiment
## ----------------------------------------------------------------------------
body YadaRunnableDocument::selectExperiment {{experimentName ""}} {
  set box $itk_component(experimentBox)
  if {$experimentName == ""} {
    set experimentName [$box get]
    if {$experimentName == ""} {
      set experimentName "<none>"
    }
  }

  set listOfEntries [[$box component list] get 0 end]
  set index [lsearch $listOfEntries $experimentName]

  if {$index < 0} {
    selectExperiment "<none>"
  } else {
    $box clear entry
    $box insert entry 0 "$experimentName"
    $box selection set $index $index
    foreach wordgraph [array names _doneWordgraph] {
      unset _doneWordgraph($wordgraph)
    }
    clear
  }
  displayTitle
}

## ----------------------------------------------------------------------------
## _showContextMenu
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_showContextMenu {x y xroot yroot} {
  set _selectedContextWordgraph [$itk_component(table) getCell @0,$y]
  if {$_selectedContextWordgraph == ""} {
    return
  }

  tk_popup $itk_component(contextMenu) $xroot $yroot
}

## ----------------------------------------------------------------------------
## display logfile of a experiment.
## This method reads the logfiles of one or more experiments given a selected wordgraph
## in _selectedContextWordgraph.
## \param experiments the optional list of experiments that we want to display. if the
##	  list is empty the we fetch the currently selected experiments via _getSelectedExperiment().
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_viewLogfile {{experiments ""}} {

  if {$_selectedContextWordgraph == ""} {
    return
  }

  if {[llength $experiments] == 0} {
    lappend experiments [_getSelectedExperiment]
    if {[llength $experiments] == 0} {
      return
    }
  }

  if {!$_consoleFlag} {
    setToggle "console" 1
  }

  # read computed data of all experiments
  $itk_component(log) clear
  foreach experiment $experiments {
    set fileName [$experiment getLogFileName $_selectedContextWordgraph]
    print "INFO: reading $fileName\n"
    if {![file exists $fileName]} {
      printMessage "File not found!" "error"
      return
    }

    # read
    set fd [open "|zcat $fileName" r]
    set data [read $fd]
    close $fd

    # print
    print $data
  }
  $itk_component(log) see end
  
  set _selectedContextWordgraph ""
}

## ----------------------------------------------------------------------------
## _viewXmlfile
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_viewXmlfile {{experiments ""}} {

  if {$_selectedContextWordgraph == ""} {
    return
  }

  if {[llength $experiments] == 0} {
    lappend experiments [_getSelectedExperiment]
    if {[llength $experiments] == 0} {
      return
    }
  }

  if {!$_consoleFlag} {
    setToggle "console" 1
  }

  # read computed data of all experiments
  $itk_component(log) clear
  foreach experiment $experiments {
    set fileName [$experiment getXmlFileName $_selectedContextWordgraph]
    print "INFO: reading $fileName\n"
    if {![file exists $fileName]} {
      printMessage "File not found!" "error"
      return
    }

    # read
    set fd [open "|zcat $fileName" r]
    set data [read $fd]
    close $fd

    # print
    print $data
  }
  $itk_component(log) see end
  
  set _selectedContextWordgraph ""
}


## ----------------------------------------------------------------------------
## _clearRow
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_clearRow {} {

  set row [_getRowOfWordgraph $_selectedContextWordgraph]
  set table $itk_component(table)
  set cmax [$table cget -cols]
  $table clear tags $row,1 $row,$cmax
  $table erase $row,1 $row,$cmax

  set _selectedContextWordgraph ""
}

## ----------------------------------------------------------------------------
## _traceProcess
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_traceProcess {} {

  # check for activity
  set row [_getRowOfWordgraph $_selectedContextWordgraph]
  if {![$itk_component(table) tag includes runner $row,0]} {
    _viewLogfile
    return
  }

  # search job of selected wordgraph
  set found 0
  set experiment [_getSelectedExperiment]
  foreach job $_currentJobs($experiment) {
    if {[$job cget -wordgraphName] == $_selectedContextWordgraph} {
      set found 1
      break;
    }
  }
  if {!$found} {
    print "ERROR: wordgraph $wordgraphName not found for [$experiment cget -name]\n"
    return
  }

  # get the corresponding process
  set process [$job cget -process]

  if {$process != ""} {

    # switch off the last progess
    if {$_progressProcess != ""} {
      if {[info commands $_progressProcess] != ""} {
	set oldJob [$_progressProcess cget -job]
	print  "INFO: switching off trace of [$oldJob cget -wordgraphName]\n"
	$_progressProcess configure -showProgress 0
      } else {
      }
    }

    # switch on new trace
    if {$_progressProcess != $process} {
      set _progressProcess $process
      set experiment [$job cget -experiment]
      set logFile [$experiment getLogFileName $_selectedContextWordgraph]
      regsub {^(.*)\.gz$} $logFile {\1} clearLogFile
      set fd [open "$clearLogFile" r]
      set buffer [read $fd]
      close $fd
      $itk_component(log) clear
      print "$buffer\n"
      $process configure -showProgress 1
      print  "INFO: switching on trace of [$job cget -wordgraphName]\n"
    } else {
      set _progressProcess ""
    }
  }

  set _selectedContextWordgraph ""
}

## ----------------------------------------------------------------------------
## _deleteData
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_deleteData {} {
  
  # check
  set experiment [_getSelectedExperiment]
  if {$experiment == ""} {
    return
  }

  # are there files available
  set fileNames ""
  set fileName [$experiment getLogFileName $_selectedContextWordgraph]
  if {[file exists $fileName]} {
    lappend fileNames $fileName
  }
  set fileName [$experiment getXmlFileName $_selectedContextWordgraph]
  if {[file exists $fileName]} {
    lappend fileNames $fileName
  }

  if {[llength $fileNames] == 0} {
    printMessage "File not found, nothing to delete!" "error"
    return
  }

  # ask for confirmation
  set msg [iwidgets::messagedialog .msg \
    -text \
"Are you sure that you want to delete 
[join $fileNames \ and\ \n] ?" \
    -title "Yada - Question"  \
    -image [YadaImages::get question] \
    -modality application \
    -master $itk_interior]
    
  $msg buttonconfigure OK -text "Yes"
  $msg buttonconfigure Cancel -text "No"
  $msg center .
  
  set result [$msg activate]

  destroy $msg
  if {!$result} {
    return 0
  }

  foreach fileName $fileNames {
    file delete $fileName
  }
}

## ----------------------------------------------------------------------------
## setToggle
## ----------------------------------------------------------------------------
body YadaRunnableDocument::setToggle {name {value ""}} {
  switch $name {
    "table" {
      if {$value == 0 || $value == 1} {
	set _tableFlag $value
      }
      if {$_tableFlag} {
	$itk_component(pane) show 0
	if {$_consoleFlag} {
	  $itk_component(pane) fraction 80 20
	} 
      } else {
	$itk_component(pane) hide 0
      }
      return $_tableFlag
    }
    "console" {
      if {$value == 0 || $value == 1} {
	set _consoleFlag $value
      }
      if {$_consoleFlag} {
	$itk_component(pane) show 1
	if {$_tableFlag} {
	  $itk_component(pane) fraction 80 20
	} 
      } else {
	$itk_component(pane) hide 1
      }
      return $_consoleFlag
    }
    "follow" {
      if {$value == 0 || $value == 1} {
	set _followDataFlag $value
      }
      $itk_component(log) configure -followData $_followDataFlag
      return $_followDataFlag
    }
    "clear" {
      if {$value == 0 || $value == 1} {
	set _clearLogFlag $value
      }
    }
    "orientation" {
      if {$value == "horizontal" || $value == "vertical"} {
	set _orientFlag $value
      }
      if {$_orientFlag == "horizontal"} {
	$itk_component(pane) configure \
	  -orient $_orientFlag \
	  -sashheight 8 -sashwidth 18
      } else {
	$itk_component(pane) configure \
	  -orient $_orientFlag \
	  -sashheight 18 -sashwidth 8
      }
    }
    default {
      return [chain $name $value]
    }
  }
}

## ----------------------------------------------------------------------------
## _resetDataOf
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_resetDataOf {wordgraphs} {
  foreach wordgraph $wordgraphs {
    set _doneWordgraph($wordgraph) 0
  }
}


## ----------------------------------------------------------------------------
## Returns wordgraph file name for selected experiment
## ----------------------------------------------------------------------------
body YadaRunnableDocument::_getXMLFileName { wordgraph } {

  set experimentName [$itk_component(experimentBox) get]
  set experiment [.main getExperiment $experimentName]
   
  return [$experiment getXmlFileName $wordgraph]

}


## ----------------------------------------------------------------------------
## Overridden method
## ----------------------------------------------------------------------------
body YadaRunnableDocument::save {} {
    chain
    print "INFO: Save document to `[getPersistanceFileName]'\n"
}